# file: wsj/scripts/perl/wsj_subroutines/wsj_subs.pm
#
# this file contains the subroutines that pertain to running experiments with 
# the prototype system. Typically these will be wrappers around the prototype
# system functions. Note that all functions in this package return 0 if there
# is an error and 1 otherwise. This is chosen to coincide with the notion of
# true and false. Note that this is different than the return values for 
# system calls where 1 is an error and 0 is no error
#

# system libraries
#
use FileHandle;
autoflush STDOUT;

# global variables
#

# define the utilities that will be needed for this script
#
$CREATE_MODELS = "/cavs/hse/ies/isip/tools/proto/bin/scripts/create_models";
$CREATE_TRIPHONE_MAP = "/cavs/hse/ies/isip/tools/proto/bin/scripts/create_triphone_map";
$BW_TRAIN = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/bw_train";
$CONVERT_MMF = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/convert_mmf";
$CREATE_MIXTURES = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/create_mixtures";
$HMM_TRAIN = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/hmm_train";
$INIT_HMM = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/init_hmm";
$INIT_TRIPHONES = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/init_triphones";
$TIE_STATE = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/tie_state";
$TRACE_PROJECTOR = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/trace_projector";
$GRAMMAR_COMPILER = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/grammar_compiler";
$EXTRACT_FEATURE = "/cavs/hse/ies/isip/tools/proto/bin/i686-pc-linux-gnu/extract_feature";
$ISIP_EVAL = "/cavs/hse/ies/isip/users/raghavan/tools_r12_00/bin/scripts/isip_eval";

# define modes for the utilities
#
$TI_BINARY_MODE = "binary";
$TI_ASCII_MODE = "ascii";


#----------------------------------------------------------------------------
#
# wrappers around prototype utilities
#
#----------------------------------------------------------------------------

# subroutine: run_extract_feature
#
# arguments:
#  @_ : (delta, acc, cms, use_energy, zero_mean, energy_norm, num_mfcc,
#       delta_window, lifter_coeff, window_dur, frame_dur, num_fbanks,
#	pre_emph_coeff, swap_bytes, sample_freq, window_type, output_mode,
#       input_file_list, output_file_list, $cpu)
#
# description: this routine runs extract_feature with the given inputs
#
# return: 0 if error 1 if no error
#
sub run_extract_feature ($$$$$$$$$$$$$$$$$$$$) {

    # parse the argument list
    #
    my $delta = shift @_;
    my $acc = shift @_;
    my $cms = shift @_;
    my $use_energy = shift @_;
    my $zero_mean = shift @_;
    my $energy_norm = shift @_;
    my $num_mfcc = shift @_;
    my $delta_window = shift @_;
    my $lifter_coeff = shift @_;
    my $window_dur = shift @_;
    my $frame_dur = shift @_;
    my $num_fbanks = shift @_;
    my $pre_emph_coeff = shift @_;
    my $swap_bytes = shift @_;
    my $sample_freq = shift @_;
    my $window_type = shift @_;
    my $output_mode = shift @_;
    my $input_file_list = shift @_;
    my $output_file_list = shift @_;
    my $cpu = shift @_;

    if (($output_file_list eq "") || ($input_file_list eq "") 
	|| ($cpu eq "")) {
	print "ERROR: run_extract_feature - bad argument list\n";
	return 0;
    }

    # build the argument list
    #
    @args = ($EXTRACT_FEATURE);

    # check for the optional flags
    #
    if ($delta ne "") {
	push (@args, "-delta");
    }
    if ($acc ne "") {
	push (@args, "-acc");
    }
    if ($cms ne "") {
	push (@args, "-cms");
    }
    if ($use_energy ne "") {
	push (@args, "-energy");
    }    
    if ($zero_mean ne "") {
	push (@args, "-zero_mean");
    }    
    if ($energy_norm ne "") {
	push (@args, "-energy_norm");
    }    
    if ($swap_bytes ne "") {
	push (@args, "-swap_bytes");
    }    

    # check for the optional arguments
    #
    if ($num_mfcc ne "") {
	push (@args, "-mfcc", $num_mfcc);
    }
    if ($delta_window ne "") {
	push (@args, "-delta_win", $delta_window);
    }
    if ($lifter_coeff ne "") {
	push (@args, "-lifter_coeff", $lifter_coeff);
    }
    if ($window_dur ne "") {
	push (@args, "-window_dur", $window_dur);
    }
    if ($frame_dur ne "") {
	push (@args, "-frame_dur", $frame_dur);
    }
    if ($num_fbanks ne "") {
	push (@args, "-num_fbanks", $num_fbanks);
    }
    if ($pre_emph_coeff ne "") {
	push (@args, "-pre_emph_coeff", $pre_emph_coeff);
    }
    if ($sample_freq ne "") {
	push (@args, "-sf", $sample_freq);
    }
    if ($window_type ne "") {
	push (@args, "-window_type", $window_type);
    }
    if ($output_mode ne "") {
	push (@args, "-output_mode", $output_mode);
    }

    # add the input and output list to the command line
    #
    push (@args, "-input", $input_file_list);    
    push (@args, "-output", $output_file_list);    

    # display the command 
    #
    $out = join(" ", @args);
    print "COMMAND: rsh $cpu $out\n";
    
    # run extract_feature
    #
    ((`rsh $cpu $out`) == 0) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: run_init_triphones
#
# arguments:
#  @_ : (monophones_file, input_models, input_states, input_trans, 
#        cluster_list, output_models, output_states_file, output_param_file,
#        $cpu)
#
# description: this routine initializes triphone models using the
# init_triphones utility
#
# return: 0 if error 1 if no error
#
sub run_init_triphones ($$$$$$$$$) {

    # parse the argument list
    #
    my $monophones_file = shift @_;
    my $input_models_file = shift @_;
    my $input_states_file = shift @_;
    my $input_trans_file = shift @_;
    my $cluster_list = shift @_;
    my $output_models_file = shift @_;
    my $output_states_file = shift @_;
    my $output_param_file = shift @_;
    my $cpu = shift @_;

    if (($output_param_file eq "") || ($cpu eq ""))  {
	print "ERROR: run_init_triphones - bad argument list\n";
	return 0;
    }

    # parse the paramater file
    #
    $output_param_string = $ITRI_PARAM_FILE_TEMPLATE;

    # try to open the output parameter file
    #
    (isip_open(fp_out, ">$output_param_file") == 1) or return 0;

    # loop over all lines and do regular expression replacements to create the
    # new file
    #
    $output_param_string =~ s/__CI_PHONES_FILE__/$monophones_file/g;
    $output_param_string =~ s/__IN_MODELS_FILE__/$input_models_file/g;
    $output_param_string =~ s/__IN_STATES_FILE__/$input_states_file/g;
    $output_param_string =~ s/__IN_TRANS_FILE__/$input_trans_file/g;
    $output_param_string =~ s/__TRI_CLIST_FILE__/$cluster_list/g;
    $output_param_string =~ s/__OUT_MODELS_FILE__/$output_models_file/g;
    $output_param_string =~ s/__OUT_STATES_FILE__/$output_states_file/g;
    print fp_out "$output_param_string";

    # close the parameter file
    #
    (isip_close(fp_out) == 1) or return 0;

    # build the argument list for init_triphones
    #
    @args = ($INIT_TRIPHONES, "-param", $output_param_file);
    $out = join(" ", @args);
    print "COMMAND: rsh $cpu $out\n";
    
    # run bw_train
    #
    if (((`rsh $cpu $out`) != 0) ||
	(file_exists($output_states_file) != 1) ||
	(file_exists($output_models_file) != 1)) {
	
	return 0;
    }

    # exit gracefully
    #
    return 1;
}


# subroutine: run_tie_state
#
# arguments:
#  @_ : (monophones_file, input_models, input_states, occupancy_file,
#        output_trees, output_models, output_states_file, output_cluster_list,
#        input_mode, output_mode, questions_file, special_models_file, 
#        split_thresh, merge_thresh, num_occ_thresh, num_context,
#        output_param_file, $cpu)
#
# description: this routine performs state-tying
# 
# return: 0 if error 1 if no error
#
sub run_tie_state_train ($$$$$$$$$$$$$$$$$$) {

    # get parameters from arg list
    #
    my $monophones_file = shift @_;
    my $input_models_file = shift @_;
    my $input_states_file = shift @_;
    my $occupancy_file = shift @_;
    my $output_trees_file = shift @_;
    my $output_models_file = shift @_;
    my $output_states_file = shift @_;
    my $output_cluster_list = shift @_;
    my $input_mode = shift @_;
    my $output_mode = shift @_;
    my $questions_file = shift @_;
    my $special_models_file = shift @_;
    my $split_thresh = shift @_;
    my $merge_thresh = shift @_;
    my $num_occ_thresh = shift @_;
    my $num_context = shift @_;
    my $output_param_file = shift @_;
    my $cpu = shift @_;

    if (($output_param_file eq "") || ($cpu eq ""))  {
	print "ERROR: run_tie_state_train - bad argument list\n";
	return 0;
    }

    # try to open the output parameter file
    #
    (isip_open(fp_out, ">$output_param_file") == 1) or return 0;

    # loop over all lines and do regular expression replacements to create the
    # new file
    #
    $output_param_string = $TS_TRAIN_PARAM_FILE_TEMPLATE;
    $output_param_string =~ s/__MONOPHONES_FILE__/$monophones_file/g;
    $output_param_string =~ s/__MEMBERSHIP_FILE__/$questions_file/g;
    $output_param_string =~ s/__INPUT_MODE__/$input_mode/g;
    $output_param_string =~ s/__OUTPUT_MODE__/$output_mode/g;
    $output_param_string =~ s/__IN_STATES_FILE__/$input_states_file/g;
    $output_param_string =~ s/__IN_MODELS_FILE__/$input_models_file/g;
    $output_param_string =~ s/__SPECIAL_MODELS_FILE__/$special_models_file/g;
    $output_param_string =~ s/__OCCUPANCY_FILE__/$occupancy_file/g;
    $output_param_string =~ s/__SPLIT_THRESHOLD__/$split_thresh/g;
    $output_param_string =~ s/__MERGE_THRESHOLD__/$merge_thresh/g;
    $output_param_string =~ s/__OCC_THRESHOLD__/$num_occ_thresh/g;
    $output_param_string =~ s/__NUM_CONTEXT__/$num_context/g;
    $output_param_string =~ s/__OUTPUT_TREES_FILE__/$output_trees_file/g;
    $output_param_string =~ s/__OUTPUT_STATES_FILE__/$output_states_file/g;
    $output_param_string =~ s/__OUTPUT_MODELS_FILE__/$output_models_file/g;
    $output_param_string =~ s/__OUTPUT_CLIST_FILE__/$output_cluster_list/g;
    print fp_out "$output_param_string";
    
    # close the parameter file
    #
    (isip_close(fp_out) == 1) or return 0;

    # build the argument list for tie_state
    #
    @args = ($TIE_STATE, "-p", $output_param_file);
    $out = join(" ", @args);
    print "COMMAND: rsh $cpu $out\n";
    
    # run tie_state
    #
    if (((`rsh $cpu $out`) != 0) ||
	(file_exists($output_trees_file) != 1) ||
	(file_exists($output_cluster_list) != 1) ||
	(file_exists($output_states_file) != 1) ||
	(file_exists($output_models_file) != 1)) {

	return 0;
    }

    # exit gracefully
    #
    return 1;
}


# subroutine: run_tie_state_test
#
# arguments:
#  @_ : (monophones_file, input_models, input_states, input_trees,
#        input_model_list, output_models, output_cluster_list, input_mode,
#        output_mode, questions_file, special_models_file, split_thresh,
#        merge_thresh, num_occ_thresh, num_context, output_param_file, $cpu)
#
# description: this routine performs state-tying in test mode
# 
# return: 0 if error 1 if no error
#
sub run_tie_state_test ($$$$$$$$$$$$$$$$$) {

    # get parameters from arg list
    #
    my $monophones_file = shift @_;
    my $input_models_file = shift @_;
    my $input_states_file = shift @_;
    my $input_trees_file = shift @_;
    my $input_model_list = shift @_;
    my $output_models_file = shift @_;
    my $output_cluster_list = shift @_;
    my $input_mode = shift @_;
    my $output_mode = shift @_;
    my $questions_file = shift @_;
    my $special_models_file = shift @_;
    my $split_thresh = shift @_;
    my $merge_thresh = shift @_;
    my $num_occ_thresh = shift @_;
    my $num_context = shift @_;
    my $output_param_file = shift @_;
    my $cpu = shift @_;

    if (($output_param_file eq "") || ($cpu eq ""))   {
	print "ERROR: run_tie_state_test - bad argument list\n";
	return 0;
    }

    # try to open the output parameter file
    #
    (isip_open(fp_out, ">$output_param_file") == 1) or return 0;

    # loop over all lines and do regular expression replacements to create the
    # new file
    #
    $output_param_string = $TS_TEST_PARAM_FILE_TEMPLATE;
    $output_param_string =~ s/__MONOPHONES_FILE__/$monophones_file/g;
    $output_param_string =~ s/__MEMBERSHIP_FILE__/$questions_file/g;
    $output_param_string =~ s/__INPUT_MODE__/$input_mode/g;
    $output_param_string =~ s/__OUTPUT_MODE__/$output_mode/g;
    $output_param_string =~ s/__IN_STATES_FILE__/$input_states_file/g;
    $output_param_string =~ s/__IN_MODELS_FILE__/$input_models_file/g;
    $output_param_string =~ s/__SPECIAL_MODELS_FILE__/$special_models_file/g;
    $output_param_string =~ s/__SPLIT_THRESHOLD__/$split_thresh/g;
    $output_param_string =~ s/__MERGE_THRESHOLD__/$merge_thresh/g;
    $output_param_string =~ s/__OCC_THRESHOLD__/$num_occ_thresh/g;
    $output_param_string =~ s/__NUM_CONTEXT__/$num_context/g;
    $output_param_string =~ s/__INPUT_TREE_FILE__/$input_trees_file/g;
    $output_param_string =~ s/__INPUT_PHONE_LIST_FILE__/$input_model_list/g;
    $output_param_string =~ s/__OUTPUT_STATES_FILE__/$output_states_file/g;
    $output_param_string =~ s/__OUTPUT_MODELS_FILE__/$output_models_file/g;
    $output_param_string =~ s/__OUTPUT_CLIST_FILE__/$output_cluster_list/g;
    print fp_out "$output_param_string";

    # close the parameter file
    #
    (isip_close(fp_out) == 1) or return 0;

    # build the argument list for tie_state
    #
    @args = ($TIE_STATE, "-p", $output_param_file);
    $out = join(" ", @args);
    print "COMMAND: rsh $cpu $out\n";
    
    # run tie_state
    #
    if (((`rsh $cpu $out`) != 0) ||
	(file_exists($output_cluster_list) != 1) ||
	(file_exists($output_models_file) != 1)) {

	return 0;
    }

    # exit gracefully
    #
    return 1;
}


# subroutine: run_hmm_train
#
# arguments:
#  @_ : (train_mode, batch_stats_file, acc_list_file, mfcc_list, output_list,
#        mlf_file, mlf_mode, monophones_file, special_models_file,
#        models_file, phones_file, lexicon_file, input_states, input_trans,
#        output_states, output_trans, align_mode, input_mode, output_mode,
#        context_mode, nbest_size, context_size, state_beam, model_beam,
#        word_beam, mapmi_limit, occupancy_file, output_param_file)
#
# description: this routine creates an hmm_train parameter file from a template
#              and then runs hmm_train
#
# return: 0 if error 1 if no error
#
sub run_hmm_train ($$$$$$$$$$$$$$$$$$$$$$$$$$$) {

    # parse the argument list
    #
    my $train_mode = shift @_;
    my $batch_stats_file = shift @_;
    my $acc_list_file = shift @_;
    my $mfcc_list = shift @_;
    my $output_list = shift @_;
    my $mlf_file = shift @_;
    my $mlf_mode = shift @_;
    my $monophones_file = shift @_;
    my $special_models_file = shift @_;
    my $models_file = shift @_;
    my $phones_file = shift @_;
    my $lexicon_file = shift @_;
    my $input_states = shift @_;
    my $input_trans = shift @_;
    my $output_states = shift @_;
    my $output_trans = shift @_;
    my $align_mode = shift @_;
    my $input_mode = shift @_;
    my $output_mode = shift @_;
    my $context_mode = shift @_;
    my $nbest_size = shift @_;
    my $context_size = shift @_;
    my $state_beam = shift @_;
    my $model_beam = shift @_;
    my$word_beam = shift @_;
    my $mapmi_limit = shift @_;
    my $occupancy_file = shift @_;
    my $output_param_file = shift @_;

    if ($output_param_file eq "")  {
	print "ERROR: run_hmm_train - bad argument list\n";
	return 0;
    }

    my $context_flag = "CI";
    if ($context_mode ne "monophone") {
	$context_flag = "CD";
    }

    my $occ_mode = "off";
    if ($occupancy_file ne "") {
	$occ_mode = "on";
    }

    # try to open the output parameter file
    #
    (isip_open(fp_out, ">$output_param_file") == 1) or return 0;

    # loop over all lines and do regular expression replacements to create the
    # new file
    #
    $output_params_string = $HT_PARAM_FILE_TEMPLATE;
    $output_params_string =~ s/__TRAIN_MODE__/$train_mode/g;
    $output_params_string =~ s/__BATCH_STATS_FILE__/$batch_stats_file/g;
    $output_params_string =~ s/__ACCUMULATOR_LIST_FILE__/$acc_list_file/g;
    $output_params_string =~ s/__MFCC_LIST__/$mfcc_list/g;
    $output_params_string =~ s/__OUTPUT_LIST__/$output_list/g;
    $output_params_string =~ s/__MLF_FILE__/$mlf_file/g;
    $output_params_string =~ s/__MLF_MODE__/$mlf_mode/g;
    $output_params_string =~ s/__MONOPHONES_FILE__/$monophones_file/g;
    $output_params_string =~ s/__SPECIAL_MODELS_FILE__/$special_models_file/g;
    $output_params_string =~ s/__MODELS_FILE__/$models_file/g;
    $output_params_string =~ s/__PHONES_FILE__/$phones_file/g;
    $output_params_string =~ s/__LEXICON_FILE__/$lexicon_file/g;
    $output_params_string =~ s/__INPUT_STATES__/$input_states/g;
    $output_params_string =~ s/__INPUT_TRANSITIONS__/$input_trans/g;
    $output_params_string =~ s/__OUTPUT_STATES__/$output_states/g;
    $output_params_string =~ s/__OUTPUT_TRANSITIONS__/$output_trans/g;
    $output_params_string =~ s/__ALIGN_MODE__/$align_mode/g;
    $output_params_string =~ s/__INPUT_MODE__/$input_mode/g;
    $output_params_string =~ s/__OUTPUT_MODE__/$output_mode/g;
    $output_params_string =~ s/__CONTEXT_MODE__/$context_mode/g;
    $output_params_string =~ s/__NBEST_SIZE__/$nbest_size/g;
    $output_params_string =~ s/__CONTEXT_SIZE__/$context_size/g;
    $output_params_string =~ s/__STATE_BEAM_WIDTH__/$state_beam/g;
    $output_params_string =~ s/__MODEL_BEAM_WIDTH__/$model_beam/g;
    $output_params_string =~ s/__WORD_BEAM_WIDTH__/$word_beam/g;
    $output_params_string =~ s/__MAPMI_THRESH__/$mapmi_limit/g;
    $output_params_string =~ s/__STATE_OCCUPANCY_FILE__/$occupancy_file/g;
    $output_params_string =~ s/__STATE_OCC_MODE__/$occ_mode/g;
    print fp_out "$output_params_string";

    # close the parameter file
    #
    (isip_close(fp_out) == 1) or return 0;

    # build the argument list for hmm_train
    #
    @args = ($HMM_TRAIN, "-p", $output_param_file, "-c", $context_flag);
    $out = join(" ", @args);
    print "COMMAND: $out\n";
    
    # run hmm_train
    #
    if (system(@args) != 0) {
	print "ERROR: run_hmm_train - call to hmm_train failed\n";
	return 0;
    }

    if ($train_mode ne "batch") {
	if ((file_exists($output_states) != 1) ||
	    (file_exists($output_trans) != 1)) {
	    return 0;
	}
    }

    # exit gracefully
    #
    return 1;
}


# subroutine: run_bw_train
#
# arguments:
#  @_ : (train_mode, mfcc_list, label_list, monophones_file, models_file,
#        phones_file, lexicon_file, input_states, input_trans, output_states,
#        output_trans, input_mode, output_mode, context_mode, min_mpd,
#        beam_width, min_occupancy, min_model_count, vfloor_file,
#        state_occ_file, acc_file, acc_list_file, output_param_file, 
#        input_feature_format, compute_delta, compute_acc, delta_win, cpu,
#        shell_script)
#
# description: this routine creates a bw_train parameter file from a template
#              and then runs bw_train
#
# return: 0 if error 1 if no error
#
sub run_bw_train ($$$$$$$$$$$$$$$$$$$$$$$$$$$$$$) {

    # get parameters from arg list
    #
    my $train_mode = shift @_;
    my $mfcc_list = shift @_;
    my $label_list = shift @_;
    my $monophones_file = shift @_;
    my $models_file = shift @_;
    my $phones_file = shift @_;
    my $lexicon_file = shift @_;
    my $input_states = shift @_;
    my $input_trans = shift @_;
    my $output_states = shift @_;
    my $output_trans = shift @_;
    my $input_mode = shift @_;
    my $output_mode = shift @_;
    my $context_mode = shift @_;
    my $min_mpd = shift @_;
    my $beam_width = shift @_;
    my $min_occupancy = shift @_;
    my $min_model_count = shift @_;
    my $vfloor_file = shift @_;
    my $state_occ_file = shift @_;
    my $acc_file = shift @_;
    my $acc_list_file = shift @_;
    my $output_param_file = shift @_;
    my $input_feature_format = shift @_;
    my $compute_delta = shift @_;
    my $compute_acc = shift @_;
    my $delta_win = shift @_;
    my $cpu = shift @_;
    my $shell_script = shift @_;
    my $walltime = shift @_;
    
    if (($output_param_file eq "") || ($cpu eq ""))  {
	print "ERROR: run_bw_train - bad argument list\n";
	return 0;
    }

    $state_occ_mode = "off";
    if ($state_occ_file ne "") {
	$state_occ_mode = "on";
    }

    # try to open the output parameter file
    #
    (isip_open(fp_out, ">$output_param_file") == 1) or return 0;

    # loop over all lines and do regular expression replacements to create the
    # new file
    #
    $param_file_string = $BW_TRAIN_PARAM_FILE_TEMPLATE;
    $param_file_string =~ s/__TRAIN_MODE__/$train_mode/g;
    $param_file_string =~ s/__ACCUMULATOR_FILE__/$acc_file/g;
    $param_file_string =~ s/__ACCUMULATOR_LIST_FILE__/$acc_list_file/g;
    $param_file_string =~ s/__MFCC_LIST__/$mfcc_list/g;
    $param_file_string =~ s/__LABEL_LIST__/$label_list/g;
    $param_file_string =~ s/__MONOPHONES_FILE__/$monophones_file/g;
    $param_file_string =~ s/__MODELS_FILE__/$models_file/g;
    $param_file_string =~ s/__PHONES_FILE__/$phones_file/g;
    $param_file_string =~ s/__LEXICON_FILE__/$lexicon_file/g;
    $param_file_string =~ s/__INPUT_STATES__/$input_states/g;
    $param_file_string =~ s/__INPUT_TRANSITIONS__/$input_trans/g;
    $param_file_string =~ s/__OUTPUT_STATES__/$output_states/g;
    $param_file_string =~ s/__OUTPUT_TRANSITIONS__/$output_trans/g;
    $param_file_string =~ s/__INPUT_MODE__/$input_mode/g;
    $param_file_string =~ s/__OUTPUT_MODE__/$output_mode/g;
    $param_file_string =~ s/__CONTEXT_MODE__/$context_mode/g;
    $param_file_string =~ s/__MIN_MPD__/$min_mpd/g;
    $param_file_string =~ s/__BEAM_WIDTH__/$beam_width/g;
    $param_file_string =~ s/__MIN_OCC__/$min_occupancy/g;
    $param_file_string =~ s/__MIN_MODEL_COUNT__/$min_model_count/g;
    $param_file_string =~ s/__VFLOOR_FILE__/$vfloor_file/g;
    $param_file_string =~ s/__STATE_OCC_FILE__/$state_occ_file/g;
    $param_file_string =~ s/__STATE_OCC_MODE__/$state_occ_mode/g;
    $param_file_string =~ s/__INPUT_FEATURE_FORMAT__/$input_feature_format/g;
    $param_file_string =~ s/__COMPUTE_DELTA__/$compute_delta/g;
    $param_file_string =~ s/__COMPUTE_ACC__/$compute_acc/g;
    $param_file_string =~ s/__DELTA_WIN__/$delta_win/g;
    print fp_out "$param_file_string";

    # close the parameter file
    #
    (isip_close(fp_out) == 1) or return 0;

    # build the argument list for the call to bw_train
    #
    @args = ($BW_TRAIN, "-p", $output_param_file);
    $out = join(" ", @args);

    # run shell script that would run the bw_train in batch_mode
    #
    if ($train_mode eq "batch") {
	
        # print the debug message
	#
	print "COMMAND: sh $shell_script &\n";
	
	# run the training using the shell script
	#
	if ($walltime eq "") {
	    if (system("sh $shell_script &") != 0) {
		return 0;
	    }
	}
	if ($walltime ne "") {
	    if (system("qsub -j oe $shell_script &") != 0) {
		return 0;
	    }
	}
    }
    
    
    # run shell script that would run the bw_train in combine mode
    #
    if ($train_mode eq "combine") {
	
        # print the debug message
	#
	print "COMMAND: $out\n";
	
	# run the training using the shell script
	#
	if ((`$out`) != 0) {
	    return 0;
	}
	
	# check if the training was successful
	#
	if ((file_exists($output_states) != 1) || 
	    (file_exists($output_trans) != 1)) {
	    return 0;
	}
    }

    # exit gracefully
    #
    return 1;
}


# subroutine: run_trace_projector
#
# arguments:
#  @_ : (decode_mode, mfcc_list, output_list, input_lattice_list, 
#        transcription_file, lattice_wer_file, output_lattice_list,
#        grammar_list_file, monophones_file, models_file,
#        phones_file, lexicon_file, input_states, input_trans, align_mode,
#        lm_file, ngram_order, lm_scale, word_penalty, history_paths,
#        nbest_paths, input_format, context_mode, state_beam, 
#        model_beam, word_beam, mapmi_limit, num_allowed_words,
#        output_param_file, input_feature_format, compute_delta, compute_acc,
#        delta_win, shell_script)
#
# description: this routine creates a trace_projector parameter file from a
#              template and then runs trace_projector
#
# return: 0 if error 1 if no error
#
sub run_trace_projector ($$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$) {

    # get parameters from arg list
    #
    my $decode_mode = shift @_;
    my $mfcc_list = shift @_;
    my $output_list = shift @_;
    my $input_lattice_list = shift @_;
    my $transcription_file = shift @_;
    my $lattice_wer_file = shift @_;
    my $output_lattice_list = shift @_;
    my $grammar_list_file = shift @_;
    my $monophones_file = shift @_;
    my $models_file = shift @_;
    my $phones_file = shift @_;
    my $lexicon_file = shift @_;
    my $input_states = shift @_;
    my $input_trans = shift @_;
    my $align_mode = shift @_;
    my $lm_file = shift @_;
    my $ngram_order = shift @_;
    my $lm_scale = shift @_;
    my $word_penalty = shift @_;
    my $history_paths = shift @_;
    my $nbest_paths = shift @_;
    my $input_format = shift @_;
    my $context_mode = shift @_;
    my $state_beam = shift @_;
    my $model_beam = shift @_;
    my $word_beam = shift @_;
    my $mapmi_limit = shift @_;
    my $num_allowed_words = shift @_;
    my $output_param_file = shift @_;
    my $input_feature_format = shift @_;
    my $compute_delta = shift @_;
    my $compute_acc = shift @_;
    my $delta_win = shift @_;
    my $shell_script = shift @_;

    if (($output_param_file eq "") || ($shell_script eq ""))  {
	print "ERROR: run_trace_projector - bad argument list\n";
	return 0;
    }

    # try to open the output parameter file
    #
    (isip_open(fp_out, ">$output_param_file") == 1) or return 0;

    # loop over all lines and do regular expression replacements to create the
    # new file
    #
    $param_file_string = $TP_PARAM_FILE_TEMPLATE;
    $param_file_string =~ s/__DECODE_MODE__/$decode_mode/g;
    $param_file_string =~ s/__MFCC_LIST__/$mfcc_list/g;
    $param_file_string =~ s/__OUTPUT_LIST__/$output_list/g;
    $param_file_string =~ s/__INPUT_LATTICE_LIST__/$input_lattice_list/g;
    $param_file_string =~ s/__TRANSCRIPTION_FILE__/$transcription_file/g;
    $param_file_string =~ s/__LATTICE_WER_FILE__/$lattice_wer_file/g;
    $param_file_string =~ s/__OUTPUT_LATTICE_LIST__/$output_lattice_list/g;
    $param_file_string =~ s/__GRAMMAR_LIST_FILE__/$grammar_list_file/g;
    $param_file_string =~ s/__MONOPHONES_FILE__/$monophones_file/g;
    $param_file_string =~ s/__MODELS_FILE__/$models_file/g;
    $param_file_string =~ s/__PHONES_FILE__/$phones_file/g;
    $param_file_string =~ s/__LEXICON_FILE__/$lexicon_file/g;
    $param_file_string =~ s/__STATES_FILE__/$input_states/g;
    $param_file_string =~ s/__TRANSITIONS_FILE__/$input_trans/g;
    $param_file_string =~ s/__ALIGN_MODE__/$align_mode/g;
    $param_file_string =~ s/__LM_FILE__/$lm_file/g;
    $param_file_string =~ s/__NGRAM_ORDER__/$ngram_order/g;
    $param_file_string =~ s/__LM_SCALE__/$lm_scale/g;
    $param_file_string =~ s/__WORD_PENALTY__/$word_penalty/g;
    $param_file_string =~ s/__HISTORY_PATHS__/$history_paths/g;
    $param_file_string =~ s/__NBEST_PATHS__/$nbest_paths/g;
    $param_file_string =~ s/__INPUT_FORMAT__/$input_format/g;
    $param_file_string =~ s/__CONTEXT_MODE__/$context_mode/g;
    $param_file_string =~ s/__STATE_BEAM__/$state_beam/g;
    $param_file_string =~ s/__MODEL_BEAM__/$model_beam/g;
    $param_file_string =~ s/__WORD_BEAM__/$word_beam/g;
    $param_file_string =~ s/__MAPMI_LIMIT__/$mapmi_limit/g;
    $param_file_string =~ s/__NUM_ALLOWED_WORDS__/$num_allowed_words/g;
    $param_file_string =~ s/__INPUT_FEATURE_FORMAT__/$input_feature_format/g;
    $param_file_string =~ s/__COMPUTE_DELTA__/$compute_delta/g;
    $param_file_string =~ s/__COMPUTE_ACC__/$compute_acc/g;
    $param_file_string =~ s/__DELTA_WIN__/$delta_win/g;
    print fp_out "$param_file_string";

    # close the parameter file
    #
    (isip_close(fp_out) == 1) or return 0;

    # build the argument list for the call to bw_train
    #
    @args = ($TRACE_PROJECTOR, "-p", $output_param_file);
    $out = join(" ", @args);
    print "COMMAND: sh $shell_script &\n";
    
    # run trace_projector
    #
    (system("sh $shell_script &") == 0) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: run_create_models
#
# arguments:
#  @_ : (num_states_file, output_models_file)
#
# description: this routine runs create_models and checks for error
# conditions
#
# return: 0 if error 1 if no error
#
sub run_create_models ($$) {

    # parse argument list
    #
    my $num_states_file = shift @_;
    my $output_models_file = shift @_;

    # verify that all necessary parameters are specified
    #
    if (($num_states_file eq "") || ($output_models_file eq "")) {
	print "ERROR: run_create_models - bad argument list\n";
	return 0;
    }

    # make sure the states file exists
    #
    if (file_exists($num_states_file) != 1) {
	print "ERROR: run_create_models - $num_states_file not found\n";
	return 0;
    }

    # build the argument list for the call to create_models
    #
    @args = ($CREATE_MODELS, "-states", "$num_states_file", "-output",
	     "$output_models_file");
    $out = join(" ", @args);
    print "COMMAND: $out\n";
    
    # run create_models on the cpu
    #
    if ((system(@args) != 0) || 
	(file_exists($output_models_file) != 1)) {
	print "ERROR: run_create_models - call to create_models failed\n";
	return 0;
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: run_create_shell_script_train
#
# arguments:
#  @_ : (shell_script, param_file, log_file, completion_file, cpu)
#
# description: this creates shell script to run bw_train using rsh
#
# return: 0 if error 1 if no error
#
sub run_create_shell_script_train ($$$$$$) {

    # parse argument list
    #
    my $shell_script = shift @_;
    my $param_file = shift @_;
    my $log_file = shift @_;
    my $completion_file = shift @_;
    my $cpu = shift @_;
    my $walltime = shift @_;
    
    # verify that all necessary parameters are specified
    #
    if (($shell_script eq "") || ($param_file eq "") || ($log_file eq "") || 
	($cpu eq "")) {
	print "ERROR: run_create_shell_script - bad argument list\n";
	return 0;
    }
    
    # open the shell script
    #
    (isip_open(fp_shell, ">$shell_script") == 1) or return 0;
    
    # for cluster the script is different
    #
    
    if ($walltime eq "") {
	
	# build the shell script
	#
	print fp_shell "#!/bin/sh\n";
	print fp_shell "if (test -f $completion_file)\n"; 
	print fp_shell "then\n";
	print fp_shell "rm $completion_file\n";
	print fp_shell "fi\n";
	print fp_shell "rsh $cpu nice -19 $BW_TRAIN -p $param_file > $log_file\n";
	print fp_shell "touch $completion_file\n";
    }
    if ($walltime ne "") {
	
	# build the shell script
	#
	print fp_shell "#!/bin/sh\n";
	print fp_shell "#PBS -V\n";
	print fp_shell "#PBS -N wsj_train\n";
	print fp_shell "#PBS -W umask=002\n";
	print fp_shell "#PBS -l nodes=1:ppn=1\n";
	print fp_shell "#PBS -l walltime=$walltime:00:00\n";
	print fp_shell "cd $PBS_O_WORKDIR\n";
	print fp_shell "if (test -f $completion_file)\n"; 
	print fp_shell "then\n";
	print fp_shell "rm $completion_file\n";
	print fp_shell "fi\n";
	print fp_shell "$BW_TRAIN -p $param_file > $log_file\n";
	print fp_shell "touch $completion_file\n";
    }
    
    # close the shell script
    #
    (isip_close(fp_shell) == 1) or return 0;
    
    # exit gracefully
    #
    return 1;
}

# subroutine: run_create_shell_script_test
#
# arguments:
#  @_ : (shell_script, param_file, current_path, 
#        log_file, completion_file, cpu)
#
# description: this creates shell script to run trace_projector using rsh
#
# return: 0 if error 1 if no error
#
sub run_create_shell_script_test ($$$$$$) {

    # parse argument list
    #
    my $shell_script = shift @_;
    my $param_file = shift @_;
    my $current_path = shift @_;
    my $log_file = shift @_;
    my $completion_file = shift @_;
    my $cpu = shift @_;
    
    # verify that all necessary parameters are specified
    #
    if (($shell_script eq "") || ($param_file eq "") || ($current_path eq "")
	|| ($log_file eq "") || ($cpu eq "")) {
	print "ERROR: run_create_shell_script - bad argument list\n";
	return 0;
    }
    
    # open the shell script
    #
    (isip_open(fp_shell, ">$shell_script") == 1) or return 0;
    
    # build the shell script
    #
    print fp_shell "#!/bin/sh\n";
    print fp_shell "if (test -f $completion_file)\n"; 
    print fp_shell "then\n";
    print fp_shell "rm $completion_file\n";
    print fp_shell "fi\n";
    print fp_shell "rsh $cpu \"cd $current_path;nice -19 $TRACE_PROJECTOR -p $param_file > $log_file\"\n";
    print fp_shell "touch $completion_file\n";

    # close the shell script
    #
    (isip_close(fp_shell) == 1) or return 0;
    
    # exit gracefully
    #
    return 1;
}

# subroutine: run_isip_eval
#
# arguments:
#  @_ : (mode, input_file_list, reference_file, hypothesis_file)
#
# description: this routine runs isip_eval and checks for error
#              conditions
#
# return: 0 if error 1 if no error
#
sub run_isip_eval ($$$$) {

    # parse argument list
    #
    my $mode = shift @_;
    my $input_file_list = shift @_;
    my $reference_file = shift @_;
    my $hypothesis_file = shift @_;

    # verify that all necessary parameters are specified
    #
    if (($mode eq "") || ($input_file_list eq "") ||
	($reference_file eq "") || ($hypothesis_file eq "")) {
	print "ERROR: run_isip_eval - bad argument list\n";
	return 0;
    }

    # build the argument list for the call to isip_eval
    #
    @args = ($ISIP_EVAL, $mode, $input_file_list, $reference_file, 
	     $hypothesis_file);
    $out = join(" ", @args);
    print "COMMAND: $out\n";
    
    # run isip_eval
    #
    if ((system(@args) != 0) || 
	(file_exists("$hypothesis_file.score") != 1)) {
	print "ERROR: run_isip_eval - call to isip_eval failed\n";
	return 0;
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: run_grammar_compiler
#
# arguments:
#  @_ : (input_grammar, output_lattice, $cpu)
#
# description: this routine runs grammar_compiler to create a lattice from a 
#              grammar
# conditions
#
# return: 0 if error 1 if no error
#
sub run_grammar_compiler ($$$) {

    # parse argument list
    #
    my $input_grammar = shift @_;
    my $output_lattice = shift @_;
    my $cpu = shift @_;

    # verify that all necessary parameters are specified
    #
    if (($input_grammar eq "") || ($output_lattice eq "") || 
	($cpu eq "")) {
	print "ERROR: run_grammar_compiler - bad argument list\n";
	return 0;
    }

    # build the argument list for the call to grammar_compiler
    #
    @args = ($GRAMMAR_COMPILER, "-input", $input_grammar, "-output",
	     $output_lattice);	
    $out = join(" ", @args);
    print "COMMAND: rsh $cpu $out\n";
    
    # run grammar_compiler
    #
    if (((`rsh $cpu $out`) != 0) || 
	(file_exists($output_lattice) != 1)) {
	print "ERROR: run_grammar_compiler - call to grammar_compiler";
	print " failed\n";
	return 0;
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: run_create_triphone_map
#
# arguments:
#  @_ : (monophones_list, cluster_file, context_mode, models_file, 
#        output_phones_file)
#
# description: this routine runs create_triphone_map and checks for error
# conditions
#
# return: 0 if error 1 if no error
#
sub run_create_triphone_map ($$$$$) {

    # parse the argument list
    #
    my $monophones_file = shift @_;
    my $cluster_file = shift @_;
    my $context_mode = shift @_;
    my $models_file = shift @_;
    my $output_phones_file = shift @_;

    if (($monophones_file eq "") || ($cluster_file eq "") ||
	($context_mode eq "") || ($models_file eq "") ||
	($output_phones_file eq "")) {

	print "ERROR: run_create_triphone_map - bad argument list\n";
	return 0;
    }

    # build the argument list for the call to create_triphone_map
    #
    @args = ($CREATE_TRIPHONE_MAP, "-mono", "$monophones_file", "-clist",
	     "$cluster_file", "-context", "$context_mode", "-models",
	     "$models_file",  "-output", "$output_phones_file");
    $out = join(" ", @args);
    print "COMMAND: $out\n";
    
    # run create_triphone_map
    #
    if ((system(@args) != 0) || 
	(file_exists("$output_phones_file") != 1)) {
	print "ERROR: run_create_triphone_map - call to create_triphone_map ";
	print "failed\n";
	return 0;
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: run_init_hmm
#
# arguments:
#  @_ : (mfcc_list, models_file, output_transitions_file, output_states_file,
#        input_mode, vfloor_file, vfloor, num_features, input_feature_format, 
#        $cpu)
#
# description: this routine runs init_hmm and checks for error conditions
#
# return: 0 if error 1 if no error
#
sub run_init_hmm ($$$$$$$$$$$$) {

    # parse the argument list
    #
    my $mfcc_list = shift @_;
    my $models_file = shift @_;
    my $output_transitions_file = shift @_;
    my $output_states_file = shift @_;
    my $input_mode = shift @_;
    my $vfloor_file = shift @_;
    my $vfloor = shift @_;
    my $num_features = shift @_;
    my $input_feature_format = shift @_;
    my $mode = shift @_;
    my $delta_win = shift @_;
    my $cpu = shift @_;

    if (($mfcc_list eq "") || ($models_file eq "") ||
	($output_transitions_file eq "") || ($output_states_file eq "") ||
	($input_mode eq "") || ($vfloor_file eq "") || ($vfloor eq "") ||
	($num_features eq "") || ($cpu eq "")) {

	print "ERROR: run_init_hmm - bad argument list\n";
	return 0;
    }

    @mode_format = ("-",$mode);
    $mode = join("",@mode_format);
	
    # make suitable arguments and run init_hmm for htk format
    #
    if ($input_feature_format eq "htk") {
	
        # build the input_feature_format and mode_format suitable for init_hmm
        #
        @feat_format = ("-",$input_feature_format);
	$input_feature_format = join("",@feat_format);
 
	# remove mode and delta_win if none from the init_hmm commandline
	#
	if ($mode eq "-none") {
	    # build the argument list for the call to init_hmm
	    #
	    @args = ($INIT_HMM, "-input", "$mfcc_list", "-models",
		     "$models_file", "-trans", "$output_transitions_file",
		     "-state", "$output_states_file", "-mode", "$input_mode",
		     "-vfloor_file", "$vfloor_file", "-var_floor", "$vfloor",
		     "-num_features", $num_features, $input_feature_format);
	    $out = join(" ", @args);
	    print "COMMAND: rsh $cpu $out\n";
	}
	
	elsif ($mode eq "-acc") {
	    # build the argument list for the call to init_hmm
	    #
	    @args = ($INIT_HMM, "-input", "$mfcc_list", "-models",
		     "$models_file", "-trans", "$output_transitions_file",
		     "-state", "$output_states_file", "-mode", "$input_mode",
		     "-vfloor_file", "$vfloor_file", "-var_floor", "$vfloor",
		     "-num_features", $num_features, $mode, "-delta", 
		     "-delta_win", $delta_win, $input_feature_format);
	    $out = join(" ", @args);
	    print "COMMAND: rsh $cpu $out\n";
	}
	
	elsif ($mode eq "-delta") {
	    # build the argument list for the call to init_hmm
	    #
	    @args = ($INIT_HMM, "-input", "$mfcc_list", "-models",
		     "$models_file", "-trans", "$output_transitions_file",
		     "-state", "$output_states_file", "-mode", "$input_mode",
		     "-vfloor_file", "$vfloor_file", "-var_floor", "$vfloor",
		     "-num_features", $num_features, $mode, 
		     "-delta_win", $delta_win, $input_feature_format);
	    $out = join(" ", @args);
	    print "COMMAND: rsh $cpu $out\n";    
	}
    }
    
    # make suitable arguments and run init_hmm for isip_proto format
    # 
    else {
	
	# remove mode and delta_win if none from the init_hmm commandline
	#
	if ($mode eq "-none") {
	    # build the argument list for the call to init_hmm
	    #
	    @args = ($INIT_HMM, "-input", "$mfcc_list", "-models",
		     "$models_file", "-trans", "$output_transitions_file",
		     "-state", "$output_states_file", "-mode", "$input_mode",
		     "-vfloor_file", "$vfloor_file", "-var_floor", "$vfloor",
		     "-num_features", $num_features);
	    $out = join(" ", @args);
	    print "COMMAND: rsh $cpu $out\n";
	}
	
	elsif ($mode eq "-acc") {
	    # build the argument list for the call to init_hmm
	    #
	    @args = ($INIT_HMM, "-input", "$mfcc_list", "-models",
		     "$models_file", "-trans", "$output_transitions_file",
		     "-state", "$output_states_file", "-mode", "$input_mode",
		     "-vfloor_file", "$vfloor_file", "-var_floor", "$vfloor",
		     "-num_features", $num_features, $mode, "-delta", 
		     "-delta_win", $delta_win);
	    $out = join(" ", @args);
	    print "COMMAND: rsh $cpu $out\n";
	}
	       
	elsif ($mode eq "-delta") {
	    # build the argument list for the call to init_hmm
	    #
	    @args = ($INIT_HMM, "-input", "$mfcc_list", "-models",
		     "$models_file", "-trans", "$output_transitions_file",
		     "-state", "$output_states_file", "-mode", "$input_mode",
		     "-vfloor_file", "$vfloor_file", "-var_floor", "$vfloor",
		     "-num_features", $num_features, $mode, 
		     "-delta_win", $delta_win);
	    $out = join(" ", @args);
	    print "COMMAND: rsh $cpu $out\n";
	}
	
    }
    
    # run init_hmm
    #
    if (((`rsh $cpu $out`) != 0) || 
	(file_exists("$output_transitions_file") != 1) ||
	(file_exists("$output_states_file") != 1) ||
	(file_exists("$vfloor_file") != 1)) {
	
	print "ERROR: run_init_hmm - call to init_hmm failed\n";
	return 0;
	
    }
    
    # exit gracefully
    #
    return 1;
}

# subroutine: run_convert_mmf
#
# arguments:
#  @_ : (input_mode, output_mode, input_file, output_file, cpu)
#
# description: this routine runs convert_mmf and checks for error
#              conditions
#
# return: 0 if error 1 if no error
#
sub run_convert_mmf ($$$$$) {

    # parse the argument list
    #
    my $input_mode = shift @_;
    my $output_mode = shift @_;
    my $input_file = shift @_;
    my $output_file = shift @_;
    my $cpu = shift @_;

    if (($input_mode eq "") || ($output_mode eq "") ||
	($input_file eq "") || ($output_file eq "") || ($cpu eq "")) {
	print "ERROR: run_convert_mmf - bad argument list\n";
	return 0;
    }

    # build the argument list for the call to convert_mmf
    #
    @args = ($CONVERT_MMF, "-input_mode", "$input_mode", "-output_mode",
	     "$output_mode", "$input_file", "$output_file");
    $out = join(" ", @args);
    print "COMMAND: rsh $cpu $out\n";
    
    # run convert_mmf
    #
    if (((`rsh $cpu $out`) != 0) || 
	(file_exists("$output_file") != 1)) {
	return 0;
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: run_create_mixtures
#
# arguments:
#  @_ : (input_states, output_states, num_mixtures, cpu)
#
# description: this routine runs create_mixtures and checks for error
#              conditions
#
# return: 0 if error 1 if no error
#
sub run_create_mixtures ($$$$) {

    # parse the argument list
    #
    my $input_states = shift @_;
    my $output_states = shift @_;
    my $num_mixtures = shift @_;
    my $cpu = shift @_;

    if (($input_states eq "") || ($output_states eq "") ||
	($num_mixtures eq "") || ($cpu eq "")) {

	print "ERROR: run_create_mixtures - bad argument list\n";
	return 0;
    }

    # build the argument list for the call to create_mixtures
    #
    @args = ($CREATE_MIXTURES, "-input", "$input_states", "-output",
	     "$output_states", "-num_mixtures", "$num_mixtures");
    $out = join(" ", @args);
    print "\n rsh $cpu $out\n";
    
    # run create_mixtures
    #
    if (((`rsh $cpu $out`) != 0) || 
	(file_exists("$output_states") != 1)) {
	return 0;
    }

    # exit gracefully
    #
    return 1;
}


#----------------------------------------------------------------------------
#
# prototype system subroutines
#
#----------------------------------------------------------------------------

# subroutine: build_xwrd_transcription
#
# arguments:
#  @_ : (mono_trans_file, xwrd_trans_file);
#
# description: this routine takes a set of monophone transcriptions and 
#              creates a corresponding set of cross-word triphone
#              transcriptions
#
# return: 0 if error 1 if no error
#
sub build_xwrd_transcription ($$) {

    # get parameters from arg list
    #
    my $mono_trans_file = shift @_;
    my $xwrd_trans_file = shift @_;

    if (($mono_trans_file eq "") || ($xwrd_trans_file eq "")) {
	print "ERROR: build_xwrd_transcription - bad argument list\n";
	return 0;
    }

    # open the input and output files
    #
    (isip_open(fp_mono_in, "$mono_trans_file")) or return 0;
    (isip_open(fp_xwrd_out, ">$xwrd_trans_file")) or return 0;

    # loop over all monophone transcriptions and create their cross-word
    # counterparts
    #
    $NULL = "null";
    while(<fp_mono_in>) {

	# split the line into monophones
	#
	chop;
	@phones = split;
    
	# loop over all of the monophones
	#
	for ($i = 0; $i <= $#phones; $i++) {

	    # if first phone of the transcription is silence then just print
	    # it since silence is not expanded
	    #
	    if (($phones[$i] eq "sil") && ($i == 0)) {
		print fp_xwrd_out "$phones[$i] ";
		$left = $phones[$i];
	    }

	    # if central phone is silence then just print it since 
	    # silence is not expanded
	    #
	    elsif ($phones[$i] eq "sil") {
		$phn = $phones[$i];
		if ($i == $#phones) {
		    print fp_xwrd_out "$phn\n";
		}
		else {
		    print fp_xwrd_out "$phn ";
		    $left = $phn;
		}
	    }

	    # if central phone is short pause then print it but don't update
	    # the left context since short pause can not be a context
	    #
	    elsif ($phones[$i] eq "sp") {

		print fp_xwrd_out "$phones[$i] ";
		
		# do not update the left context
		#
		$left = $left;
	    }

	    # if central phone is neither silence nor short pause
	    #
	    elsif ($i != $#phones) {

		# get the left context
		#
		if ($left ne $NULL) {
		    $phn = join("-", $left, $phones[$i]);
		}
		else {
		    $phn = $phones[$i];
		}

		# need to look at the next phone
		#
		if ($phones[$i+1] ne "sp") {
		    $phn = join("+", $phn, $phones[$i+1]);
		}
            
		# need to look at the first phone of next word for right
		# context
		#
		else {
		    $phn = join("+", $phn, $phones[$i+2]);
		}
		print fp_xwrd_out "$phn ";
		$left = $phones[$i];
	    }
	}
    }

    # close the input and output files
    #
    (isip_close(fp_mono_in) == 1) or return 0;
    (isip_close(fp_xwrd_out) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: run_model_alignment
#
# arguments:
#  @_ : (mfcc_list, output_list, transcription_list, monophones_file,
#        models_file, phones_file, lexicon_file, input_states,
#        input_transition, input_format, context_mode, state_beam, model_beam,
#        word_beam, mapmi_limit, num_allowed_words, output_dir,
#        aligned_trans_file, aligned_mfcc_file, align_param_file, $cpu);
#
# description: this routine creates monophone forced alignments
#
# return: 0 if error 1 if no error
#
sub run_model_alignment ($$$$$$$$$$$$$$$$$$$$$) {

    # get parameters from arg list
    #
    $mfcc_list = shift @_;
    $output_list = shift @_;
    $transcription_list = shift @_;
    $monophones_file = shift @_;
    $models_file = shift @_;
    $phones_file = shift @_;
    $lexicon_file = shift @_;
    $input_states = shift @_;
    $input_transition = shift @_;
    $input_format = shift @_;
    $context_mode = shift @_;
    $state_beam = shift @_;
    $model_beam = shift @_;
    $word_beam = shift @_;
    $mapmi_limit = shift @_;
    $num_allowed_words = shift @_;
    $output_dir = shift @_;
    $aligned_trans_file = shift @_;
    $aligned_mfcc_file = shift @_;
    $align_param_file = shift @_;
    $cpu = shift @_;

    if (($align_param_file eq "") || ($output_dir eq "")  ||
	($aligned_trans_file eq "") || ($aligned_mfcc_file eq "")
	|| ($cpu eq "")) {
	print "ERROR: run_model_alignment - bad argument list\n";
	return 0;
    }

    # try to open the output parameter file
    #
    (isip_open(fp_out, ">$align_param_file") == 1) or return 0;

    # loop over all lines and do regular expression replacements to create the
    # new file
    #
    $line = $ALIGN_PARAM_FILE_TEMPLATE;
    $line =~ s/__MFCC_LIST__/$mfcc_list/g;
    $line =~ s/__OUTPUT_LIST__/$output_list/g;
    $line =~ s/__TRANSCRIPTION_FILE__/$transcription_list/g;
    $line =~ s/__MONOPHONES_FILE__/$monophones_file/g;
    $line =~ s/__MODELS_FILE__/$models_file/g;
    $line =~ s/__PHONES_FILE__/$phones_file/g;
    $line =~ s/__LEXICON_FILE__/$lexicon_file/g;
    $line =~ s/__INPUT_STATES__/$input_states/g;
    $line =~ s/__INPUT_TRANSITIONS__/$input_transition/g;
    $line =~ s/__INPUT_FORMAT__/$input_format/g;
    $line =~ s/__CONTEXT_MODE__/$context_mode/g;
    $line =~ s/__STATE_BEAM__/$state_beam/g;
    $line =~ s/__MODEL_BEAM__/$model_beam/g;
    $line =~ s/__WORD_BEAM__/$word_beam/g;
    $line =~ s/__MAPMI_LIMIT__/$mapmi_limit/g;
    $line =~ s/__NUM_ALLOWED_WORDS__/$num_allowed_words/g;
    print fp_out "$line";

    # close the parameter file
    #
    (isip_close(fp_out) == 1) or return 0;

    # create the output directories if necessary
    #
    (create_directories($output_list) == 1) or return 0;

    # run the trace_projector utility to create the alignments
    #
    @args = ($TRACE_PROJECTOR, "-p", $align_param_file);
    $out = join(" ", @args);
    print "COMMAND: rsh $cpu $out\n";
    
    # run alignment
    #
    (`rsh $cou $out` == 0) or return 0;

    # filter out utterances that could not be aligned
    #
    (filter_nonaligned($output_list, $mfcc_list, $aligned_trans_file,
		       $aligned_mfcc_file) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: filter_nonaligned
#
# arguments:
#  @_ : (aligned_list_file, mfccc_file, aligned_trans_file, aligned_mfcc_file);
#
# description: this routine takes forced alignments and removes any 
#              utterances that could not be aligned. The output of this 
#              script are correspoding feature file list and transcription
#              lists of all files that could be aligned
#
# return: 0 if error 1 if no error
#
sub filter_nonaligned ($$$$) {

    # get parameters from arg list
    #
    my $aligned_list_file = shift @_;
    my $mfcc_file = shift @_;
    my $aligned_trans_file = shift @_;
    my $aligned_mfcc_file = shift @_;

    if (($aligned_list_file eq "") || ($mfcc_file eq "") ||
	($aligned_trans_file eq "")  || ($aligned_mfcc_file eq "")) {
	print "ERROR: filter_nonaligned - bad argument list\n";
	return 0;
    }

    # open the input and output files
    #
    (isip_open(fp_align_in, "$aligned_list_file") == 1) or return 0;
    (isip_open(fp_mfcc_in, "$mfcc_file") == 1) or return 0;
    (isip_open(fp_trans_out, ">$aligned_trans_file") == 1) or return 0;
    (isip_open(fp_mfcc_out, ">$aligned_mfcc_file") == 1) or return 0;

    # loop over each utterance and see if it aligned properly. if so, then
    # output the transcription and feature file name
    #
    while (<fp_align_in>) {
	
	# get the corresponding mfcc file
	#
	$curr_mfcc_file = <fp_mfcc_in>;

	# find the next file to get the transcription for and open it
	#
	chop;
	$file = $_;
	(isip_open(fp_alignment_file, "$file") == 1) or return 0;
	
	# generate the phone-level transcription from the output file
	#
	$flag = 1;
	while(<fp_alignment_file>) {

	    $line = $_;
	    chop($line);
	    @data = split(/\s+/, $line);
	    $len = @data;

	    # see if this is a word hypothesis
	    #
	    if (($len == 7) && ($data[4] eq "!SENT_START")) {

		# grab the model name
		#
		$sent = $data[2];
		
		# now we are in the sentence so start building the trans.
		#
		while(<fp_alignment_file>) {
		    
		    $line1 = $_;
		    chop($line1);
		    @data = split(/\s+/, $line1);
		    $len1 = @data;
		    $sent = join(" ", $sent, $data[2]);

		    # word end "sp"
		    #
		    if (($len1 == 7) && ($data[2] ne "sil") &&
			($data[2] ne "sp")) {
			$sent = join(" ", $sent, "sp");
		    }

		    # done with sentence
		    #
		    if ($data[4] eq "!SENT_END") {
			$flag = 2;
		    }
		}

		# if we found a valid sentence then print it out
		#
		if ($flag == 2) {
		    print fp_trans_out "$sent\n";
		    print fp_mfcc_out "$curr_mfcc_file";
		}
	    }
	}
    }

    (isip_close(fp_align_in) == 1) or return 0;
    (isip_close(fp_mfcc_in) == 1) or return 0;
    (isip_close(fp_trans_out) == 1) or return 0;
    (isip_close(fp_mfcc_out) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: build_wint_transcription
#
# arguments:
#  @_ : (mono_trans_file, wint_trans_file);
#
# description: this routine takes a set of monophone transcriptions and 
#              creates a corresponding set of word-internal triphone
#              transcriptions
#
# return: 0 if error 1 if no error
#
sub build_wint_transcription ($$) {

    # get parameters from arg list
    #
    my $mono_trans_file = shift @_;
    my $wint_trans_file = shift @_;

    if (($mono_trans_file eq "") || ($wint_trans_file eq "")) {
	print "ERROR: build_wint_transcription - bad argument list\n";
	return 0;
    }

    # open the input and output files
    #
    (isip_open(fp_mono_in, "$mono_trans_file") == 1) or return 0;
    (isip_open(fp_wint_out, ">$wint_trans_file") == 1) or return 0;

    # loop over all monophone transcriptions and create their word-internal
    # counterparts
    #
    $NULL = "null";
    while(<fp_mono_in>) {

	# split the line into monophones
	#
	chop;
	@phones = split;
    
	# loop over all of the monophones
	#
	for ($i = 0; $i <= $#phones; $i++) {
	    
	    # if first phone of the transcription is silence then just print
	    # it since silence is not expanded
	    #
	    if (($phones[$i] eq "sil") && ($i == 0)) {
		print fp_wint_out "$phones[$i] ";
		$left = $NULL;
	    }

	    # if central phone is silence then just print it since 
	    # silence is not expanded
	    #
	    elsif ($phones[$i] eq "sil") {
		$phn = $phones[$i];
		if ($i == $#phones) {
		    print fp_wint_out "$phn\n";
		}
		else {
		    print fp_wint_out "$phn ";
		    $left = $NULL;
		}
	    }

	    # if central phone is short pause then print it but don't update
	    # the left context since short pause can not be a context
	    #
	    elsif ($phones[$i] eq "sp") {
		print fp_wint_out "$phones[$i] ";
		$left = $NULL;
	    }

	    # if central phone is neither silence nor short pause
	    #
	    elsif ($i != $#phones) {

		# get the left context
		#
		if ($left ne $NULL) {
		    $phn = join("-", $left, $phones[$i]);
		}
		else {
		    $phn = $phones[$i];
		}

		# need to look at the next phone
		#
		if (($phones[$i+1] ne "sp") && ($phones[$i+1] ne "sil")) {
		    $phn = join("+", $phn, $phones[$i+1]);
		}
		print fp_wint_out "$phn ";
		$left = $phones[$i];
	    }
	}
    }

    # close the input and output files
    #
    (isip_close(fp_mono_in) == 1) or return 0;
    (isip_close(fp_wint_out) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: build_phone_list
#
# arguments:
#  @_ : (phone_trans_file, phone_list_file);
#
# description: this routine takes a set of phone transcriptions and 
#              finds all unique phones in the transcriptions. The set of
#              unique phones is written to the phone_list_file
#
# return: 0 if error 1 if no error
#
sub build_phone_list ($$) {

    # get parameters from arg list
    #
    my $phone_trans_file = shift @_;
    my $phone_list_file = shift @_;

    if (($phone_trans_file eq "") || ($phone_list_file eq "")) {
	print "ERROR: build_phone_list - bad argument list\n";
	return 0;
    }

    # open the input and output files
    #
    (isip_open(fp_in, "$phone_trans_file")) or return 0;
    (isip_open(fp_out, ">$phone_list_file")) or return 0;

    # loop over each transcription and gather the phones
    #
    while (<fp_in>) {

	# split the line into a list of phones
	#
	chop $_;
        @phones = split(/\s+/, $_);

        # accumulate the unique phones
        #
        for ($i = 0; $i <= $#phones; $i++) {
            if (!defined($yes{$phones[$i]})) {
                $yes{$phones[$i]} = 1;
            }
        }
    }

    # print the unique phones to the output list
    #
    foreach $triphone (sort keys(%yes)){
	print fp_out "$triphone\n";
    }

    # close the input and output files
    #
    (isip_close(fp_in) == 1) or return 0;
    (isip_close(fp_out) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: add_sp_and_sil_trans
#
# arguments:
#  @_ : (input_trans_file, output_trans_file, input_models_file,
#        output_models_file)
#
# description: this routine ties the sp model to the silence model
#              and adds skip transitions to the silence model. the sil
#              model is assumed to be three state, left-to-right hmm and the
#              sp model is assumed to be a one-state model.
#
# return: 0 if error 1 if no error
#
sub add_sp_and_sil_trans ($$$$) {

    # parse the command line
    #
    my $input_trans_file = shift @_;
    my $output_trans_file = shift @_;
    my $input_models_file = shift @_;
    my $output_models_file = shift @_;

    # verify that all necessary parameters are specified
    #
    if (($input_trans_file eq "") || ($output_trans_file eq "") ||
	($input_models_file eq "") || ($output_models_file eq "")) {
	print "ERROR: add_sp_trans - bad argument list\n";
	return 0;
    }

    # initialize local variables
    #
    my $sp_trans_index = -1;
    my $sil_trans_index = -1;
    my $sp_model_index = -1;
    my $sil_model_index = -1;
    my $sil_central_state_index = -1;
    
    # open the input models file and parse out the necessary information
    #
    if (! open (fp_in_models, "$input_models_file")) {
	print "ERROR: add_sp_trans - unable to open $input_models_file ";
	print "for reading\n";
	return 0;
    }

    # read in all lines of the models file - it is rarely large
    #
    @lines = <fp_in_models>;
    isip_close(fp_in_models);

    # find the model information for the sp and sil models
    #
    $current_ind = -1;
    for ($i = 0; $i <= $#lines; $i++) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;

	# see if the line is the start of a model definition
	#
	if ($_ =~ "index: ") { 

	    # split the line up and find the model index
	    #
	    @toks = split(/\s+/, $_);
	    $current_ind = $toks[1];
	}

	# see if we are at the phone portion of the model definition
	#
	elsif ($_ =~ "phone: ") { 

	    # split the line up and find the model name
	    #
	    @toks = split(/\s+/, $_);
	    $current_mod = $toks[1];

	    # if this is the sil model then we can find the remaining 
	    # information on the sil model
	    #
	    if ($current_mod eq "sil") {

		# get the sil index
		#
		$sil_model_index = $current_ind;

		# skip the "num_states" line
		#
		$_ = $lines[++$i];

		# find the transition index
		# 
		$_ = $lines[++$i];
		@toks = split(/\s+/, $_);
		$sil_trans_index = $toks[1];
		
		# find the central state of the sil model
		#
		$_ = $lines[++$i];
		@toks = split(/\s+/, $_);
		$sil_central_state_index = $toks[3];
	    }

	    # if this is the sp model then set the sp index
	    #
	    elsif ($current_mod eq "sp") {

		# get the sp index
		#
		$sp_model_index = $current_ind;
	    }
	}	
    }
	
    # now parse the transitions file, find the transitions for the sil
    # model and the sp model and replace them with the desired form
    #
    (isip_open(fp_in_trans, "$input_trans_file") == 1) or return 0;
    (isip_open(fp_out_trans, ">$output_trans_file") == 1) or return 0;

    # read in all of the lines and close
    #
    @lines = <fp_in_trans>;
    isip_close(fp_in_trans);

    # loop forward until we find the tag line with the number of 
    # transitions
    #
    $i = 0;
    while ($i <= $#lines) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;
	
	# look for the tag string
	#
	if ($_ =~ "num_transitions") {

	    # split up the line and grab the number of transitions
	    #
	    @toks = split(/\s+/, $_);
	    $num_trans = $toks[2];

	    # determine what the sp transition index should be and then 
	    # increment the number of transitions
	    #
	    $sp_trans_index = $num_trans;
	    $num_trans++;

	    # write out the new number of transitions
	    #
	    print fp_out_trans "num_transitions = $num_trans\n\n";
	    last;
	}
	$i++;
    }
    
    # next line
    #
    $i++;

    # loop over the transition file and change the sil and sp transitions
    #
    $current_ind = -1;
    for (; $i <= $#lines; $i++) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;
	
	# see if the line is the start of a transition definition
	#
	if ($_ ne "") {

	    # find the token index
	    @toks = split(/\s+/, $_);
	    $current_ind = $toks[0];
	    $current_ind =~ s/\.//;

	    # get the length of the transition matrix
	    #
	    $size = $toks[1];

	    # print the transition index line
	    #
	    print fp_out_trans "$_\n";
	    $i++;
	    
	    # is this is not the sil model transition then just print it
	    #
	    if ($current_ind != $sil_trans_index) {
		for ($k = 0; $k < $size; $k++) {
		    print fp_out_trans "$lines[$i+$k]";
		}
		$i += $size - 1;
	    }

	    # if this is the sil model then change the transition matrix
	    # except for the transitions out of state 2.
	    #
	    else {

		print fp_out_trans " 0.000000e+00 1.000000e+00 0.000000e+00 0.000000e+00 0.000000e+00\n";
		$i++;

		print fp_out_trans " 0.000000e+00 5.000000e-01 3.000000e-01 2.000000e-01 0.000000e+00\n";
		$i++;

		print fp_out_trans "$lines[$i]";

		print fp_out_trans " 0.000000e+00 2.000000e-01 0.000000e+00 5.000000e-01 3.000000e-01\n";
		$i++;

		print fp_out_trans " 0.000000e+00 0.000000e+00 0.000000e+00 0.000000e+00 0.000000e+00\n";
		$i++;
	    }
	}

	# else print a blank line
	#
	else {
	    print fp_out_trans "\n";
	}
    }
    
    # print the sp transition matrix making the transitions equi-probable
    #
    print fp_out_trans "$sp_trans_index. 3\n";
    print fp_out_trans "0.000000e+00 5.000000e-01 5.000000e-01\n";
    print fp_out_trans "0.000000e+00 5.000000e-01 5.000000e-01\n";
    print fp_out_trans "0.000000e+00 0.000000e+00 0.000000e+00\n";

    # close the transitions file
    #
    isip_close(fp_out_trans);

    # now go back and rebuild the model file
    #
    # open the input models file and parse out the necessary information
    #
    (isip_open(fp_in_models, "$input_models_file") == 1) or return 0;
    (isip_open(fp_out_models, ">$output_models_file") == 1) or return 0;

    # read in all of the lines and close the file
    #
    @lines = <fp_in_models>;
    isip_close(fp_in_models);

    # find the sp model information
    #
    $current_ind = -1;
    for ($i = 0; $i <= $#lines; $i++) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;

	# see if the line is the start of a model definition
	#
	if ($_ =~ "index: ") { 
	    
	    # split the line up and find the model index
	    #
	    @toks = split(/\s+/, $_);
	    $current_ind = $toks[1];
	}

	# see if this is the sp model index
	#
	if ($current_ind == $sp_model_index) {

	    # write out the sp model information
	    #
	    print fp_out_models "index: $sp_model_index\n";
	    print fp_out_models "phone: sp\n";
	    print fp_out_models "num_states: 3\n";
	    print fp_out_models "transitions: $sp_trans_index\n";
	    print fp_out_models "states: 0 $sil_central_state_index 0\n";
	    print fp_out_models "\n";
	    $i += 5;
	    $current_index = -1;
	}
	
	# all other lines are just passed through
	#
	else {
	    print fp_out_models "$_\n";
	}
    }
	
    # close the output files
    #
    isip_close(fp_out_models);

    # exit gracefully
    #
    return 1;
}

# subroutine: tie_sp_to_sil
#
# arguments:
#  @_ : (input_trans_file, output_trans_file, input_models_file,
#        output_models_file)
#
# description: this routine ties the sp model to the central state of the
#              silence model. the sp model is assumed to be a one-state model.
#
# return: 0 if error 1 if no error
#
sub tie_sp_to_sil ($$$$) {

    # parse the command line
    #
    my $input_trans_file = shift @_;
    my $output_trans_file = shift @_;
    my $input_models_file = shift @_;
    my $output_models_file = shift @_;

    # verify that all necessary parameters are specified
    #
    if (($input_trans_file eq "") || ($output_trans_file eq "") ||
	($input_models_file eq "") || ($output_models_file eq "")) {
	print "ERROR: tie_sp_to_sil - bad argument list\n";
	return 0;
    }

    # initialize local variables
    #
    my $sp_trans_index = -1;
    my $sil_trans_index = -1;
    my $sp_model_index = -1;
    my $sil_model_index = -1;
    my $sil_central_state_index = -1;
    
    # open the input models file and parse out the necessary information
    #
    (isip_open(fp_in_models, "$input_models_file") == 1) or return 0;

    # read in all lines of the models file - it is rarely large
    #
    @lines = <fp_in_models>;
    isip_close(fp_in_models);

    # find the model information for the sp and sil models
    #
    $current_ind = -1;
    for ($i = 0; $i <= $#lines; $i++) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;

	# see if the line is the start of a model definition
	#
	if ($_ =~ "index: ") { 

	    # split the line up and find the model index
	    #
	    @toks = split(/\s+/, $_);
	    $current_ind = $toks[1];
	}

	# see if we are at the phone portion of the model definition
	#
	elsif ($_ =~ "phone: ") { 

	    # split the line up and find the model name
	    #
	    @toks = split(/\s+/, $_);
	    $current_mod = $toks[1];

	    # if this is the sil model then we can find the remaining 
	    # information on the sil model
	    #
	    if ($current_mod eq "sil") {

		# get the sil index
		#
		$sil_model_index = $current_ind;

		# skip the "num_states" line
		#
		$_ = $lines[++$i];

		# find the transition index
		# 
		$_ = $lines[++$i];
		@toks = split(/\s+/, $_);
		$sil_trans_index = $toks[1];
		
		# find the central state of the sil model
		#
		$_ = $lines[++$i];
		@toks = split(/\s+/, $_);
		$num_states = $#toks - 1;
		$central_ind = int ($num_states / 2) + int ($num_states % 2);
		$sil_central_state_index = $toks[$central_ind];
	    }

	    # if this is the sp model then set the sp index
	    #
	    elsif ($current_mod eq "sp") {

		# get the sp index
		#
		$sp_model_index = $current_ind;
	    }
	}	
    }
	
    # now parse the transitions file, find the transitions for the sil
    # model and the sp model and replace them with the desired form
    #
    (isip_open(fp_in_trans, "$input_trans_file") == 1) or return 0;
    (isip_open(fp_out_trans, ">$output_trans_file") == 1) or return 0;

    # read in all of the lines and close
    #
    @lines = <fp_in_trans>;
    isip_close(fp_in_trans);

    # loop forward until we find the tag line with the number of 
    # transitions
    #
    $i = 0;
    while ($i <= $#lines) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;
	
	# look for the tag string
	#
	if ($_ =~ "num_transitions") {

	    # split up the line and grab the number of transitions
	    #
	    @toks = split(/\s+/, $_);
	    $num_trans = $toks[2];

	    # determine what the sp transition index should be and then 
	    # increment the number of transitions
	    #
	    $sp_trans_index = $num_trans;
	    $num_trans++;

	    # write out the new number of transitions
	    #
	    print fp_out_trans "num_transitions = $num_trans\n\n";
	    last;
	}
	$i++;
    }
    
    # next line
    #
    $i++;

    # loop over the transition file and change the sp transitions
    #
    $current_ind = -1;
    for (; $i <= $#lines; $i++) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;
	
	# see if the line is the start of a transition definition
	#
	if ($_ ne "") {

	    # find the token index
	    @toks = split(/\s+/, $_);
	    $current_ind = $toks[0];
	    $current_ind =~ s/\.//;

	    # get the length of the transition matrix
	    #
	    $size = $toks[1];

	    # print the transition index line
	    #
	    print fp_out_trans "$_\n";
	    $i++;
	    
	    # is this is not the sil model transition then just print it
	    #
	    for ($k = 0; $k < $size; $k++) {
		print fp_out_trans "$lines[$i+$k]";
	    }
	    $i += $size - 1;
	}
	# else print a blank line
	#
	else {
	    print fp_out_trans "\n";
	}
    }
    
    # print the sp transition matrix making the transitions equi-probable
    #
    print fp_out_trans "$sp_trans_index. 3\n";
    print fp_out_trans "0.000000e+00 5.000000e-01 5.000000e-01\n";
    print fp_out_trans "0.000000e+00 5.000000e-01 5.000000e-01\n";
    print fp_out_trans "0.000000e+00 0.000000e+00 0.000000e+00\n";

    # close the transitions file
    #
    isip_close(fp_out_trans);

    # now go back and rebuild the model file
    #
    # open the input models file and parse out the necessary information
    #
    (isip_open(fp_in_models, "$input_models_file") == 1) or return 0;
    (isip_open(fp_out_models, ">$output_models_file") == 1) or return 0;

    # read in all of the lines and close the file
    #
    @lines = <fp_in_models>;
    isip_close(fp_in_models);

    # find the sp model information
    #
    $current_ind = -1;
    for ($i = 0; $i <= $#lines; $i++) {

	# remove whitespace
	#
	$_ = $lines[$i];
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+/ /g;

	# see if the line is the start of a model definition
	#
	if ($_ =~ "index: ") { 
	    
	    # split the line up and find the model index
	    #
	    @toks = split(/\s+/, $_);
	    $current_ind = $toks[1];
	}

	# see if this is the sp model index
	#
	if ($current_ind == $sp_model_index) {

	    # write out the sp model information
	    #
	    print fp_out_models "index: $sp_model_index\n";
	    print fp_out_models "phone: sp\n";
	    print fp_out_models "num_states: 3\n";
	    print fp_out_models "transitions: $sp_trans_index\n";
	    print fp_out_models "states: 0 $sil_central_state_index 0\n";
	    print fp_out_models "\n";
	    $i += 5;
	    $current_index = -1;
	}
	
	# all other lines are just passed through
	#
	else {
	    print fp_out_models "$_\n";
	}
    }
	
    # close the output files
    #
    isip_close(fp_out_models);

    # exit gracefully
    #
    return 1;
}

#----------------------------------------------------------------------------
#
# generally useful subroutines like file opens, file copies, directory 
# creation, file existence checks, etc.
#
#----------------------------------------------------------------------------

# subroutine: print_file
#
# arguments:
#  $_ : (input_file)
#
# description: this routine prints a file to stdout
#
# return: 0 if error 1 if no error
#
sub print_file ($) {
    
    # parse the argument list
    #
    $input_file = shift @_;
    
    if ($input_file eq "") {
	print "ERROR: print_file - bad argument list\n";
	return 0;
    }

    # open the input file
    #
    (isip_open(fp_in, "$input_file") == 1) or return 0;

    # print each line
    #
    while (<fp_in>) {
	print "$_";
    }

    # close the file
    #
    (isip_close(fp_in) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}


# subroutine: check_files_in_list
#
# arguments:
#  $_ : (file_list)
#
# description: this routine ensures that the files listed in the input file
#              exist
#
# return: 0 if error 1 if no error
#
sub check_files_in_list ($) {
    
    # parse the argument list
    #
    $file_list = shift @_;
    
    if ($file_list eq "") {
	print "ERROR: check_files_in_list - bad argument list\n";
	return 0;
    }

    # open the input file
    #
    (isip_open(fp_in, "$file_list") == 1) or return 0;

    # loop over each line and verify that the file exists
    #
    while (<fp_in>) {

	# strip off whitespace
	#
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+//g;
	
	if ($_ ne "") {

	    if (file_exists($_) != 1) {
		
		# we didn't find a file
		#
		print "INFO: check_files_in_list - file not found: $_\n";
		return 0;
	    }
	}
    }

    # exit gracefully
    #
    return 1;
}


# subroutine: file_exists
#
# arguments:
#  @_ : (filename)
#
# description: this routine ensures that a file exists
#
# return: 0 if error 1 if no error
#
sub file_exists($) {

    # get the filename
    #
    my $file = shift @_;

    # make sure the name is non-null
    #
    if ($file eq "") {
	print "ERROR: file_exists - file not given\n";
	return 0;
    }

    # make sure the file exists
    #
    if (-e $file) {
	return 1;
    }
    else {
	return 0;
    }
}


# subroutine: isip_open
#
# arguments:
#  @_ : (file_pointer, file_name)
#
# description: this routine attempts to open a file the same way that 
#              perl's open does
#
# return: 0 if error 1 if no error
#
sub isip_open (*;$) {

    my $fp = shift @_;
    my $fname = shift @_;

    if (!open($fp, $fname)) {
	
	# find out what mode we tried to open in
	#
	$mode = "read";
	if (index($fname, ">") == 0) {
	    $mode = "write";
	}

	# print an error message
	#
	print "ERROR: isip_open - error opening file $fname in $mode mode\n";
	return 0;
    }
    
    return 1;
}

# subroutine: isip_close
#
# arguments:
#  @_ : (file_pointer)
#
# description: this routine attempts to close a file the same way that 
#              perl's close does
#
# return: 0 if error 1 if no error
#
sub isip_close (*) {

    my $fp = shift @_;

    if (!close($fp)) {
	
	# print an error message
	#
	print "ERROR: isip_close - error closing file\n";
	return 0;
    }
    
    return 1;
}

# subroutine: isip_copy
#
# arguments:
#  @_ : (src_file, destination_file)
#
# description: this routine attempts to copy the src_file to the dst_file
#
# return: 0 if error 1 if no error
#
sub isip_copy ($$) {

    # parse the argument list
    #
    my $src_file = shift @_;
    my $dst_file = shift @_;

    if (($src_file eq "") || ($dst_file eq "")) {
	print "ERROR: isip_copy - bad argument list\n";
	return 0;
    }

    if (copy($src_file, $dst_file) == 0) {
	print "ERROR: isip_copy - error copying  $src_file ";
	print "to $dst_file\n";
	return 0;
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: isip_move
#
# arguments:
#  @_ : (src_file, destination_file)
#
# description: this routine attempts to move the src_file to the dst_file
#
# return: 0 if error 1 if no error
#
sub isip_move ($$) {

    # parse the argument list
    #
    my $src_file = shift @_;
    my $dst_file = shift @_;

    if (($src_file eq "") || ($dst_file eq "")) {
	print "ERROR: isip_move - bad argument list\n";
	return 0;
    }

    if (move($src_file, $dst_file) == 0) {
	print "ERROR: isip_move - error copying  $src_file ";
	print "to $dst_file\n";
	return 0;
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: isip_mkdir
#
# arguments:
#  @_ : (needed_dir)
#
# description: this routine attempts to create the given directory
#
# return: 0 if error 1 if no error
#
sub isip_mkdir ($) {

    # parse the argument list
    #
    my $needed_dir = shift @_;

    if ($needed_dir eq "") {
	print "ERROR: isip_mkdir - bad argument list\n";
	return 0;
    }

    # make the directory and any other parent directories needed to make that
    # directory
    if (! -d $needed_dir) {
	@successes = mkpath($needed_dir);
	if ($#successes < 0) {
	    print "ERROR: isip_mkdir - could not create directory:\n";
	    print "ERROR: $needed_dir\n";
	    return 0;
	}
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: file_replace_string
#
# arguments:
#  @_ : (search_string, replace_string, input_file, output_file)
#
# description: finds all instances of a search_string in the input file and
#              replaces them with the replace_string, writing to the
#              output_file
#
# return: 0 if error 1 if no error
#
sub file_replace_string ($$$$) {

    # parse the argument list
    #
    $search_string = shift @_;
    $replace_string = shift @_;
    $input_file = shift @_;
    $output_file = shift @_;

    if (($input_file eq "") || ($output_file eq "")) {
	print "ERROR: file_replace_string - bad argument list\n";
	return 0
    }

    # open the input and output files
    #
    (isip_open(fp_in, "$input_file") == 1) or return 0;
    (isip_open(fp_out, ">$output_file") == 1) or return 0;

    # loop over the file and replace
    #
    while (<fp_in>) {

	# filter the line
	#
	$_ =~ s/$search_string/$replace_string/g;
	print fp_out "$_";
    }

    # close the input and output files
    #
    (isip_close(fp_in) == 1) or return 0;
    (isip_close(fp_out) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}

# subroutine: map_wsj
#
# arguments:
#  @_ : (input_file_list, output_file_list)
#
# description: maps the files in the input list to the WSJ transcriptions 
#              on to the files in the output list
#
# return: 0 if error 1 if no error
#
sub map_wsj ($$) {
    
    # parse the argument list
    #
    $input_file_list = shift @_;
    $output_file_list = shift @_;

    if (($input_file_list eq "") || ($output_file_list eq "")) {
	print "ERROR: map_wsj - bad argument list\n";
	return 0;
    }

    # open the input and output files
    #
    (isip_open(fp_in, "$input_file_list") == 1) or return 0;
    (isip_open(fp_out, "$output_file_list") == 1) or return 0;

    # loop over the files in the list and replace
    #
    @in_files = <fp_in>;
    @out_files = <fp_out>;

    for ($i = 0; $i <= $#in_files; $i++) {	

	$input_file = $in_files[$i];       
	$output_file = $out_files[$i];
	chop $input_file;
	chop $output_file;
	
	# open the input and output files
	#
	(isip_open(fp_in_file, "$input_file") == 1) or return 0;
	(isip_open(fp_out_file, ">$output_file") == 1) or return 0;

	# loop over the file and replace
	#
	while (<fp_in_file>) {
	    
	    # filter the line
	    #
	    $_ =~ s/\s.PERIOD\s/ PERIOD /g;
	    $_ =~ s/\s.POINT\s/ POINT /g;
	    $_ =~ s/\s%PERCENT\s/ PERCENT /g;
	    $_ =~ s/\s\/SLASH\s/ SLASH /g;
	    print fp_out_file "$_";
	}
	
	# close the input and output files
	#
	(isip_close(fp_in_file) == 1) or return 0;
	(isip_close(fp_out_file) == 1) or return 0;
    }
    
    # close the input and output files
    #
    (isip_close(fp_in) == 1) or return 0;
    (isip_close(fp_out) == 1) or return 0;
    
    # exit gracefully
    #
    return 1;
}

# subroutine: create_directories
#
# arguments:
#  @_ : (list_file)
#
# description: for each line in the file, try to create the directory needed
#              to generate the file in the list
#
# return: 0 if error 1 if no error
#
sub create_directories ($) {

    # parse the argument list
    #
    $file_list = shift @_;

    if ($file_list eq "") {
	print "ERROR: create_base_directories - bad argument list\n";
	return 0
    }

    # open the input and output files
    #
    (isip_open(fp_in, "$file_list") == 1) or return 0;

    # loop over each line, treating it as a file, and try to make the directory
    # for the file
    #
    while (<fp_in>) {

	# remove whitespace
	#
	chop $_;
	$_ =~ s/^\s+//g;
	$_ =~ s/\s+//g;

	# if the line is empty then do nothing else create the base directory
	# if necessary
	#
	if ($_ ne "") {

	    # get the subdirectory name
	    #
	    $subdir = dirname($_);
	    
	    # if the directory does not exist, then create it
	    #
	    if (! -d $subdir) {
		(isip_mkdir($subdir) == 1) or return 0;
	    }
	}
    }

    # close the input files
    #
    (isip_close(fp_in) == 1) or return 0;

    # exit gracefully
    #
    return 1;
}

# subroutine: select_transcriptions
#
# arguments:
#  @_ : (mfcc_list_file, base_trans_file, trans_file);
#
# description: this routine takes the base transcription file and mfcc file 
#              list and created the transcription file corresponding to the 
#              files in the mfcc file list 
#
# return: 0 if error 1 if no error
#
sub select_transcriptions ($$$) {

    # get parameters from arg list
    #
    my $mfcc_list_file = shift @_;
    my $base_trans_file = shift @_;
    my $trans_file = shift @_;

    if (($mfcc_list_file eq "") || ($base_trans_file eq "") ||
	($trans_file eq "")) {
	print "ERROR: select_transcriptions - bad argument list\n";
	return 0;
    }

    # open the input and output files
    #
    (isip_open(fp_mfcc_in, "$mfcc_list_file") == 1) or return 0;
    (isip_open(fp_base_trans_in, "$base_trans_file") == 1) or return 0;
    (isip_open(fp_trans_out, ">$trans_file") == 1) or return 0;
    
    # hash the base transcription file
    #
    while(<fp_base_trans_in>) {
	chop;
	@cols = split(/\s+/, $_, 2);
	$trans{$cols[0]} = $cols[1];
    }
    
    # read the mfcc list file
    #
    @mfcc_list = <fp_mfcc_in>;
    
    # loop over each utterance in the mfcc list file and select the 
    # corresponding transcription. write this transcription to the trans file
    # 
    
    for ($i = 0; $i <= $#mfcc_list; $i++) {
	
	$line = $mfcc_list[$i];
	chop($line);
	
	$line =~ s/\s+//g;
	if ($line eq "") {
	    next;
	}

	@data = split(/\//, $line);
	$len_mfc = @data;
	
	$name_mfcc = $data[$len_mfc-1];
	@name_data = split(/\./, $name_mfcc);
	$len = @name_data;
	
	$name = $name_data[$len-2];	
	
	# get the transcription from the hash table
	#
	
	$out = $trans{"$name"};
	
	
	if ($out eq "") {
	    print("ERROR: File $name not in the data base\n");
	    return 0;
	}
	
	# if we found a valid transcription then print it out
	#
	print fp_trans_out "$out\n";
    }
    (isip_close(fp_mfcc_in) == 1) or return 0;
    (isip_close(fp_base_trans_in) == 1) or return 0;
    (isip_close(fp_trans_out) == 1) or return 0;
    
    # exit gracefully
    #
    return 1;
}

# subroutine: create_N_lists
#
# arguments:
#  @_ : (cpus, main_list, PATH_TO_EXP, sub_dir, out_filename, type);
#
# description: this routine takes the # of cpus, master list that 
# needs to be divided in N ways, path to the experiment-directory, 
# the output file name and the type of output file
#
# return: 0 if error 1 if no error
#
sub create_N_lists ($$$$$) {

    # get parameters from arg list
    #
    my $cpus = shift @_;
    my $main_list = shift @_;
    my $PATH_TO_EXP = shift @_;
    my $out_filename = shift @_;
    my $type = shift @_;

    if (($cpus eq "") || ($main_list eq "") || ($type eq "") ||
	($PATH_TO_EXP eq "") || ($out_filename eq "")) {
	print "ERROR: create_N_lists - bad argument list\n";
	return 0;
    }
    
    # open the input file
    #
    (isip_open(fp_main_in, $main_list) == 1) or return 0;

    # create the output lists
    #
    for ($i = 0; $i < $cpus; $i++) {
	$output_file[$i] = 
	    "$PATH_TO_EXP/$out_filename\_$i\.$type";
    }
    
    # loop through the master file list to get the # of lines
    # 
    $line_num = 0;    
    foreach $line (<fp_main_in>) {
	
	# count number of line and hash the lines to be divided
	#
	$lines{$line_num} = $line;
	$line_num++;
    }
    
    # close the input file
    #
    (isip_close(fp_main_in) == 1) or return 0;

    # divide the input file N ways by looping through the main list
    # 
    for ($j = 0; $j < $cpus; $j++) {
	$cpu_file_num[$j] = 0;
    }
    $file_count = 0;
    $j = 0;
    while ($file_count < $line_num) {
	$cpu_file_num[$j]++;
	$j++;
	$file_count++;
	if ($j >= $cpus) {
	    $j = 0;
	}
    }

    $line_index = 0;
    for ($i = 0; $i < $cpus ; $i++){
	
	# open the output file
        #
	open(fp_output, ">$output_file[$i]");
	
	for ($j = 0 ; $j < $cpu_file_num[$i]; $j++) {
	    if ($line_index > $line_num) { 
		last;
	    }
	    print fp_output $lines{$line_index};
	    $line_index++;
	}
	close (fp_output);
    }

    # exit gracefully
    #
    return 1;
}

# subroutine: build_output_list
#
# arguments:
#  @_ : (path_to_the_output, mfcc_list_file, out_list_file);
#
# description: this routine takes the path to the output, gets the file id 
# from the mfcc file list and creates the output file
#
# return: 0 if error 1 if no error
#
sub build_output_list ($$$) {
    
    # get parameters from arg list
    #
    my $out_path = shift @_;
    my $mfcc_list_file = shift @_;
    my $out_list_file = shift @_;
    
    if (($out_path eq "") || ($mfcc_list_file eq "") 
	|| ($out_list_file eq "")) {
	print "ERROR: build_output_list - bad argument list\n";
	return 0;
    }
    
    # open the input and output files
    #
    (isip_open(fp_mfcc_in, "$mfcc_list_file") == 1) or return 0;
    (isip_open(fp_out_out, ">$out_list_file") == 1) or return 0;
    
    # read the mfcc list file
    #
    @mfcc_list = <fp_mfcc_in>;


    # loop over each utterance in the mfcc list file and select the 
    # corresponding file id. write this file to out_list_file
    # 
    for ($i = 0; $i <= $#mfcc_list; $i++) {
    
	$line = $mfcc_list[$i];
	chop($line);
	
	@data = split(/\//, $line);
	$len_mfc = @data;
	
	$name_mfcc = $data[$len_mfc-1];
	@name_data = split(/\./, $name_mfcc);
	$len = @name_data;
	
	$name = $name_data[$len-2];	
	
	# join the path
	#
	$out = join ("/", $out_path, $name);
	$out1 = join (".", $out, "out");
	
	# if we found a valid id then print it out
	#
	print fp_out_out "$out1\n";
    }
    
    (isip_close(fp_mfcc_in) == 1) or return 0;
    (isip_close(fp_out_out) == 1) or return 0;
    
    # exit gracefully
    #
    return 1;
}

# subroutine: strip_silence
#
# arguments:
#  @_ : (utterance_timing_list, mfcc_list_file, output_path, extension, 
#         feature_format, dimension, frame_duration);
#
# description: this routine takes the utterance timings, input mfc file list,
# extension of the output stripped mfc files and strips the silences. The 
# output stripped mfc files are placed at the output_path. If the output path 
# is not given the output stripped mfc files are placed in the same directory
# as the input mfc file.
#
# return: 0 if error 1 if no error
#
sub strip_silence ($$$$$$$) {
    
    # get parameters from arg list
    #
    my $utterance_timing_list = shift @_;
    my $mfcc_list_file = shift @_;
    my $output_path = shift @_;
    my $extension = shift @_;    
    my $feature_format = shift @_;    
    my $dimension = shift @_;    
    my $frame_duration = shift @_;    

    if (($utterance_timing_list eq "") || ($mfcc_list_file eq "") 
	|| ($extension eq "") || ($feature_format eq "") 
	|| ($dimension eq "") || ($frame_duration eq "")) {
	print "ERROR: strip_silence - bad argument list\n";
	return 0;
    }

    # reset the path flag of the output_path is not specified
    $path_flag = 1;
    if ($output_path eq "") {
	$path_flag = 0;
    }
        
    # open the utterance timing file and  input files 
    #
    (isip_open(fp_timing_in, "$utterance_timing_list") == 1) or return 0;
    (isip_open(fp_mfcc_in, "$mfcc_list_file") == 1) or return 0;
    
    # read the mfcc list file
    #
    @mfcc_list = <fp_mfcc_in>;

    # hash the timing file for silence frames at begining and total frames 
    # for utterance 
    #
    while(<fp_timing_in>) {
	chop;
	@cols = split(/\s+/, $_, 3);
	@id = split(/\//, $cols[0]);
	$begin_hash{$id[1]} = $cols[1] * (1/$frame_duration) - 1;
	$total_hash{$id[1]} = ($cols[2] * (1/$frame_duration)) 
	    - ($cols[1] * (1/$frame_duration)) + 1 ;
    }
    
    # loop over each utterance in the mfcc list file and strip the silence
    # 
    for ($i = 0; $i <= $#mfcc_list; $i++) {
	
	$line = $mfcc_list[$i];
	chop($line);
	
	@data = split(/\//, $line);
	$len_mfc = @data;
	
        # get the path of the output_path is not specified
        #
	if ($path_flag == 0) {
	    $out_path = "";
	    for ($j = 0; $j <= $#data-1 ; $j++) { 
		$out_path = join("/", $out_path, $data[$j]);
	    }
	}
	
	# create the subdirectory if the output path is not provided
	#
	if ($path_flag == 1) {
	    $out_path = join("/", $output_path, $data[$len_mfc-2]);
	    (isip_mkdir("$out_path") == 1)
		or isip_die("ERROR: could not create necessary output directories...exiting");	    
	} 

	# get the id of the mfc file
	#
	$name_mfcc = $data[$len_mfc-1];
	@name_data = split(/\./, $name_mfcc);
	$len = @name_data;
	
	$name = $name_data[$len-2];	
	
	# get the begining frame and total frames from the hash table
	#
	$begin_frame = $begin_hash{"$name"};
	$total_frame = $total_hash{"$name"};
	
	# join the path
	#
	$out_path = join ("/", $out_path, $name);
	$out_path = join (".", $out_path, $extension);
	
	# strip the silence at start and end for HTK format features
	# 
        if ($feature_format eq "htk") {
	    
	    # get the size of each frame for HTK feaures
	    #
	    $size_frame = $dimension*4;
	    
	    # open the input and output feature file in the binary mode
	    #
	    open INPUT, "$line" or isip_die("Can't open $line\n");
	    open OUTPUT, ">$out_path" or isip_die("Can't open $out_path\n");
	    binmode INPUT;
	    binmode OUTPUT;
	    
	    # read the 12 bytes HTK header and write it to the output file
	    #
	    read INPUT, $buf, 12;
	    print OUTPUT $buf;
	    
	    # write the stripped utterance to the ouput file
	    #
	    $start = 12 + ("$begin_frame" * "$size_frame"); 	    
	    $total = $total_frame * $size_frame;
	    seek(INPUT, $start, 0);
	    read(INPUT, $buf, $total) ;	    
	    print OUTPUT $buf;	 	    
	}

	# strip the silence at start and end for isip_proto format features
	# 
        if ($feature_format eq "isip_proto") {
	    
	    # get the size of each frame for isip_proto feaures (doubles)
	    #
	    $size_frame = $dimension*8;
	    
	    # open the input and output feature file in the binary mode
	    #
	    open INPUT, "$line" or isip_die("Can't open $line\n");
	    open OUTPUT, ">$out_path" or isip_die("Can't open $out_path\n");
	    binmode INPUT;
	    binmode OUTPUT;
	    
	    # write the stripped utterance to the ouput file
	    #
	    $start = 12 + ("$begin_frame" * "$size_frame"); 	    
	    $total = $total_frame * $size_frame;
	    seek(INPUT, $start, 0);
	    read(INPUT, $buf, $total) ;	    
	    print OUTPUT $buf;	 	    
	}

	# close the input and output_files
	# 
	(isip_close(INPUT) == 1) or return 0;
	(isip_close(OUTPUT) == 1) or return 0;
	
    }
    
    (isip_close(fp_mfcc_in) == 1) or return 0;
    (isip_close(fp_timing_in) == 1) or return 0;
    
    # exit gracefully
    #
    return 1;
}

#----------------------------------------------------------------------------
#
# parameter file templates
#
#----------------------------------------------------------------------------

# create a parameter file template for bw_train
#
$BW_TRAIN_PARAM_FILE_TEMPLATE = <<__BW_PARAM_FILE_END__;
# parameter file for Baum Welch training
#

# define training mode (normal, batch or combine)
# - in normal mode, the parameters are updated on the given input data
# - in batch mode, accumulator files are output but no parameter updating
#   is done. combine mode must be used to combine the accumulator files and
#   perform the reestimation
# - in combine mode, accumulator files are combined and used to update the
#   HMM parameters
#
# train_mode = combine
# train_mode = batch
train_mode = __TRAIN_MODE__

# define accumulator file 
#
acc_file = __ACCUMULATOR_FILE__

# in combine mode you define an accumulator list containing all of the 
# accumulator filenames produced in batch mode
#
acclist_file = __ACCUMULATOR_LIST_FILE__

# define input file list
#
mfcclist_file = __MFCC_LIST__

# define model-level transcription file
#
lablist_file = __LABEL_LIST__

# define the file containing the list of context-independent models
#
monophones_file = __MONOPHONES_FILE__

# define the model definition file
#
models_file = __MODELS_FILE__

# define phone mapping file created by create_triphone_map
#
phones_file = __PHONES_FILE__

# define lexicon file
#
lexicon_file = __LEXICON_FILE__

# define the input states file
#
states_file = __INPUT_STATES__

# define the input transitions file
#
transitions_file = __INPUT_TRANSITIONS__

# in combine mode or normal mode we define the states file which will 
# contain the updated HMM states
#
updated_states_file = __OUTPUT_STATES__

# in combine mode or normal mode we define the transitions file which will
# contain the updated HMM transitions
#
updated_trans_file = __OUTPUT_TRANSITIONS__

# define input(mfcc)/output(states) mode file format (ascii or binary)
# accumulator files are always written in binary
#
input_mode = __INPUT_MODE__
output_mode = __OUTPUT_MODE__

# define context mode (monophone, word_internal, or cross_word)
# 
context_mode = __CONTEXT_MODE__

# define training thresholds.
# - min_mpd is the minimum model probability deviance
# - beam_width is the beam width used for beta pruning
# - min_occupancy is a floor on the occupancy probability
# - min_model_count is the minimum number of times a model must occur before
#   parameters of that model are updated
#
min_mpd = __MIN_MPD__
beam_width = __BEAM_WIDTH__
min_occupancy = __MIN_OCC__
min_model_count = __MIN_MODEL_COUNT__

# define variance floor file
# 
variance_floor_file = __VFLOOR_FILE__

# define the state occupancy file. if this mode is turned on then the
# utility will dump state occupancies at the end of the training run. this 
# is a necessary input to the state tying utility.
# the occupancy mode is either "on" or "off"
#
state_occupancy_file = __STATE_OCC_FILE__
state_occupancy = __STATE_OCC_MODE__

# define input feature file format (isip_proto or htk(default))
#
input_feature_format = __INPUT_FEATURE_FORMAT__

# define parameters for the computation of delta and acceleration features
# options for compute_delta and compute_acc are yes/no 
#
compute_delta = __COMPUTE_DELTA__
compute_acc = __COMPUTE_ACC__
delta_win = __DELTA_WIN__
__BW_PARAM_FILE_END__


#-----------------------------------------------------------------------------
#
# create a parameter file template for init_triphones
#
$ITRI_PARAM_FILE_TEMPLATE = <<__ITRI_PARAM_FILE_END__;
# parameter file for init_triphones
#

# context-independent parameters
#
in_models_file = __IN_MODELS_FILE__
in_states_file = __IN_STATES_FILE__
in_trans_file = __IN_TRANS_FILE__
ci_phones_file = __CI_PHONES_FILE__

# clustered triphones list
#
tri_clist_file = __TRI_CLIST_FILE__

# output files
#
out_models_file = __OUT_MODELS_FILE__
out_states_file = __OUT_STATES_FILE__
__ITRI_PARAM_FILE_END__

#-----------------------------------------------------------------------------
#
# create a parameter file template for train mode of tie_state
#
$TS_TRAIN_PARAM_FILE_TEMPLATE = <<__TS_TRAIN_PARAM_FILE_END__;
# parameters file for state tying using decision tree 
#

# define the mode 
#
mode = train

# common parameters for both mode
# 
# define monophones file, state file, membership, out model  file
# 
monophone_file = __MONOPHONES_FILE__
membership_file = __MEMBERSHIP_FILE__

# define input/ouput mode for the states file
#
input_mode = __INPUT_MODE__
output_mode = __OUTPUT_MODE__

# debug level
#
debug_level = 0

# define the input state and model file
#
in_states_file = __IN_STATES_FILE__
in_models_file = __IN_MODELS_FILE__

# define the input state and model file
#
special_models_file = __SPECIAL_MODELS_FILE__

# define state occupancy file
#
occupancy_file = __OCCUPANCY_FILE__

# define thresholds
#
split_threshold = __SPLIT_THRESHOLD__
merge_threshold = __MERGE_THRESHOLD__
num_occ_threshold = __OCC_THRESHOLD__
num_context = __NUM_CONTEXT__

# define the output tree file, state file, models file, clist file
#
out_tree_file = __OUTPUT_TREES_FILE__
out_states_file = __OUTPUT_STATES_FILE__
out_models_file = __OUTPUT_MODELS_FILE__
out_clist_file = __OUTPUT_CLIST_FILE__

#
# end of file
__TS_TRAIN_PARAM_FILE_END__


#-----------------------------------------------------------------------------
#
# create a parameter file template for test mode of tie_state
#
$TS_TEST_PARAM_FILE_TEMPLATE = <<__TS_TEST_PARAM_FILE_END__;
# parameters file for state tying using decision tree 
#

# define the mode 
#
mode = test

# common parameters for both mode
# 
# define monophones file, state file, membership, out model  file
# 
monophone_file = __MONOPHONES_FILE__
membership_file = __MEMBERSHIP_FILE__

# define input/ouput mode for the states file
#
input_mode = __INPUT_MODE__
output_mode = __OUTPUT_MODE__

# debug level
#
debug_level = 0

# define the input state and model file
#
in_states_file = __IN_STATES_FILE__
in_models_file = __IN_MODELS_FILE__

# define the input state and model file
#
special_models_file = __SPECIAL_MODELS_FILE__

# define thresholds
#
split_threshold = __SPLIT_THRESHOLD__
merge_threshold = __MERGE_THRESHOLD__
num_occ_threshold = __OCC_THRESHOLD__
num_context = __NUM_CONTEXT__

# define the input tree file and phone file
#
in_tree_file = __INPUT_TREE_FILE__
in_phlist_file = __INPUT_PHONE_LIST_FILE__

# define the output tree file, state file, models file, clist file
#
out_models_file = __OUTPUT_MODELS_FILE__
out_clist_file = __OUTPUT_CLIST_FILE__

#
# end of file
__TS_TEST_PARAM_FILE_END__


#-----------------------------------------------------------------------------
#
# create a parameter file template for hmm_train
#
$HT_PARAM_FILE_TEMPLATE = <<__HT_PARAM_FILE_END__;
# parameter file for Viterbi training
#
  
# define training mode (normal, batch or combine)
# - in normal mode, the parameters are updated on the given input data
# - in batch mode, accumulator files are output but no parameter updating
#   is done. combine mode must be used to combine the accumulator files and
#   perform the reestimation
# - in combine mode, accumulator files are combined and used to update the
#   HMM parameters
#
train_mode = __TRAIN_MODE__
  
# define accumulator file 
#
batch_stats_file = __BATCH_STATS_FILE__
  
# in combine mode you define an accumulator list containing all of the 
# accumulator filenames produced in batch mode
#
acc_list_file = __ACCUMULATOR_LIST_FILE__
  
# define input file list
#
input_file = __MFCC_LIST__

# define output file list of best paths (if this is not set then no
# output will be written)
#
output_file = __OUTPUT_LIST__

# define the transcript file and the type of transcription mode.
# options are word_mlf and model_mlf
#
mlf_file = __MLF_FILE__
mlf_mode = __MLF_MODE__

# define the file containing the list of context-independent models
#
monophones_file = __MONOPHONES_FILE__

# define the file containing a list of models which are not expanded
# for context-dependency. typically these are silence models
#
special_model_list = __SPECIAL_MODELS_FILE__

# define the model definition file
#
models_file = __MODELS_FILE__

# define model mapping file created by create_triphone_map
#
phones_file = __PHONES_FILE__

# define lexicon file
#
lexicon_file = __LEXICON_FILE__

# define the input states file
#
states_file = __INPUT_STATES__

# define the input transitions file
#
transitions_file = __INPUT_TRANSITIONS__

# in combine mode or normal mode we define the states file which will 
# contain the updated HMM states
#
updated_states_file = __OUTPUT_STATES__

# in combine mode or normal mode we define the transitions file which will
# contain the updated HMM transitions
#
updated_trans_file = __OUTPUT_TRANSITIONS__

# define alignment mode : model or state (always use state for training)
#
align_mode = __ALIGN_MODE__

# define input(mfcc)/output(states) mode file format (ascii or binary)
# accumulator files are always written in binary
#
input_mode = __INPUT_MODE__
output_mode = __OUTPUT_MODE__

# define context mode (monophone, word_internal, or cross_word)
# 
context_mode = __CONTEXT_MODE__
  
# define parameters for Viterbi decoding
# - n-best determines how many paths should be kept at each word-end
# - context_size specifies the size of context expansion (3 = triphones)
# - beam_width specifies the beam pruning thresholds (state, model, word)
# - max_active_phones specifies the maximum number of active phone instances
#
n-best = __NBEST_SIZE__
context_size = __CONTEXT_SIZE__
beam_width = __STATE_BEAM_WIDTH__ __MODEL_BEAM_WIDTH__ __WORD_BEAM_WIDTH__
max_active_phones = __MAPMI_THRESH__

# define the state occupancy file. if this mode is turned on then the
# utility will dump state occupancies at the end of the training run. this 
# is a necessary input to the state tying utility.
# the occupancy mode is either "on" or "off"
#
state_occupancy_file =  __STATE_OCCUPANCY_FILE__
state_occupancy = __STATE_OCC_MODE__
__HT_PARAM_FILE_END__

#-----------------------------------------------------------------------------
#
# create a parameter file template for forced alignment
#
$ALIGN_PARAM_FILE_TEMPLATE = <<__ALIGN_PARAM_FILE_END__;
# parameters file for the isip decoder
#

# define function mode
# - forced_alignment: time align a set of known transcriptions using a 
#   Viterbi DP
#
function = forced_alignment
   
# define input acoustic feature file list. this parameter is required
# in all modes but lattice_error_rate
#
mfcc_list = __MFCC_LIST__

# define the output hypothesis file list. this list should match one-to-one
# with the mfcc_list. this parameter is required in all modes which 
# generate hypotheses: forced_alignment, lattice_rescoring, ngram_decoding,
# lattice_generation, lattice_verification, lattice_to_lattice
#
output_list = __OUTPUT_LIST__

# define transcription file. this list should match one-to-one with the 
# input lists. this parameter is required only in lattice_error_rate mode
# and forced_alignment mode
#
transcription_file = __TRANSCRIPTION_FILE__

# define the file containing the list of context-independent models.
# required for all modes but lattice_error_rate
#
monophones_file = __MONOPHONES_FILE__

# define the model definition file
# required for all modes but lattice_error_rate
#
models_file = __MODELS_FILE__

# define phone mapping file created by create_triphone_map
# required for all modes but lattice_error_rate
#
phones_file = __PHONES_FILE__
  
# define lexicon file
# required for all modes but lattice_error_rate
#
lexicon_file = __LEXICON_FILE__

# define the input states file
# required for all modes but lattice_error_rate
#
states_file = __INPUT_STATES__

# define the input transitions file
# required for all modes but lattice_error_rate
#
transitions_file = __INPUT_TRANSITIONS__

# define the output alignment mode (word, model, state). this parameter
# determines how detailed the hypothesis alignments are. in state mode
# the alignments are a state-frame mapping. beware that in state mode,
# more information is stored by the decoder so the memory consumption
# may be greater. also note that align_mode must be set to word when 
# doing n-best generation. this parameter is required for all modes
# which produce hypothesis output: forced_alignment, lattice_rescoring,
# ngram_decoding, lattice_generation, lattice_verification,
# lattice_to_lattice
#
align_mode = model

# define input acoustic feature file format (ascii or binary). note
# that the states file should always be in binary mode. this parameter
# is required for all modes which use acoustic feature file input: 
# forced_alignment, lattice_rescoring, ngram_decoding, lattice_generation,
# lattice_verification, lattice_to_lattice
#
input_format = __INPUT_FORMAT__

# define context mode (monophone, word_internal, or cross_word). this 
# parameter is required for all modes
#
context_mode = __CONTEXT_MODE__

# define pruning thresholds
# - beam_pruning specifies the beam pruning thresholds (state, model, word)
# - mapmi_limit specifies the maximum number of active phone instances
# - num_allowed_words specifies the maximum number of hypothesized word ends
#   which are allowed to be extended at any time frame
# these parameters are required for all modes which perform acoustic
# decoding: forced_alignment, lattice_rescoring, ngram_decoding, 
# lattice_generation, lattice_verification, lattice_to_lattice
# 
beam_pruning = __STATE_BEAM__ __MODEL_BEAM__ __WORD_BEAM__
mapmi_limit = __MAPMI_LIMIT__
num_allowed_words = __NUM_ALLOWED_WORDS__
__ALIGN_PARAM_FILE_END__

#-----------------------------------------------------------------------------
#
# create a parameter file template for trace_projector
#
$TP_PARAM_FILE_TEMPLATE = <<__TP_PARAM_FILE_END__;
# parameters file for the isip decoder
#

# define function mode
# - forced_alignment: time align a set of known transcriptions using a 
#   Viterbi DP
# - lattice_rescoring: perform an acoustic rescoring of lattice files
#   the language model scores are taken from the lattice file
# - ngram_decoding: decode using an n-gram language model. note that for
#   most applications this is very memory intensive
# - lattice_generation: generate a lattice representing a narrowed
#   search space. typically we use an n-gram language model to generate
#   the lattice and then run with a stronger n-gram language model and 
#   stronger acoustic models in "lattice_verification" mode
# - lattice_verification: performs language model and acoustic model 
#   rescoring of lattice files. this is typically the followup stage to
#   running in lattice_generation mode
# - lattice_to_lattice: generate a lattice from a lattice. similar to 
#   lattice_generation mode but uses an existing lattice as the initial
#   search space rather than the n-gram search space
# - lattice_error_rate: calculates the error rate in the lattice by doing
#   a DP alignment of a reference transcription to the lattice
#
# function = forced_alignment
# function = lattice_rescoring
# function = ngram_decoding
# function = lattice_generation
# function = lattice_verification
# function = lattice_to_lattice
# function = lattice_error_rate
function = __DECODE_MODE__
   
# define input acoustic feature file list. this parameter is required
# in all modes but lattice_error_rate
#
mfcc_list = __MFCC_LIST__

# define the output hypothesis file list. this list should match one-to-one
# with the mfcc_list. this parameter is required in all modes which 
# generate hypotheses: forced_alignment, lattice_rescoring, ngram_decoding,
# lattice_generation, lattice_verification, lattice_to_lattice
#
output_list = __OUTPUT_LIST__

# define input lattice list. this list should match one-to-one with the
# mfcc_list. this parameter is required in all modes which accept lattices
# as input: lattice_rescoring, lattice_verification, lattice_to_lattice,
# lattice_error_rate
#
input_lattice_list = __INPUT_LATTICE_LIST__

# define transcription file. this list should match one-to-one with the 
# input lists. this parameter is required only in lattice_error_rate mode
# and forced_alignment mode
#
transcription_file = __TRANSCRIPTION_FILE__

# define lattice_error_rate output file. this parameter is required
# only in lattice_error_rate mode
#
lattice_wer_file = __LATTICE_WER_FILE__

# define output lattice list. this list should match one-to-one with the
# mfcc_list. this parameter is required in any mode which generates a 
# lattice: lattice_generation, lattice_to_lattice
#
output_lattice_list = __OUTPUT_LATTICE_LIST__

# define grammar list file. this is a list of defined grammars in the 
# dynamic language models. 
#
grammar_list_file = __GRAMMAR_LIST_FILE__

# define the file containing the list of context-independent models.
# required for all modes but lattice_error_rate
#
monophones_file = __MONOPHONES_FILE__

# define the model definition file
# required for all modes but lattice_error_rate
#
models_file = __MODELS_FILE__

# define phone mapping file created by create_triphone_map
# required for all modes but lattice_error_rate
#
phones_file = __PHONES_FILE__
  
# define lexicon file
# required for all modes but lattice_error_rate
#
lexicon_file = __LEXICON_FILE__

# define the input states file
# required for all modes but lattice_error_rate
#
states_file = __STATES_FILE__

# define the input transitions file
# required for all modes but lattice_error_rate
#
transitions_file = __TRANSITIONS_FILE__

# define the output alignment mode (word, model, state). this parameter
# determines how detailed the hypothesis alignments are. in state mode
# the alignments are a state-frame mapping. beware that in state mode,
# more information is stored by the decoder so the memory consumption
# may be greater. also note that align_mode must be set to word when 
# doing n-best generation. this parameter is required for all modes
# which produce hypothesis output: forced_alignment, lattice_rescoring,
# ngram_decoding, lattice_generation, lattice_verification,
# lattice_to_lattice
#
align_mode = __ALIGN_MODE__

# define an n-gram backoff language model and relevant n-gram parameters:
# - lm_file: the language model file is in ARPA n-gram backoff format
# - ngram_order: N >= 1. 1=unigram, 2=bigram, 3=trigram
# - lm_scale: is a prescaling factor used to scale the n-gram score before
#   incorporating it into the likelihood function during the search
# - word_penalty: is a factor to add to any hypothesis which arrives at a 
#   word end. this has the effect of discouraging the hypothesis of a long
#   string of short duration words (common in LVCSR systems)
# these parameters should be defined for any mode which uses an n-gram
# language model: ngram_decoding, lattice_generation, lattice_verification
#
lm_file = __LM_FILE__
ngram_order = __NGRAM_ORDER__
lm_scale = __LM_SCALE__
word_penalty = __WORD_PENALTY__

# define number of history paths and n-best paths to preserve (N > 0)
# - history_paths defines the number of n-best paths to keep at each 
#   point in the search space. note that the history_paths should be set
#   to 1 in n-gram decoding mode. this is required for modes where lattices
#   are generated: lattice_generation, lattice_to_lattice and for 
#   ngram_decoding mode
# - nbest_paths defines the number of final sentence hypotheses to output.
#   this parameter is required for any mode which produces a hypothesis:
#   lattice_rescoring, ngram_decoding, lattice_generation,
#   lattice_verification, lattice_to_lattice
#
history_paths = __HISTORY_PATHS__
nbest_paths = __NBEST_PATHS__

# define input acoustic feature file format (ascii or binary). note
# that the states file should always be in binary mode. this parameter
# is required for all modes which use acoustic feature file input: 
# forced_alignment, lattice_rescoring, ngram_decoding, lattice_generation,
# lattice_verification, lattice_to_lattice
#
input_format = __INPUT_FORMAT__

# define context mode (monophone, word_internal, or cross_word). this 
# parameter is required for all modes
#
context_mode = __CONTEXT_MODE__

# define pruning thresholds
# - beam_pruning specifies the beam pruning thresholds (state, model, word)
# - mapmi_limit specifies the maximum number of active phone instances
# - num_allowed_words specifies the maximum number of hypothesized word ends
#   which are allowed to be extended at any time frame
# these parameters are required for all modes which perform acoustic
# decoding: forced_alignment, lattice_rescoring, ngram_decoding, 
# lattice_generation, lattice_verification, lattice_to_lattice
# 
beam_pruning = __STATE_BEAM__ __MODEL_BEAM__ __WORD_BEAM__
mapmi_limit = __MAPMI_LIMIT__
num_allowed_words = __NUM_ALLOWED_WORDS__

# define input feature file format (isip_proto or htk(default))
#
input_feature_format = __INPUT_FEATURE_FORMAT__

# define parameters for the computation of delta and acceleration features
# options for compute_delta and compute_acc are yes/no 
#
compute_delta = __COMPUTE_DELTA__
compute_acc = __COMPUTE_ACC__
delta_win = __DELTA_WIN__
__TP_PARAM_FILE_END__

#-----------------------------------------------------------------------------
#
# return 1 to make perl happy
#
return 1;

#
# end of file
