#!/usr/bin/perl
#
# file: wsj_exp/wsj_decode.pl.in
#
$ISIP_HELP_FILE = <<__ISIP_HELP_HERE_FILE__;
name: wsj_decode
synopsis: 
   wsj_decode -mode <mode> -feature_format <format> 
   -split_threshold <threshold_to_split_a_node_in_a_decision_tree>
   -merge_threshold <threshold_to_merge_two_nodes_in_a_decision_tree>
   -num_occ_threshold <min_num_of_occupancy_to_split_a_node_in_a_decision_tree>
   -lm <language_model> -lm_scale <language_model_scale> 
   -word_penalty <word_insertion_penalty> 
   -test_state_beam_pruning <state_level_beam_pruning_for_testing>
   -test_model_beam_pruning <model_level_beam_pruning_for_testing>
   -test_word_beam_pruning <word_level_beam_pruning_for_testing>
   -models_path <base_directory_of_models>
   -test_mfc_list <list_of_testing_feature_files> 
   -cpus_test <cpus_used_for_testing>


descr: 
   this program decodes and evaluates using the models built by the 
   training module.

example: 
   wsj_decode  -mode acc -feature_format htk -lm devel 
   -test_mfc_list test.list -cpus_test isip204 isip205

options:
   -mode <mode> : 
      this option tells the script whether you want to compute deltas or both 
      delta/accelerations or neither on the fly. Choices are:
      delta    : computes deltas on the fly
      acc      : computes accelerations on the fly
      none     : does not computes deltas or accelerations on the fly (default)

   -feature_format <format> : 
      specifies the feature format to use. Choices are:
      htk        : HTK format features (default)
      isip_proto : ISIP format features

  -split_threshold <threshold_to_split_a_node_in_decision_tree> : 
      specifies the minimum increase in likelihood to split a node in the 
      decision tree during the state tying process in test mode(default is 165)

   -merge_threshold <threshold_to_merge_two_node_in_decision_tree> : 
      specifies the maximum decrease in likelihood to merge two nodes in the
      decision tree during the state tying process in test mode(default is 165)

   -num_occ_threshold <min_num_of_occupancy_to_split_a_node_in_decision_tree> :
      specifies the minimum number of states for splitting a node in the 
      decision tree during the state tying process in test mode(default is 840)

   -lm <language_model> : 
      specifies the language model for testing. Choices are:
      devel : language model for WSJ devtest set (default)
      eval  : language model for WSJ evaltest set

   -lm_scale <language_model_scale> : 
      specifies the weight for the langauge model score (default is 18)

   -word_penalty <word_insertion_penalty> : 
      specifies the penalty applied to each word that is hypothesized 
      (default is 10)

   -test_state_beam_pruning <state_level_beam_pruning_for_testing> : 
      specifies the state-level beam pruning during decoding (default is 300) 

   -test_model_beam_pruning <model_level_beam_pruning_for_testing> : 
      specifies the model-level beam pruning during decoding (default is 250) 

   -test_word_beam_pruning <word_level_beam_pruning_for_testing> : 
      specifies the word-level beam pruning during decoding (default is 250) 

   -test_mfc_list <list_of_test_feature_files> :
      specifies the testing feature files list

   -cpus_test <cpus_used_to_test>
      specifies the cpus to be employed for testing

   -models_path <base_directory_of_models>
      specifies the models base-directory to be employed for testing
      (default is the current directory)

arguments: none

man page: none
__ISIP_HELP_HERE_FILE__

# define the path to this script
#
$WSJ_PATH = "/cavs/hse/ies/isip/users/raghavan/aurora/cluster_enabled/asr_wsj_tutorial_v1.4.2";

# system packages
#
use File::Basename;
use File::Path;
use File::Copy;
use FileHandle;
autoflush STDOUT;

# isip libraries
#
require "$WSJ_PATH/lib/scripts/perl/command_line.pm";
require "$WSJ_PATH/lib/scripts/perl/wsj_subs.pm";

#----------------------------------------------------------------------------
#
# constant and global variable definitions
#
#----------------------------------------------------------------------------

# define the program name
#
@tmp = split(/\//, $0);
$ISIP_PROG = pop(@tmp);

# define the program to check for endian-ness
#
$ISIP_BINARY_NAME = $ENV{ISIP_BINARY};
$BIG_ENDIAN = "0";
$LITTLE_ENDIAN = "1";
$BIG_ENDIAN_CHECKER = "$WSJ_PATH/bin/$ISIP_BINARY_NAME/check_endian";

# define a separator for output comments
#
$TI_LINE_SEP = "-------------------  $ISIP_PROG  ----------------";

# get the current working directory
#
$TI_PATH_TO_CUR_DIR = `pwd`;
chop($TI_PATH_TO_CUR_DIR);

# define files and locations to be used by this script
#
$TI_PATH_TO_DATA = "$WSJ_PATH/data";
$TI_PATH_TO_DEC_DATA = "$TI_PATH_TO_DATA/decode";
$TI_PATH_TO_DECODE = "$TI_PATH_TO_CUR_DIR/decode";
$TI_PATH_TO_TRAIN = "$TI_PATH_TO_CUR_DIR/train";
$TI_PATH_TO_BW = "$TI_PATH_TO_TRAIN/baum_welch";
$TI_PATH_TO_MFC = "$TI_PATH_TO_CUR_DIR/data_generation";

$TI_PATH_TO_OUTPUT_LISTS = "$TI_PATH_TO_MFC/lists";
$TI_PATH_TO_TEST_LIST = "$TI_PATH_TO_OUTPUT_LISTS/test_mfc.list";
$TI_PATH_TO_CPUS_TEST = "$TI_PATH_TO_OUTPUT_LISTS/cpus_test.list";
$TI_PATH_TO_DATA_TEST = "$WSJ_PATH/data/decode";

# define decoder functional modes
#
$TP_FUNC_NGRAM_DEC = "ngram_decoding";

# define command-line options
#
$TI_OPT_MODE = "-mode";
$TI_OPT_FEAT_FORMAT = "-feature_format";
$TI_OPT_SP_THRESH = "-split_threshold";
$TI_OPT_ME_THRESH = "-merge_threshold";
$TI_OPT_NUM_OCC_THRESH = "-num_occ_threshold";
$TI_OPT_LM = "-lm";
$TI_OPT_LM_SCALE = "-lm_scale";
$TI_OPT_WORD_PEN = "-word_penalty";
$TI_OPT_TE_ST_BEAM_PRUN = "-test_state_beam_pruning";
$TI_OPT_TE_MO_BEAM_PRUN = "-test_model_beam_pruning";
$TI_OPT_TE_WORD_BEAM_PRUN = "-test_word_beam_pruning";
$TI_OPT_TEST_MFC_LIST = "-test_mfc_list";
$TI_OPT_CPUS_TEST = "-cpus_test";
$TI_OPT_MODELS_PATH = "-models_path";

$TI_MODE_DELTA = "delta";
$TI_MODE_ACC = "acc";
$TI_MODE_NONE = "none";
@TI_MODE_TYPES = ($TI_MODE_DELTA, $TI_MODE_ACC, $TI_MODE_NONE);

$TI_FEAT_FORMAT_HTK = "htk";
$TI_FEAT_FORMAT_ISIP = "isip_proto";
@TI_FEAT_FORMAT_TYPES = ($TI_FEAT_FORMAT_HTK, $TI_FEAT_FORMAT_ISIP);

$TI_LM_DEVEL = "devel";
$TI_LM_EVAL = "eval";
@TI_LM_TYPES = ($TI_LM_DEVEL, $TI_LM_EVAL);

# define defaults
#
$TI_MODE_DEF = $TI_MODE_NONE;
$TI_FEAT_FORMAT_DEF = $TI_FEAT_FORMAT_HTK;
$TI_SP_THRESH_DEF = "165";
$TI_ME_THRESH_DEF = "165";
$TI_NUM_OCC_THRESH_DEF = "840";
$TI_LM_DEF = $TI_LM_DEVEL;
$TI_LM_SCALE_DEF = "18";
$TI_WORD_PEN_DEF = "10";
$TI_TE_ST_BEAM_PRUN_DEF = "300";
$TI_TE_MO_BEAM_PRUN_DEF = "250";
$TI_TE_WORD_BEAM_PRUN_DEF = "250";

#----------------------------------------------------------------------------
#
# command line processing
#
#----------------------------------------------------------------------------
# print a debugging message
#
print "$TI_LINE_SEP\n";
print "INFO: wsj_decode - checking command-line arguments\n\n";

# parse the command line
#
my ( $mode, $feature_format, $split_threshold, $merge_threshold, 
     $num_occ_threshold, $lm, $lm_scale, $word_penalty, 
     $test_state_beam_pruning, $test_model_beam_pruning,
     $test_word_beam_pruning, $test_mfc_list, $cpus_test, $models_path) 
    =  command_line(0, $TI_OPT_MODE, 1, $TI_OPT_FEAT_FORMAT, 1, 
		    $TI_OPT_SP_THRESH, 1, $TI_OPT_ME_THRESH, 1,
                    $TI_OPT_NUM_OCC_THRESH, 1, $TI_OPT_LM, 1, 
		    $TI_OPT_LM_SCALE, 1, $TI_OPT_WORD_PEN, 1,
		    $TI_OPT_TE_ST_BEAM_PRUN, 1, $TI_OPT_TE_MO_BEAM_PRUN, 1, 
		    $TI_OPT_TE_WORD_BEAM_PRUN, 1, $TI_OPT_TEST_MFC_LIST, 1,
		    $TI_OPT_CPUS_TEST, -1, $TI_OPT_MODELS_PATH, 1);

# set defaults if necessary
#
$model_type = "xwrd_triphone";
$train_mode = "baum_welch";

if ($mode eq "") {
    $mode = $TI_MODE_DEF;
}

if ($feature_format eq "") {
    $feature_fromat = $TI_FEAT_FORMAT_DEF;
}

if ($split_threshold eq "") {
    $split_threshold = $TI_SP_THRESH_DEF;
}

if ($merge_threshold eq "") {
    $merge_threshold = $TI_ME_THRESH_DEF;
}

if ($num_occ_threshold eq "") {
    $num_occ_threshold = $TI_NUM_OCC_THRESH_DEF;
}

if ($lm eq "") {
    $lm = $TI_LM_DEF;
}

if ($lm_scale eq "") {
    $lm_scale = $TI_LM_SCALE_DEF;
}

if ($word_penalty eq "") {
    $word_penalty = $TI_WORD_PEN_DEF;
}

if ($test_state_beam_pruning eq "") {
    $test_state_beam_pruning = $TI_TE_ST_BEAM_PRUN_DEF;
}

if ($test_model_beam_pruning eq "") {
    $test_model_beam_pruning = $TI_TE_MO_BEAM_PRUN_DEF;
}

if ($test_word_beam_pruning eq "") {
    $test_word_beam_pruning = $TI_TE_WORD_BEAM_PRUN_DEF;
}

if ($cpus_test eq "") {
    isip_die ("ERROR: cpus_test not specified");
}
if ($models_path ne  "") {
    $TI_PATH_TO_TRAIN = "$models_path/train";
    $TI_PATH_TO_BW = "$models_path/baum_welch";
}

# complete the path of the training list if absolute path is not given
# from the commandline
#
#if (substr($test_mfc_list, 0, 1) ne "/") {
#    $test_mfc_list = "$WSJ_PATH/$test_mfc_list";
#}
    
# create the list of cpus used for testing
#
@cpus_test_list = split(/\s+/, $cpus_test);

# create sub-directory name for testing cpus
#
$j = 0;
for ($i = 0; $i <= $#cpus_test_list; $i++) {
 
    if ($cpus_test_list[$i] ne $cpus_test_list[$i-1]) {
       $j = 0;
    }

    if (($cpus_test_list[$i] eq $cpus_test_list[$i-1]) && ($i != 0)) {
       $j++; 
    }

    $cpus_test_num[$i] = join("_", $cpus_test_list[$i], $j);
}

# don'tsort the testing cpus
#
@sort_cpus_test = @cpus_test_num;

#----------------------------------------------------------------------------
#
# file and directory verification
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_decode - checking for input data directory and files\n";
print "INFO: checking for input data directory and files\n";

# verify the input data directory exists
#
if (! -d "$TI_PATH_TO_DATA_TEST") {
    isip_die ("ERROR: input data directory not found: $TI_PATH_TO_DATA_TEST\n\t. you may have an incomplete distribution...exiting");
}

# verify that the input mfcc files exist
#
(check_files_in_list($test_mfc_list) == 1) 
    or isip_die("ERROR: some input test files not found...exiting");

# print a debugging message
#
print "INFO: wsj_decode - all decoding mfc files exist\n";

# print a debugging message
#
print "INFO: wsj_decode - creating local data directories\n";

# create directory structure for the output lists and transcriptions
#
if (! -d "$TI_PATH_TO_MFC") {
    (isip_mkdir($TI_PATH_TO_MFC) == 1) 
	 or isip_die ("ERROR: error creating directory $TI_PATH_TO_MFC ...exiting");
}
if (! -d "$TI_PATH_TO_OUTPUT_LISTS") {
    (isip_mkdir($TI_PATH_TO_OUTPUT_LISTS) == 1)
	 or isip_die ("ERROR: error creating directory $TI_PATH_TO_OUTPUT_LISTS ...exiting");
 }

# create directory structure for the output lists and transcriptions
# equal to the number of cpu's
#
for ($num = 0; $num <= $#sort_cpus_test; $num++) {
    
    if (! -d "$TI_PATH_TO_CUR_DIR/$sort_cpus_test[$num]") {
	
	(isip_mkdir("$TI_PATH_TO_CUR_DIR/$sort_cpus_test[$num]") == 1) or
	    isip_die ("ERROR: error creating directory $TI_PATH_TO_CUR_DIR/$sort_cpus_test[$num]...exiting");
	
	(isip_mkdir("$TI_PATH_TO_CUR_DIR/$sort_cpus_test[$num]/data_generation") 
	 == 1) or
	     isip_die ("ERROR: error creating directory $TI_PATH_TO_CUR_DIR/$sort_cpus_test[$num]/data_generation...exiting");
	
	(isip_mkdir("$TI_PATH_TO_CUR_DIR/$sort_cpus_test[$num]/data_generation/lists") == 1) or
	    isip_die ("ERROR: error creating directory $TI_PATH_TO_CUR_DIR/$sort_cpus_test[$num]/data_generation/lists...exiting");
    }
}

#----------------------------------------------------------------------------
#
# constructing lists for the decoding lists according to the number of cpu's
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_decode - copying and decoding lists to exp directory\n";

# copying the testing list to experimental subdirectory
#
(isip_copy($test_mfc_list, $TI_PATH_TO_TEST_LIST))
    or isip_die("ERROR: wsj_decode - testing mfc list not copied...exiting");

# print a debugging message
#
print "INFO: wsj_decode - creating decoding lists\n";
print "INFO: wsj_decode - equal to the number of corresponding cpu's\n";

# divide the testing lists N ways to build the decoding lists
#
(create_N_lists(($#sort_cpus_test + 1), $TI_PATH_TO_TEST_LIST, 
		$TI_PATH_TO_OUTPUT_LISTS, "test_mfc", "list") == 1) 
    or isip_die("ERROR: wsj_decode - test mfc lists not built...exiting");

#----------------------------------------------------------------------------
#
# moving the training lists to cpus directories
#
#----------------------------------------------------------------------------

print "INFO: wsj_decode - moving testing mfcc to respective subdirectories\n";
print "INFO: corresponding to the cpus..\n";

for( $i = 0; $i <= $#sort_cpus_test; $i++) {
    (isip_move("$TI_PATH_TO_OUTPUT_LISTS/test_mfc_$i.list", "$TI_PATH_TO_CUR_DIR/$sort_cpus_test[$i]/data_generation/lists/test_mfc.list"))
     or isip_die("ERROR: wsj_decode - test mfc sublists not moved...exiting");
}

#----------------------------------------------------------------------------
#
# checking endianness
#
#----------------------------------------------------------------------------

print "INFO: wsj_decode - checking endian-ness of your system...\n";

# check for endian-ness. note that the input data is big-endian.
#
$swap_value = `$BIG_ENDIAN_CHECKER`;
chop $swap_value;
$swap_value =~ s/^\s+//g;
$swap_value =~ s/\s+//g;

if ($swap_value eq $LITTLE_ENDIAN) {
    $swap_flag = "1";
    print "INFO: wsj_decode - found a little endian architecture\n";
    print "INFO: wsj_decode - Please check that the mfcc's are in little endian format\n";
##    print "INFO: wsj_decode - adding swap_bytes flag to command line\n";
}
else {
    $swap_flag = "";
    print "INFO: wsj_decode - found a big endian architecture\n";
    print "INFO: wsj_decode - Please check that the mfcc's are in big endian format\n";
}

#----------------------------------------------------------------------------
#
# build the commadline for the decoding to be displayed
#
#----------------------------------------------------------------------------

print "INFO: wsj_decode - running with parameters: \n";
$out_line = join (" ", $TI_OPT_MODE, $mode, $TI_OPT_FEAT_FORMAT, 
		  $feature_format,$TI_OPT_SP_THRESH, $split_threshold, 
                  $TI_OPT_ME_THRESH, $merge_threshold, 
                  $TI_OPT_NUM_OCC_THRESH, $num_occ_threshold, 
		  $TI_OPT_LM, $lm, $TI_OPT_LM_SCALE, 
		  $lm_scale, $TI_OPT_WORD_PEN, $word_penalty, 
                  $TI_OPT_TE_ST_BEAM_PRUN, $test_state_beam_pruning, 
                  $TI_OPT_MO_ST_BEAM_PRUN, $test_model_beam_pruning, 
                  $TI_OPT_WORD_ST_BEAM_PRUN, $test_word_beam_pruning, 
                  $TI_OPT_TEST_MFC_LIST, $test_mfc_list, 
		  $TI_OPT_CPUS_TEST, $cpus_test);

print "$out_line\n";

#----------------------------------------------------------------------------
#
# generate filenames and modes for the decoder depending on the training mode
#
#----------------------------------------------------------------------------

# determine the mode of operation of the decoder and set up decoder files
# needed by that mode
#
$dec_function = "";
$output_path = "";

# build the path to the working directory
#
$output_path = $TI_PATH_TO_DECODE;
$output_path = join("/", $output_path, "baum_welch");
$output_path = join("/", $output_path, "xwrd_tri");

# build the path to the working directory corresponding to the cpus
#
for ($i = 0; $i <= $#sort_cpus_test; $i++) {
    $output_path_cpus[$i] = "$TI_PATH_TO_CUR_DIR/$sort_cpus_test[$i]";
    $output_path_cpus[$i] = join("/", $output_path_cpus[$i], "decode");
    $output_path_cpus[$i] = join("/", $output_path_cpus[$i], "baum_welch");
    $output_path_cpus[$i] = join("/", $output_path_cpus[$i], "xwrd_tri");
}

# using the bigram language model
#
if ($dec_mode eq $TI_DEC_BIGRAM) {
    $dec_function = $TP_FUNC_NGRAM_DEC;
    $output_path = join("/", $output_path, "bigram_decoding");
    for ($i = 0; $i <= $#sort_cpus_test; $i++) {
	$output_path_cpus[$i] 
	    = join("/", $output_path_cpus[$i], "bigram_decoding");
    }
}
else {
    isip_die ("ERROR: invalid decoding mode $dec_mode");
}

print "INFO: wsj_decode - using working directory $output_path\n";

# make the output directories for the main working directory
#
((isip_mkdir($output_path) == 1) &&
 (isip_mkdir("$output_path/lists") == 1) &&
 (isip_mkdir("$output_path/output") == 1) &&
 (isip_mkdir("$output_path/lattices") == 1)) 
    or isip_die("ERROR: could not create necessary output directories...exiting");

# make the output directories for the working directories corresponding to the
# testing cpus
#
for ($i = 0; $i <= $#sort_cpus_test; $i++) {
    ((isip_mkdir($output_path_cpus[$i]) == 1) &&
     (isip_mkdir("$output_path_cpus[$i]/lists") == 1) &&
     (isip_mkdir("$output_path_cpus[$i]/output") == 1) &&
     (isip_mkdir("$output_path_cpus[$i]/lattices") == 1)) 
	or isip_die("ERROR: could not create necessary output directories...exiting");
}

# build a path to the training models and set parameters specific to that
# mode
#
$training_path = $TI_PATH_TO_TRAIN;
$training_path = join("/", $training_path, "baum_welch");

$training_path = join("/", $training_path, "xwrd_tri");
$context_mode = "cross_word";


print "INFO: wsj_decode - using models from $training_path\n";

# specify common parameters for each mode. all other parameters are specified
# in the mode block below
#
$mfcc_list = $TI_PATH_TO_TEST_LIST;
for ($i = 0; $i <= $#sort_cpus_test; $i++) {
    $mfcc_list_cpus[$i] = 
	"$TI_PATH_TO_CUR_DIR/$sort_cpus_test[$i]/data_generation/lists/test_mfc.list";
}
$monophones_file= "$training_path/final_models/monophones.text";
$models_file = "$training_path/state_tying/tied_models.text";
$phones_file = "$training_path/state_tying/tied_phones.text";
$states_file = "$training_path/final_models/states.bin";
$transitions_file = "$training_path/final_models/transitions.text";
$lexicon_file = "$TI_PATH_TO_DEC_DATA/decode_lexicon.text";

# specify the decoding parameters (thresholds, etc.)
#
$mapmi_limit = 10000;
$num_allowed_words = 150;
$input_format = $TI_BINARY_MODE;

# define defaults for all other parameters
#
$grammar_list_file = "";
$output_list = "";
$input_lattice_list = "";
$transcription_file = "";
$lattice_wer_file = "";
$output_lattice_list = "";
$align_mode = "word";
$ngram_order = "";
$history_paths = "1";
$nbest_paths = "1";

if ($mode eq $TI_MODE_ACC) {
    $compute_delta = "yes";
    $compute_acc = "yes";
    $delta_win = "2";
}
elsif ($mode eq $TI_MODE_DELTA) {
    $compute_delta = "yes";
    $compute_acc = "no";
    $delta_win = "2";
}
else {
    $compute_delta = "no";
    $compute_acc = "no";
    $delta_win = "";
}

$input_feature_format = $feature_format;

# set a flag to determine if we need to evaluate the output hypotheses or 
# not
#
$evaluate_hyps = 1;
$decoding_finished = 0;

#----------------------------------------------------------------------------
#
# decode with the appropriate models
#
#----------------------------------------------------------------------------

# set the bigram lm file
#
if ($lm eq $TI_LM_DEVEL) {
$lm_file = "$TI_PATH_TO_DEC_DATA/wsj_devel_bigram.arpa";
}
elsif ($lm eq $TI_LM_EVAL) {
$lm_file = "$TI_PATH_TO_DEC_DATA/wsj_eval_bigram.arpa";
}
else { isip_die ("ERROR: $lm language model doesn't exists. Incomplete distribution");
   }

$ngram_order = 2;

# build the output list
#
$output_list = "$output_path/lists/output.list";
for ($i = 0; $i <= $#sort_cpus_test; $i++) {
    $output_list_cpus[$i] = "$output_path_cpus[$i]/lists/output.list";
}

print "INFO: wsj_decode - building the output list\n";
print "INFO: wsj_decode - $output_list\n";

(build_output_list("$output_path/output", $TI_PATH_TO_TEST_LIST, 
		   $output_list) == 1)
    or isip_die ("ERROR: could not create $output_list");

print "INFO: wsj_decode - dividing the output list N ways according to the\n";
print "INFO: wsj_decode - number of test cpus\n";

# divide the output list N ways according to the number of cpus
#
(create_N_lists(($#sort_cpus_test + 1), $output_list, 
		"$output_path/lists", "output", "list") == 1) 
    or isip_die("ERROR: wsj_decode - output list not divided into N ways...exiting");

# move the testing lists from the mail decoding directory to the
# cpus directories
#
for( $i = 0; $i <= $#sort_cpus_test; $i++) {
    (isip_move("$output_path/lists/output_$i.list", "$output_list_cpus[$i]"))
	or isip_die("ERROR: wsj_decode - output lists for decoding not moved...exiting\n");
}

# build the output directories
#
print "INFO: wsj_decode - building the output directories\n";
(create_directories($output_list) == 1)
    or isip_die ("ERROR: could not create output directories");

# define the alignment mode
#
$align_mode = "word";

# define the number of history and n-best paths
#
$history_paths = 1;
$nbest_paths = 1;

# check to see if all needed outputs already exist
#
if (check_files_in_list($output_list) == 1) {
    $decoding_finished = 1;
}

#----------------------------------------------------------------------------
#
# run the decoder
#
#----------------------------------------------------------------------------

# at this point, all needed variables should be set, so just run the
# decoder. note that in perl, if a parameter is not defined, it is 
# interpreted as "" so it is safe to pass in variables that may not have
# been defined. Loop over all the cpus and start the deocoding job on each.
#
if ($decoding_finished == 0) {

    print "INFO: wsj_decode - starting decoder in $dec_function mode\n";
    
    for( $i = 0; $i <= $#cpus_test_list; $i++) {
	
	$param_file = "$output_path_cpus[$i]/params.text";
	$log_file = "$output_path_cpus[$i]/decode.log";
	$shell_script = "$output_path_cpus[$i]/run_trace_projector.sh";
	$completion_file[$i] = "$output_path_cpus[$i]/decode_job_completed.text";
	
	# create the shell script that would run the bw_train
	#
	run_create_shell_script_test($shell_script, $param_file, 
				     $TI_PATH_TO_CUR_DIR,
				     $log_file, $completion_file[$i], 
				     $cpus_test_list[$i]);
	
	(run_trace_projector($dec_function, $mfcc_list_cpus[$i], 
			     $output_list_cpus[$i], 
			     $input_lattice_list, $transcription_file,
			     $lattice_wer_file, $output_lattice_list, 
			     $grammar_list_file, $monophones_file, 
			     $models_file,
			     $phones_file, $lexicon_file, $states_file, 
			     $transitions_file, $align_mode, $lm_file,
			     $ngram_order, $lm_scale, $word_penalty,
			     $history_paths, $nbest_paths, $input_format,
			     $context_mode, $test_state_beam_pruning, 
			     $test_model_beam_pruning, $test_word_beam_pruning,
			     $mapmi_limit, $num_allowed_words, $param_file, 
			     $input_feature_format, $compute_delta, 
			     $compute_acc, $delta_win, $shell_script) 
	 == 1) 
	    or isip_die ("decoding procedure failed...exiting");
    }
}
else {
    print "INFO: wsj_decode - all output files already exist... skipping";
    print " decoding\n";
}

# wait for the jobs to finish only when output files don't exist
#
if ($decoding_finished == 0) {
    print "INFO: wsj_decode - waiting for the decoding jobs to complete\n";
    $done = 0;
    while ($done == 0) {
	$done = 1;
	sleep 1;	
	for ($i = 0; $i <= $#cpus_test_list; $i++) {
	    if ( ! -f $completion_file[$i]) {
		$done = 0;
	    }
	}
    }
}

# evaluate the hypotheses if desired
#
if ($evaluate_hyps == 1) {

    # build the mapped output list
    #
    $map_output_list = "$output_path/lists/map_output.list";

    print "INFO: wsj_decode - building the mapped output list for WSJ transcription\n";
    print "INFO: wsj_decode - $map_output_list\n";

    (build_output_list("$output_path/map_output", $TI_PATH_TO_TEST_LIST, 
		   $map_output_list) == 1)
	or isip_die ("ERROR: could not create $map_output_list");

    # build the mapped output directories
    #
    print "INFO: wsj_decode - building the mapped output directory\n";
    (create_directories($map_output_list) == 1)
	or isip_die ("ERROR: could not create mapped output directory\n");

    # map the hypotheses to WSJ transcription
    #
    (map_wsj($output_list, $map_output_list) == 1)
	or isip_die ("ERROR: could not map output hypotheses to WSJ transcription\n");
 
    $reference_file = "$TI_PATH_TO_DEC_DATA/reference.score";
    $hypothesis_file = "$output_path/hypothesis";
    print "INFO: wsj_decode - evaluating resultant hypotheses\n";
    (run_isip_eval("isip_word", $map_output_list, $reference_file, 
		   $hypothesis_file) == 1)
	or isip_die ("error evaluating the hypotheses...exiting");
}

#----------------------------------------------------------------------------
#
# finished
#
#----------------------------------------------------------------------------
print "INFO: wsj_decode - decoding complete\n";


# return no error
#
exit 0;
