#!/usr/local/bin/perl
#
# file: wsj_exp/util/wsj_scripts/wsj_preprocess.pl.in
#
$ISIP_HELP_FILE = <<__ISIP_HELP_HERE_FILE__;
name: wsj_preprocess
synopsis: 
   wsj_preprocess -train_mfc_list <list_of_training_feature_files> 
    -test_mfc_list <list_of_testing_feature_files> 
    -cpus_train <cpus_used_for_training>
    -cpus_test <cpus_used_for_testing>

descr: 
   this program is typically not meant to be run alone - rather it 
   should be called by an external script. it makes the basic directories 
   and transcriptions for the rest of the experiment

example: 
   wsj_preprocess -train_mfc_list train_list -test_mfc_list test_list
   -cpus_train isip105 isip206 isip206 -cpus_test isip204 isip205

options:     
   -train_mfc_list <list_of_training_feature_files> : 
      specifies the training feature files list

   -test_mfc_list <list_of_test_feature_files> :
      specifies the testing feature files list

   -machines <machines_to_run_experiment>
      specifies the machines on which the experiment is to be run

   -cpus_train <cpus_used_to_train>
      specifies the cpus to be employed for training

   -cpus_test <cpus_used_to_test>
      specifies the cpus to be employed for testing

arguments:  none

man page: none
__ISIP_HELP_HERE_FILE__

# define the path to the necessary information
#
$WSJ_PATH = "/isip/d206/research/aurora_front_end/asr_wsj_tutorial_v1.4";

# system packages
#
use File::Basename;
use File::Path;
use File::Copy;
use FileHandle;
autoflush STDOUT;

# isip libraries
#
require "$WSJ_PATH/lib/scripts/perl/command_line.pm";
require "$WSJ_PATH/lib/scripts/perl/wsj_subs.pm";

#----------------------------------------------------------------------------
#
# constant and global variable definitions
#
#----------------------------------------------------------------------------

# define the program name
#
@tmp = split(/\//, $0);
$ISIP_PROG = pop(@tmp);

# define the program to check for endian-ness
#
$ISIP_BINARY_NAME = $ENV{ISIP_BINARY};
$BIG_ENDIAN = "0";
$LITTLE_ENDIAN = "1";
$BIG_ENDIAN_CHECKER = "$WSJ_PATH/bin/$ISIP_BINARY_NAME/check_endian";

# define a separator for output comments
#
$TI_LINE_SEP = "-------------------  $ISIP_PROG  ----------------";

# define files to be used by this script
#
$TI_PATH_TO_EXP = "$WSJ_PATH/exp";
$TI_PATH_TO_OUTPUT = "$TI_PATH_TO_EXP/data_generation";
$TI_PATH_TO_OUTPUT_LISTS = "$TI_PATH_TO_OUTPUT/lists";
$TI_PATH_TO_TRAIN_LIST = "$TI_PATH_TO_OUTPUT_LISTS/train_mfc.list";
$TI_PATH_TO_TEST_LIST = "$TI_PATH_TO_OUTPUT_LISTS/test_mfc.list";
$TI_PATH_TO_CPUS_TRAIN = "$TI_PATH_TO_OUTPUT_LISTS/cpus_train.list";
$TI_PATH_TO_CPUS_TEST = "$TI_PATH_TO_OUTPUT_LISTS/cpus_test.list";
$TI_PATH_TO_OUTPUT_TRANS = "$TI_PATH_TO_OUTPUT/transcriptions";
$TI_PATH_TO_DATA_TRAIN = "$WSJ_PATH/data/train";
$TI_PATH_TO_DATA_TEST = "$WSJ_PATH/data/decode";
$TI_PATH_TO_TRAIN_MONOPHONE_BASE = "$TI_PATH_TO_DATA_TRAIN/monophone/base";

# define command-line options
#
$TI_OPT_TRAIN_MFC_LIST = "-train_mfc_list";
$TI_OPT_TEST_MFC_LIST = "-test_mfc_list";
$TI_OPT_CPUS_TRAIN = "-cpus_train";
$TI_OPT_CPUS_TEST = "-cpus_test";

#----------------------------------------------------------------------------
#
# command line processing
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "$TI_LINE_SEP\n";
print "INFO: wsj_preprocess - checking command-line arguments\n";

# parse the command line
#
my ($train_mfc_list, $test_mfc_list, $cpus_train, $cpus_test) =  
    command_line(0, $TI_OPT_TRAIN_MFC_LIST, 1, $TI_OPT_TEST_MFC_LIST, 1,
		 $TI_OPT_CPUS_TRAIN, -1, $TI_OPT_CPUS_TEST, -1);

if ($train_mfc_list eq "") {
    isip_die ("ERROR: train_mfc_list not specified");
}

if ($test_mfc_list eq "") {
    isip_die ("ERROR: test_mfc_list not specified");
}

if ($cpus_train eq "") {
    isip_die ("ERROR: cpus_train not specified");
}

if ($cpus_test eq "") {
    isip_die ("ERROR: cpus_test not specified");
}

# create the list of cpus that will be used to form the 
# directory structure 
#
@cpus_train_list = split(/\s+/, $cpus_train);
@cpus_test_list = split(/\s+/, $cpus_test);

# create sub-directory name for training cpus
#
$j = 0;
for ($i = 0; $i <= $#cpus_train_list; $i++) {
    
    if ($cpus_train_list[$i] ne $cpus_train_list[$i-1]) {
       $j = 0;
    }

    if (($cpus_train_list[$i] eq $cpus_train_list[$i-1]) && ($i != 0)) {
       $j++; 
    }

    $cpus_train_num[$i] = join("_", $cpus_train_list[$i], $j);
}

# create sub-directory name for testing cpus
#
$j = 0;
for ($i = 0; $i <= $#cpus_test_list; $i++) {
    
    if ($cpus_test_list[$i] ne $cpus_test_list[$i-1]) {
       $j = 0;
    }

    if (($cpus_test_list[$i] eq $cpus_test_list[$i-1]) && ($i != 0)) {
       $j++; 
    }

    $cpus_test_num[$i] = join("_", $cpus_test_list[$i], $j);
}

# sort the training and testing cpus
#
@sort_cpus_train = sort(@cpus_train_num);
@sort_cpus_test = sort(@cpus_test_num);

# find the total cpus for training and testing
#
$total_cpus_num = join(" ", @sort_cpus_train, @sort_cpus_test);
@cpus_list = split(/\s+/, $total_cpus_num);
 
# accumulate the unique cpus
#
for ($i = 0; $i <= $#cpus_list; $i++) {
    if (!defined($yes{$cpus_list[$i]})) {
	$yes{$cpus_list[$i]} = 1;
    }
}

# get the unique cpus
#
$i = 0;
foreach $unique_cpu (sort keys(%yes)) {
    $unique_cpus_num[$i] = $unique_cpu;
    $i++;
}

#----------------------------------------------------------------------------
#
# file and directory verification
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_preprocess - checking for input data directory and files\n";
print "INFO: checking for input data directory and files\n";

# verify the input data directory exists
#
if (! -d "$TI_PATH_TO_DATA_TRAIN") {
    isip_die ("ERROR: input data directory not found: $TI_PATH_TO_DATA_TRAIN\n\t. you may have an incomplete distribution...exiting");
}

if (! -d "$TI_PATH_TO_DATA_TEST") {
    isip_die ("ERROR: input data directory not found: $TI_PATH_TO_DATA_TEST\n\t. you may have an incomplete distribution...exiting");
}

# verify that the input mfcc files exist
#
(check_files_in_list($train_mfc_list) == 1) 
    or isip_die("ERROR: some input training files not found...exiting");

(check_files_in_list($test_mfc_list) == 1) 
    or isip_die("ERROR: some input test files not found...exiting");

# print a debugging message
#
print "INFO: wsj_preprocess - all training and testing mfc files exist\n";

# print a debugging message
#
print "INFO: wsj_preprocess - creating local data directories\n";

# create directory structure for the output lists and transcriptions
#
if (! -d "$TI_PATH_TO_EXP") {
    ((isip_mkdir($TI_PATH_TO_EXP) == 1) &&
     (isip_mkdir($TI_PATH_TO_OUTPUT) == 1) && 
     (isip_mkdir($TI_PATH_TO_OUTPUT_LISTS) == 1) &&
     (isip_mkdir($TI_PATH_TO_OUTPUT_TRANS) == 1))
	or isip_die ("ERROR: error creating directories...exiting");
}

# create directory structure for the output lists and transcriptions
# equal to the number of cpu's
#
for ($num = 0; $num <= $#unique_cpus_num; $num++) {

(isip_mkdir("$TI_PATH_TO_EXP/$unique_cpus_num[$num]") == 1) or
 isip_die ("ERROR: error creating directory $TI_PATH_TO_EXP/$unique_cpus_num[$num]...exiting");

(isip_mkdir("$TI_PATH_TO_EXP/$unique_cpus_num[$num]/data_generation") == 1) or
 isip_die ("ERROR: error creating directory $TI_PATH_TO_EXP/$unique_cpus_num[$num]/data_generation...exiting");

(isip_mkdir("$TI_PATH_TO_EXP/$unique_cpus_num[$num]/data_generation/lists") == 1) or
 isip_die ("ERROR: error creating directory $TI_PATH_TO_EXP/$unique_cpus_num[$num]/data_generation/lists...exiting");

(isip_mkdir("$TI_PATH_TO_EXP/$unique_cpus_num[$num]/data_generation/transcriptions") == 1)
 or isip_die ("ERROR: error creating directory $TI_PATH_TO_EXP/$unique_cpus_num[$num]/data_generation/transcriptions...exiting");
};

#----------------------------------------------------------------------------
#
# save the cpus_train and cpus_test to be used by other scripts like
# wsj_train.pl.in and wsj_test.pl.in
#
#----------------------------------------------------------------------------

# print a debugging message
#
#print "INFO: wsj_preprocess - copying the cpus_train and cpus_test\n";
#print "INFO: to be used by other scripts in this package\n";

# open the files to write
#
#(isip_open(fp_cpus_train, ">$TI_PATH_TO_CPUS_TRAIN") == 1)  
#    or isip_die ("ERROR: $TI_PATH_TO_CPUS_TRAIN not written...exiting");

#(isip_open(fp_cpus_test, ">$TI_PATH_TO_CPUS_TEST") == 1)  
#    or isip_die ("ERROR: $TI_PATH_TO_CPUS_TEST not written...exiting");


# copy the cpus_train and cpus_test
#
#for ($i = 0; $i <= $#sort_cpus_train; $i++) {
#print fp_cpus_train "$sort_cpus_train[$i]\n";
#}

#for ($i = 0; $i <= $#sort_cpus_test; $i++) {
#print fp_cpus_test "$sort_cpus_test[$i]\n";
#}

# close the files
#
#(isip_close(fp_cpus_train) == 1)  
#    or isip_die ("ERROR: $TI_PATH_TO_CPUS_TRAIN cannot be closed...exiting");

#(isip_close(fp_cpus_test) == 1)  
#    or isip_die ("ERROR: $TI_PATH_TO_CPUS_TEST cannot be closed...exiting");

#----------------------------------------------------------------------------
#
# extracting transcription form the master transcription for the training
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_preprocess - creating transcriptions\n";

# define the path to the base transcription files
#
$all_mono_trans_no_sp = "$TI_PATH_TO_TRAIN_MONOPHONE_BASE/all_mono_trans_no_sp.text"; 
$all_mono_trans_with_sp = "$TI_PATH_TO_TRAIN_MONOPHONE_BASE/all_mono_trans_with_sp.text"; 

# define the path to the transcription files to be created
#
$mono_trans_no_sp = "$TI_PATH_TO_OUTPUT_TRANS/mono_trans_no_sp.text"; 
$mono_trans_with_sp = "$TI_PATH_TO_OUTPUT_TRANS/mono_trans_with_sp.text"; 

# create the transcriptions corresponding to the mfcc list file
#
(select_transcriptions($train_mfc_list, $all_mono_trans_no_sp, $mono_trans_no_sp) == 1) 
    or isip_die("ERROR: wsj_preprocess - monophones transcriptions without sp not built...exiting");

(select_transcriptions($train_mfc_list, $all_mono_trans_with_sp, $mono_trans_with_sp) == 1) 
    or isip_die("ERROR: wsj_preprocess - monophones transcriptions with sp not built...exiting");

#----------------------------------------------------------------------------
#
# constructing lists for the training and decoding lists according to the 
# number of cpu's
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_preprocess - copying training and decoding lists\n";
print "INFO: wsj_preprocess - to exp directory\n";

# copying the training and testing list to experimental subdirectory
#
(isip_copy($train_mfc_list, $TI_PATH_TO_TRAIN_LIST))
    or isip_die("ERROR: wsj_preprocess - training mfc list not copied...exiting");

(isip_copy($test_mfc_list, $TI_PATH_TO_TEST_LIST))
    or isip_die("ERROR: wsj_preprocess - testing mfc list not copied...exiting");

# print a debugging message
#
print "INFO: wsj_preprocess - creating training and decoding lists\n";
print "INFO: wsj_preprocess - equal to the number of corresponding cpu's\n";

# divide the training and testing lists N ways to built the training lists
#
(create_N_lists(($#sort_cpus_train + 1), $TI_PATH_TO_TRAIN_LIST, 
		$TI_PATH_TO_OUTPUT_LISTS, "train_mfc", "list") == 1) 
    or isip_die("ERROR: wsj_preprocess - training mfc lists not built...exiting");

(create_N_lists(($#sort_cpus_test + 1), $TI_PATH_TO_TEST_LIST, 
		$TI_PATH_TO_OUTPUT_LISTS, "test_mfc", "list") == 1) 
    or isip_die("ERROR: wsj_preprocess - test mfc lists not built...exiting");

# print a debugging message
#
print "INFO: wsj_preprocess - dividing the monophone\n"; 
print "transcriptions equal to the number of corresponding cpus\n";

# divide the extracted transcriptions according to the number of cpu's
#
(create_N_lists(($#sort_cpus_train + 1), $mono_trans_no_sp, 
		$TI_PATH_TO_OUTPUT_TRANS, "mono_trans_no_sp", "text")  == 1) 
    or isip_die("ERROR: wsj_preprocess - monophones transcriptions without sp not divided into N ways...exiting");

(create_N_lists(($#sort_cpus_train + 1), $mono_trans_with_sp, 
		$TI_PATH_TO_OUTPUT_TRANS, "mono_trans_with_sp", "text")  == 1) 
    or isip_die("ERROR: wsj_preprocess - monophones transcriptions with sp not divided into N ways...exiting");

#----------------------------------------------------------------------------
#
# moving lists for the training and decoding mfc and transcriptions to 
# cpus directories
#
#----------------------------------------------------------------------------

print "INFO: wsj_preprocess - moving traning mfc, testing mfcc, and\n"; 
print "INFO: transcriptions to respective subdirectories corresponding\n";
print "INFO: to the cpus..\n";

for( $i = 0; $i <= $#sort_cpus_train; $i++) {
    (isip_move("$TI_PATH_TO_OUTPUT_LISTS/train_mfc_$i.list", "$TI_PATH_TO_EXP/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list")) 
     or isip_die("ERROR: wsj_preprocess - training mfc sublists not moved...exiting");
}

    
for( $i = 0; $i <= $#sort_cpus_test; $i++) {
    (isip_move("$TI_PATH_TO_OUTPUT_LISTS/test_mfc_$i.list", "$TI_PATH_TO_EXP/$sort_cpus_test[$i]/data_generation/lists/test_mfc.list"))
     or isip_die("ERROR: wsj_preprocess - test mfc sublists not moved...exiting");
}
	
for( $i = 0; $i <= $#sort_cpus_train; $i++) {
    (isip_move("$TI_PATH_TO_OUTPUT_TRANS/mono_trans_no_sp_$i.text", "$TI_PATH_TO_EXP/$sort_cpus_train[$i]/data_generation/transcriptions/mono_trans_no_sp.text"))
     or isip_die("ERROR: wsj_preprocess - monophone sub-transcriptions without sp  not moved...exiting");
}

for( $i = 0; $i <= $#sort_cpus_train; $i++) {
(isip_move("$TI_PATH_TO_OUTPUT_TRANS/mono_trans_with_sp_$i.text", "$TI_PATH_TO_EXP/$sort_cpus_train[$i]/data_generation/transcriptions/mono_trans_with_sp.text"))
     or isip_die("ERROR: wsj_preprocess - monophone transcriptions with sp  not moved...exiting");
}

#----------------------------------------------------------------------------
#
# checking endianness
#
#----------------------------------------------------------------------------

print "INFO: wsj_preprocess - checking endian-ness of your system...\n";

# check for endian-ness. note that the input data is big-endian.
#
$swap_value = `$BIG_ENDIAN_CHECKER`;
chop $swap_value;
$swap_value =~ s/^\s+//g;
$swap_value =~ s/\s+//g;

if ($swap_value eq $LITTLE_ENDIAN) {
    $swap_flag = "1";
    print "INFO: wsj_preprocess - found a little endian architecture\n";
    print "INFO: wsj_preprocess - Please check that the mfcc's are in little endian format\n";
##    print "INFO: wsj_preprocess - adding swap_bytes flag to command line\n";
}
else {
    $swap_flag = "";
    print "INFO: wsj_preprocess - found a big endian architecture\n";
    print "INFO: wsj_preprocess - Please check that the mfcc's are in big endian format\n";
}
exit 0;
