#!/usr/bin/perl
#
# file: wsj_exp/wsj_strip.pl
#
$ISIP_HELP_FILE = <<__ISIP_HELP_HERE_FILE__;
name: wsj_strip
synopsis: 
    wsj_strip -utterance_timing_list <file_containing_timing_information>
    -input_mfc_list <input_mfc_file_list_from_which_silence_to_be_stripped> 
    -output_path <output_path_for_stripped_feature_files> 
    -extension <file_extension_for_stripped_feature_files>
    -feature_format <format> -dimension <dimension> 
    -frame_duration <duration_of_frame_in_secs>

descr: This script strips the silences from the 13 dimensional feature
    files. The user needs to provide the utterances timing information
    and the path to the output stripped feature files.

example:
 
    wsj_strip -utterance_timing_list ./eval_0330.text 
    -input_mfc_list eval01_0166_v1.4.1.list 
    -output_path /isip/data/corpora/wsj0/13mfc/etsi_fe/test_ep/test_clean_wv1
    -extension endpt -format htk -dimension 13 -frame_duration 0.01
 
options:

   -utterance_timing_list <file_containing_timing_information>
      specifies the file that contains the start and end timings of the 
      utterances

    -input_mfc_list <input_mfc_file_list_from_which_silence_to_be_stripped> 
      specifies the input file list that will be stripped
  
    -output_path <output_path_for_stripped_feature_files>
      specifies the path for the stripped feature files (default is the same 
      path as the input mfc files)
   
    -extension <file_extension_for_stripped_feature_files>
      specifies file extension for the stripped feature files (default is 
      .endpt)

    -feature_format <format> : 
      specifies the feature format to use. Choices are:
      htk        : HTK format features (default)
      isip_proto : ISIP format features

   -dimension <dimension_of_input_mfc> : 
      specifies the dimension of the input features (default is 13)

   -frame_duration <duration> : 
      specifies the duration of frame in seconds (default is 0.01)
   
arguments: none

man page: none
__ISIP_HELP_HERE_FILE__

# define the path to this script
#
$WSJ_PATH = "/cavs/hse/ies/isip/users/raghavan/aurora/cluster_enabled/asr_wsj_tutorial_v1.4.2";

# system packages
#
use File::Basename;
use File::Path;
use File::Copy;
use FileHandle;
autoflush STDOUT;

# isip libraries
#
require "$WSJ_PATH/lib/scripts/perl/command_line.pm";
require "$WSJ_PATH/lib/scripts/perl/wsj_subs.pm";

#----------------------------------------------------------------------------
#
# constant and global variable definitions
#
#----------------------------------------------------------------------------

# define the program name
#
@tmp = split(/\//, $0);
$ISIP_PROG = pop(@tmp);

# define a separator for output comments
#
$TI_LINE_SEP = "-------------------  $ISIP_PROG  ----------------";

# get the current working directory
#
$TI_PATH_TO_CUR_DIR = `pwd`;
chop($TI_PATH_TO_CUR_DIR);

# define files and locations to be used by this script
#
$TI_PATH_TO_DATA = "$WSJ_PATH/data";
$TI_PATH_TO_OUTPUT = "$TI_PATH_TO_MFC/lists";

# define command-line options
#
$TI_OPT_UTTERANCE_TIMING_LIST = "-utterance_timing_list";
$TI_OPT_INPUT_MFC_LIST = "-input_mfc_list";
$TI_OPT_OUTPUT_PATH = "-output_path";
$TI_OPT_EXTENSION = "-extension";
$TI_OPT_FEAT_FORMAT = "-feature_format";
$TI_OPT_DIMENSION = "-dimension";
$TI_OPT_FRAME_DURATION = "-frame_duration";

$TI_FEAT_FORMAT_HTK = "htk";
$TI_FEAT_FORMAT_ISIP = "isip_proto";
@TI_FEAT_FORMAT_TYPES = ($TI_FEAT_FORMAT_HTK, $TI_FEAT_FORMAT_ISIP);


$TI_EXTENSION_DEF = "endpt";
$TI_FEAT_FORMAT_DEF = $TI_FEAT_FORMAT_HTK;
$TI_DIMENSION_DEF = "13";
$TI_FRAME_DURATION_DEF = "0.01";

#----------------------------------------------------------------------------
#
# command line processing
#
#----------------------------------------------------------------------------
# print a debugging message
#
print "$TI_LINE_SEP\n";
print "INFO: wsj_strip - checking command-line arguments\n\n";

# parse the command line
#
my ( $utterance_timing_list, $input_mfc_list, $output_path, $extension,
     $feature_format, $dimension, $frame_duration) 
    =  command_line(0, $TI_OPT_UTTERANCE_TIMING_LIST, 1, 
		    $TI_OPT_INPUT_MFC_LIST, 1, $TI_OPT_OUTPUT_PATH, 1, 
		    $TI_OPT_EXTENSION, 1, $TI_OPT_FEAT_FORMAT, 1, 
		    $TI_OPT_DIMENSION, 1, $TI_OPT_FRAME_DURATION, 1);

if ($extension eq "") {
    $extension = $TI_EXTENSION_DEF;
}
if ($feature_format eq "") {
    $feature_format = $TI_FEAT_FORMAT_DEF;
}
if ($dimension eq "") {
    $dimension = $TI_DIMENSION_DEF;
}
if ($frame_duration eq "") {
    $frame_duration = $TI_FRAME_DURATION_DEF;
}

if ($output_path eq "") {

# print a debugging message
#
print "INFO: wsj_strip - path to output stripped files not specified\n";
print "using the path from the input feature file list\n";
}

#----------------------------------------------------------------------------
#
# input file verification
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_strip - checking for input files\n";

# verify that the input mfcc files exist
#
(check_files_in_list($input_mfc_list) == 1) 
    or isip_die("ERROR: some input feature files not found...exiting");

# print a debugging message
#
print "INFO: wsj_strip - all input feature files exist\n";

#----------------------------------------------------------------------------
#
# endpointing the data by stripping the silences at the begining and the end
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_strip - endpointing the input feature files\n";

(strip_silence($utterance_timing_list, $input_mfc_list, $output_path, 
	       $extension, $feature_format, $dimension, $frame_duration) == 1)
    or return 0;

print "INFO: wsj_strip - Completed Endpointing\n";
