#!/usr/bin/perl
#
# file: wsj_exp/util/wsj_scripts/wsj_train.pl.in
#
$ISIP_HELP_FILE = <<__ISIP_HELP_HERE_FILE__;
name: wsj_train
synopsis: 
    wsj_train -num_features <number> -mode <mode> -feature_format <format> 
    -mixtures <num_of_mixtures_for_xwrd_models>
    -train_beam_pruning <beam_pruning_for_training>
    -split_threshold <threshold_to_split_a_node_in_a_decision_tree>
    -merge_threshold <threshold_to_merge_two_nodes_in_a_decision_tree>
    -num_occ_threshold <min_num_of_occupancy_to_split_a_node_in_a_decision_tree>   -train_mfc_list <list_of_training_feature_files> 
    -cpus_train <cpus_used_for_training>
    -cluster_train <cluster_used_for_training>
    -num_nodes <number_of_cluster_nodes>
    -walltime <expected_walltime_on_the_cluster>
    
descr: 
   this program builds 1, 4, 8, or 16 mixture crossword context-dependent HMM 
   acoustic models

example: 
   wsj_train -num_features 13 -mode acc -feature_format htk -mixtures 1 
   -train_mfc_list train.list

options:

   -num_features <number> : 
      number of features per frame in each of the input data files
      (default is 39)

   -mode <mode> : 
      this option tells the script whether you want to compute deltas or both 
      delta/accelerations or neither on the fly. Choices are:
      delta    : computes deltas on the fly
      acc      : computes accelerations on the fly
      none     : does not computes deltas or accelerations on the fly (default)

   -feature_format <format> : 
      specifies the feature format to use. Choices are:
      htk        : HTK format features (default)
      isip_proto : ISIP format features

   -mixtures <num_of_mixtures_for_xwrd_models> : 
      specifies the number of cross-word mixtures to train and use for 
      decoding. Choices are:
       1 (default)
       2
       4
       8
      16

   -train_beam_pruning <beam_pruning_for_training> : 
      specifies the beam pruning for all the three levels during training
      (default is 1000)

   -split_threshold <threshold_to_split_a_node_in_decision_tree> : 
      specifies the minimum increase in likelihood to split a node in the 
      decision tree during the state tying process (default is 165)

   -merge_threshold <threshold_to_merge_two_node_in_decision_tree> : 
      specifies the maximum decrease in likelihood to merge two nodes in the
      decision tree during the state tying process (default is 165)

   -num_occ_threshold <min_num_of_occupancy_to_split_a_node_in_decision_tree> :
      specifies the minimum number of states for splitting a node in the 
      decision tree during the state tying process (default is 840)

   -train_mfc_list <list_of_training_feature_files> : 
      specifies the training feature files list

   -cpus_train <cpus_used_to_train>
      specifies the cpus to be employed for training
    
   -cluster_train <cluster_used_for_training>
       specifies the cluster to be employed for training

   -num_nodes <number_of_cluster_nodes>
    specifies the number of nodes required to run jobs

   -walltime <expected_walltime_on_the_cluster>
       specifies the walltime expected for job completion

man page: none
__ISIP_HELP_HERE_FILE__

# define the path to the necessary information
#
$WSJ_PATH = "/cavs/hse/ies/isip/users/raghavan/aurora/cluster_enabled/asr_wsj_tutorial_v1.4.2";

# system packages
#
use File::Basename;
use File::Path;
use File::Copy;
use FileHandle;
autoflush STDOUT;

# isip libraries
#
require "$WSJ_PATH/lib/scripts/perl/command_line.pm";
require "$WSJ_PATH/lib/scripts/perl/wsj_subs.pm";

#----------------------------------------------------------------------------
#
# constant and global variable definitions
#
#----------------------------------------------------------------------------

# define the program name
#
@tmp = split(/\//, $0);
$ISIP_PROG = pop(@tmp);

# define the program to check for endian-ness
#
$ISIP_BINARY_NAME = $ENV{ISIP_BINARY};
$BIG_ENDIAN = "0";
$LITTLE_ENDIAN = "1";
$BIG_ENDIAN_CHECKER = "$WSJ_PATH/bin/$ISIP_BINARY_NAME/check_endian";

# define a separator for output comments
#
$TI_LINE_SEP = "-------------------  $ISIP_PROG  ----------------";

# get the current working directory
#
$TI_PATH_TO_CUR_DIR = `pwd`;
chop($TI_PATH_TO_CUR_DIR);

# define files and locations to be used by this script
#
$TI_PATH_TO_TRAIN = "$WSJ_PATH/data/train";
$TI_PATH_TO_OUTPUT = "$TI_PATH_TO_CUR_DIR/train";
$TI_PATH_TO_BW = "$TI_PATH_TO_OUTPUT/baum_welch";
$TI_PATH_TO_MFC = "$TI_PATH_TO_CUR_DIR/data_generation";

$TI_PATH_TO_OUTPUT_LISTS = "$TI_PATH_TO_MFC/lists";
$TI_PATH_TO_TRAIN_LIST = "$TI_PATH_TO_OUTPUT_LISTS/train_mfc.list";
$TI_PATH_TO_CPUS_TRAIN = "$TI_PATH_TO_OUTPUT_LISTS/cpus_train.list";
$TI_PATH_TO_OUTPUT_TRANS = "$TI_PATH_TO_MFC/transcriptions";
$TI_PATH_TO_TRAIN_MONOPHONE_BASE = "$TI_PATH_TO_TRAIN/monophone/base";


# define command-line options
#
$TI_OPT_NUM_FEATURES = "-num_features";
$TI_OPT_MODE = "-mode";
$TI_OPT_FEAT_FORMAT = "-feature_format";
$TI_OPT_NUM_MIX = "-mixtures";
$TI_OPT_TR_BEAM_PRUN = "-train_beam_pruning";
$TI_OPT_SP_THRESH = "-split_threshold";
$TI_OPT_ME_THRESH = "-merge_threshold";
$TI_OPT_NUM_OCC_THRESH = "-num_occ_threshold";
$TI_OPT_TRAIN_MFC_LIST = "-train_mfc_list";
$TI_OPT_CPUS_TRAIN = "-cpus_train";
$TI_OPT_CLUSTER_TRAIN = "-cluster_train";
$TI_OPT_NUM_NODES = "-num_nodes";
$TI_OPT_WALLTIME = "-walltime";

$TI_MODE_DELTA = "delta";
$TI_MODE_ACC = "acc";
$TI_MODE_NONE = "none";
@TI_MODE_TYPES = ($TI_MODE_DELTA, $TI_MODE_ACC, $TI_MODE_NONE);

$TI_FEAT_FORMAT_HTK = "htk";
$TI_FEAT_FORMAT_ISIP = "isip_proto";
@TI_FEAT_FORMAT_TYPES = ($TI_FEAT_FORMAT_HTK, $TI_FEAT_FORMAT_ISIP);

$TI_NUM_MIX_1 = "1";
$TI_NUM_MIX_2 = "2";
$TI_NUM_MIX_4 = "4";
$TI_NUM_MIX_8 = "8";
$TI_NUM_MIX_16 = "16";
@TI_NUM_MIX_TYPES = ($TI_NUM_MIX_1, $TI_NUM_MIX_2, $TI_NUM_MIX_4, 
		     $TI_NUM_MIX_8, $TI_NUM_MIX_16);

# define defaults
#
$TI_NUM_FEATURES_DEF = "39";
$TI_MODE_DEF = $TI_MODE_NONE;
$TI_FEAT_FORMAT_DEF = $TI_FEAT_FORMAT_HTK;
$TI_NUM_MIX_DEF = $TI_NUM_MIX_1;
$TI_TR_BEAM_PRUN_DEF = "1000";
$TI_SP_THRESH_DEF = "165";
$TI_ME_THRESH_DEF = "165";
$TI_NUM_OCC_THRESH_DEF = "840";

#----------------------------------------------------------------------------
#
# command line processing
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "$TI_LINE_SEP\n";
print "INFO: wsj_train - checking command-line arguments\n";

# parse the command line
#
my ($num_features, $mode, $feature_format, $mixtures, $train_beam_pruning, 
    $split_threshold, $merge_threshold, $num_occ_threshold, $train_mfc_list,
    $cpus_train, $cluster_train, $num_nodes, $walltime) =  
    command_line(0, $TI_OPT_NUM_FEATURES, 1, $TI_OPT_MODE, 1,
		 $TI_OPT_FEAT_FORMAT, 1, $TI_OPT_NUM_MIX, 1, 
		 $TI_OPT_TR_BEAM_PRUN, 1,
		 $TI_OPT_SP_THRESH, 1, $TI_OPT_ME_THRESH, 1,
		 $TI_OPT_NUM_OCC_THRESH, 1, $TI_OPT_TRAIN_MFC_LIST, 1,
		 $TI_OPT_CPUS_TRAIN, -1,
		 $TI_OPT_CLUSTER_TRAIN, 1,
		 $TI_OPT_NUM_NODES, 1,
		 $TI_OPT_WALLTIME, 1);

if ($num_features < 1) {
    isip_die ("ERROR: num_features cannot be less than 1\n");
}

if ($num_features eq "") {
    $num_features = $TI_NUM_FEATURES_DEF;
}

if ($mode eq "") {
    $mode = $TI_MODE_DEF;
}

if ($feature_format eq "") {
    $feature_format = $TI_FEAT_FORMAT_DEF;
}

if ($mixtures eq "") {
    $mixtures = $TI_NUM_MIX_DEF;
}

if ($train_beam_pruning eq "") {
    $train_beam_pruning = $TI_TR_BEAM_PRUN_DEF;
}

if ($split_threshold eq "") {
    $split_threshold = $TI_SP_THRESH_DEF;
}

if ($merge_threshold eq "") {
    $merge_threshold = $TI_ME_THRESH_DEF;
}

if ($num_occ_threshold eq "") {
    $num_occ_threshold = $TI_NUM_OCC_THRESH_DEF;
}

if ($cpus_train eq "" && $cluster_train eq "") {
    isip_die ("ERROR: cpus_train or cluster_train not specified");
}

# complete the path of the training list if absolute path is not given
# from the commandline
#
#if (substr($train_mfc_list, 0, 1) ne "/") {
#    $train_mfc_list = "$WSJ_PATH/$train_mfc_list";
#}

# create the list of cpus used for training
#
if ($cpus_train ne "") {
    @cpus_train_list = split(/\s+/, $cpus_train);
}
if ($cluster_train ne "") {
    for ($i = 0; $i <= $num_nodes; $i++) {
	$cpus_train_list[$i] = join("_", $cluster_train, $i);
    }
}

# create sub-directory name for training cpus
#
$j = 0;
for ($i = 0; $i <= $#cpus_train_list; $i++) {
    
    if ($cpus_train_list[$i] ne $cpus_train_list[$i-1]) {
	$j = 0;
    }
    
    if (($cpus_train_list[$i] eq $cpus_train_list[$i-1]) && ($i != 0)) {
	$j++; 
    }

    $cpus_train_num[$i] = join("_", $cpus_train_list[$i], $j);
}

# don't sort the training cpus
#
@sort_cpus_train = @cpus_train_num;

#----------------------------------------------------------------------------
#
# file and directory verification
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_train - checking for input data directory and files\n";
print "INFO: checking for input data directory and files\n";

# verify the input data directory exists
#
if (! -d "$TI_PATH_TO_TRAIN") {
    isip_die ("ERROR: input data directory not found: $TI_PATH_TO_TRAIN\n\t. you may have an incomplete distribution...exiting");
}

# verify that the input mfcc files exist
#
(check_files_in_list($train_mfc_list) == 1) 
    or isip_die("ERROR: some input training files not found...exiting");

# print a debugging message
#
print "INFO: wsj_train - all training and testing mfc files exist\n";

# print a debugging message
#
print "INFO: wsj_train - creating local data directories\n";

# create directory structure for the output lists and transcriptions
#
if (! -d "$TI_PATH_TO_OUTPUT") {
    ((isip_mkdir($TI_PATH_TO_MFC) == 1) && 
     (isip_mkdir($TI_PATH_TO_OUTPUT_LISTS) == 1) &&
     (isip_mkdir($TI_PATH_TO_OUTPUT_TRANS) == 1))
	or isip_die ("ERROR: error creating directories...exiting");
}

# create directory structure for the output lists and transcriptions
# equal to the number of cpu's
#
for ($num = 0; $num <= $#sort_cpus_train; $num++) {
    if (! -d "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]") {
	(isip_mkdir("$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]") == 1) or
	    isip_die ("ERROR: error creating directory $TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]...exiting");
	
	(isip_mkdir("$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]/data_generation") 
	 == 1) or
	     isip_die ("ERROR: error creating directory $TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]/data_generation...exiting");
	
	(isip_mkdir("$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]/data_generation/lists") == 1) or
	    isip_die ("ERROR: error creating directory $TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]/data_generation/lists...exiting");

	(isip_mkdir("$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]/data_generation/transcriptions") == 1)
	    or isip_die ("ERROR: error creating directory $TI_PATH_TO_CUR_DIR/$sort_cpus_train[$num]/data_generation/transcriptions...exiting");
    }
}

#----------------------------------------------------------------------------
#
# extracting transcription form the master transcription for the training
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_train - creating transcriptions\n";

# define the path to the base transcription files
#
$all_mono_trans_no_sp = "$TI_PATH_TO_TRAIN_MONOPHONE_BASE/all_mono_trans_no_sp.text"; 
$all_mono_trans_with_sp = "$TI_PATH_TO_TRAIN_MONOPHONE_BASE/all_mono_trans_with_sp.text"; 

# define the path to the transcription files to be created
#
$mono_trans_no_sp = "$TI_PATH_TO_OUTPUT_TRANS/mono_trans_no_sp.text"; 
$mono_trans_with_sp = "$TI_PATH_TO_OUTPUT_TRANS/mono_trans_with_sp.text"; 

# create the transcriptions corresponding to the mfcc list file
#
(select_transcriptions($train_mfc_list, $all_mono_trans_no_sp, $mono_trans_no_sp) == 1) 
    or isip_die("ERROR: wsj_train - monophones transcriptions without sp not built...exiting");

(select_transcriptions($train_mfc_list, $all_mono_trans_with_sp, $mono_trans_with_sp) == 1) 
    or isip_die("ERROR: wsj_train - monophones transcriptions with sp not built...exiting");

#----------------------------------------------------------------------------
#
# constructing lists for the training lists according to the number of cpu's
#
#----------------------------------------------------------------------------

# print a debugging message
#
print "INFO: wsj_train - copying training and decoding lists\n";
print "INFO: wsj_train - to exp directory\n";

# copying the training list to experimental subdirectory
#
(isip_copy($train_mfc_list, $TI_PATH_TO_TRAIN_LIST))
    or isip_die("ERROR: wsj_train - training mfc list not copied...exiting");

# print a debugging message
#
print "INFO: wsj_train - creating training lists\n";
print "INFO: wsj_train - equal to the number of corresponding cpu's\n";

# divide the training lists N ways to built the training lists
#
(create_N_lists(($#sort_cpus_train + 1), $TI_PATH_TO_TRAIN_LIST, 
		$TI_PATH_TO_OUTPUT_LISTS, "train_mfc", "list") == 1) 
    or isip_die("ERROR: wsj_train - training mfc lists not built...exiting");

# print a debugging message
#
print "INFO: wsj_train - dividing the monophone\n"; 
print "transcriptions equal to the number of corresponding cpus\n";

# divide the extracted transcriptions according to the number of cpu's
#
(create_N_lists(($#sort_cpus_train + 1), $mono_trans_no_sp, 
		$TI_PATH_TO_OUTPUT_TRANS, "mono_trans_no_sp", "text")  == 1) 
    or isip_die("ERROR: wsj_train - monophones transcriptions without sp not divided into N ways...exiting");

(create_N_lists(($#sort_cpus_train + 1), $mono_trans_with_sp, 
		$TI_PATH_TO_OUTPUT_TRANS, "mono_trans_with_sp", "text")  == 1) 
    or isip_die("ERROR: wsj_train - monophones transcriptions with sp not divided into N ways...exiting");

#----------------------------------------------------------------------------
#
# moving lists for the training mfc lists and transcriptions to 
# cpus directories
#
#----------------------------------------------------------------------------

print "INFO: wsj_train - moving traning mfc and transcriptions to\n";
print "INFO: respective subdirectories corresponding to the cpus..\n";

for( $i = 0; $i <= $#sort_cpus_train; $i++) {
    (isip_move("$TI_PATH_TO_OUTPUT_LISTS/train_mfc_$i.list", "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list")) 
     or isip_die("ERROR: wsj_train - training mfc sublists not moved...exiting");
}

for( $i = 0; $i <= $#sort_cpus_train; $i++) {
    (isip_move("$TI_PATH_TO_OUTPUT_TRANS/mono_trans_no_sp_$i.text", "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/transcriptions/mono_trans_no_sp.text"))
	or isip_die("ERROR: wsj_train - monophone sub-transcriptions without sp  not moved...exiting");
}

for( $i = 0; $i <= $#sort_cpus_train; $i++) {
(isip_move("$TI_PATH_TO_OUTPUT_TRANS/mono_trans_with_sp_$i.text", "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/transcriptions/mono_trans_with_sp.text"))
     or isip_die("ERROR: wsj_train - monophone transcriptions with sp  not moved...exiting");
}

#----------------------------------------------------------------------------
#
# checking endianness
#
#----------------------------------------------------------------------------

print "INFO: wsj_train - checking endian-ness of your system...\n";

# check for endian-ness. note that the input data is big-endian.
#
$swap_value = `$BIG_ENDIAN_CHECKER`;
chop $swap_value;
$swap_value =~ s/^\s+//g;
$swap_value =~ s/\s+//g;

if ($swap_value eq $LITTLE_ENDIAN) {
    $swap_flag = "1";
    print "INFO: wsj_train - found a little endian architecture\n";
    print "INFO: wsj_train - Please check that the mfcc's are in little endian format\n";
}
else {
    $swap_flag = "";
    print "INFO: wsj_train - found a big endian architecture\n";
    print "INFO: wsj_train - Please check that the mfcc's are in big endian format\n";
}

#----------------------------------------------------------------------------
#
# build the commandline to be displayed
#
#----------------------------------------------------------------------------

$out_line = join (" ", $TI_OPT_NUM_FEATURES, $num_features, $TI_OPT_MODE,
                  $mode, $TI_OPT_FEAT_FORMAT, $feature_format,
		  $TI_OPT_NUM_MIX, $mixtures,
                  $TI_OPT_TR_BEAM_PRUN, $train_beam_pruning, 
                  $TI_OPT_SP_THRESH, $split_threshold, 
                  $TI_OPT_ME_THRESH, $merge_threshold, 
                  $TI_OPT_NUM_OCC_THRESH, $num_occ_threshold, 
		  $TI_OPT_TRAIN_MFC_LIST, $TI_PATH_TO_TRAIN_LIST, 
		  $TI_OPT_CPUS_TRAIN, $cpus_train);

print "INFO: wsj_train - running with parameters: $out_line\n";

#----------------------------------------------------------------------------
#
# build monophones and then crossword triphones
#
#----------------------------------------------------------------------------

(bw_build_base_mono_models(@cpus_train_list) == 1) or
    isip_die ("ERROR: failed to build base monophone models...exiting");

(bw_build_xwrd_models("$mixtures", @cpus_train_list) == 1) or
    isip_die ("ERROR: failed to build crossword triphone models...exiting");

# exit gracefully (note that utilities return 0 to indicate no error
#
exit 0;


# subroutine: bw_build_base_mono_models
#
# arguments: sort_cpus_train
#
# description: this routine builds a set base monophone models from scratch
#
# return: 0 if error, 1 if no error
#
sub bw_build_base_mono_models(\@) {

    # get parameters from arg list
    #
    my @cpus_train_list = @_;
    
    # create sub-directory name for training cpus
    #
    $j = 0;
    for ($i = 0; $i <= $#cpus_train_list; $i++) {
	
	if ($cpus_train_list[$i] ne $cpus_train_list[$i-1]) {
	    $j = 0;
	}
	
	if (($cpus_train_list[$i] eq $cpus_train_list[$i-1]) && ($i != 0)) {
	    $j++; 
	}
	
	$cpus_train_num[$i] = join("_", $cpus_train_list[$i], $j);
    }
    
    # don't sort the training cpus
    #
    @sort_cpus_train = @cpus_train_num;
    
    # a path to the data directory
    #
    my $TI_PATH_TO_TRAIN_DATA = "$TI_PATH_TO_TRAIN/monophone/base";
    
    # a path to the working main directory
    #
    my $TI_PATH_TO_BASE_MONO = "$TI_PATH_TO_BW/monophone/base";
    
    # a path to the working directories for training jobs on cpus
    #
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
	$TI_PATH_TO_BASE_MONO_CPUS[$i] 
	    = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/train/baum_welch/monophone/base";
    }

    # build directories as needed
    #
    print "INFO: bw_build_base_mono_models - building necessary paths and";
    print " files...\n";
    
    # check the base directory
    #
    if ( ! -d "$TI_PATH_TO_BASE_MONO") {
	(isip_mkdir($TI_PATH_TO_BASE_MONO) == 1) or return 0;
    }
    
    # check the base directories for training jobs on cpus 
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
        if ( ! "$TI_PATH_TO_BASE_MONO_CPUS[$i]") { 
            (isip_mkdir($TI_PATH_TO_BASE_MONO_CPUS[$i]) == 1) or return 0;
	}
    }
    

    # a path to the final models directory
    #
    my $TI_PATH_TO_FINAL_MODELS = "$TI_PATH_TO_BASE_MONO/final_models";

    # the files that must be present for this function to be complete
    #
    my @TI_BASE_MONO_MODEL_DEPENDENCIES = 
	("$TI_PATH_TO_FINAL_MODELS/monophones.text",
	 "$TI_PATH_TO_FINAL_MODELS/models.text",
	 "$TI_PATH_TO_FINAL_MODELS/phones.text",
	 "$TI_PATH_TO_FINAL_MODELS/lexicon.text",
	 "$TI_PATH_TO_FINAL_MODELS/vfloor.text",
	 "$TI_PATH_TO_FINAL_MODELS/states.bin",
	 "$TI_PATH_TO_FINAL_MODELS/transitions.text");

    # check that we don't already have these models
    #
    $found_all_files = 1;
    foreach $file (@TI_BASE_MONO_MODEL_DEPENDENCIES) {
	(file_exists($file) == 1)  or $found_all_files = 0;
    }
    
    if ($found_all_files == 1) {
	print "INFO: bw_build_base_mono_models - pre-existing base monophone";
	print " models found...skipping\n";
	return 1;
    }
    
    # at least one of the necessary files does not exist, so regenerate
    # the models from scratch.
    #
    # verify that the acoustic feature files exist
    #
    (check_files_in_list($TI_PATH_TO_TRAIN_LIST) == 1) or return 0;

    # sub-directories in main-base to create for this experiment
    #
    my @TI_BASE_MONO_MODEL_DIRS = 
	("$TI_PATH_TO_BASE_MONO/transcriptions",
	 "$TI_PATH_TO_BASE_MONO/alignments",	
	 "$TI_PATH_TO_FINAL_MODELS",
	 "$TI_PATH_TO_BASE_MONO/r00/hmm0",
	 "$TI_PATH_TO_BASE_MONO/r00/hmm1",
	 "$TI_PATH_TO_BASE_MONO/r00/hmm2",
 	 "$TI_PATH_TO_BASE_MONO/r00/hmm3",
	 "$TI_PATH_TO_BASE_MONO/r00/hmm4",
	 "$TI_PATH_TO_BASE_MONO/r01/hmm0",
	 "$TI_PATH_TO_BASE_MONO/r01/hmm1",
	 "$TI_PATH_TO_BASE_MONO/r01/hmm2",
 	 "$TI_PATH_TO_BASE_MONO/r01/hmm3",
	 "$TI_PATH_TO_BASE_MONO/r01/hmm4",
	 "$TI_PATH_TO_BASE_MONO/r01/hmm5",
	 "$TI_PATH_TO_BASE_MONO/r01/hmm6",
 	 "$TI_PATH_TO_BASE_MONO/r01/hmm7",
	 "$TI_PATH_TO_BASE_MONO/r01/hmm8",
 	 "$TI_PATH_TO_BASE_MONO/r01/hmm9");

    # create all other necessary directories
    #
    foreach $dir (@TI_BASE_MONO_MODEL_DIRS) {
	if ( ! -d "$dir") {
	    (isip_mkdir($dir) == 1) or
		isip_die("ERROR: couldn't create $dir\n");
        }
    }
    
    # sub-directories in base directories for training jobs on cpus
    # to create for this experiment
    #
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
	@TI_BASE_MONO_MODEL_DIRS_CPUS = 
	    ("$TI_PATH_TO_BASE_MONO_CPUS[$i]/transcriptions",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/alignments",	
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/hmm0",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/hmm1",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/hmm2",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/hmm3",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/hmm4",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/acc",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm0",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm1",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm2",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm3",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm4",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm5",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm6",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm7",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm8",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm9",
	     "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/acc");
	
	# create all other necessary directories
	#
	foreach $dir (@TI_BASE_MONO_MODEL_DIRS_CPUS) {
	    if ( ! -d "$dir") {
		(isip_mkdir($dir) == 1) or 
		isip_die("ERROR: couldn't create $dir\n");
	    }
	}
    }	
        
    # create the accumulator list that will be used for combining the 
    # accumulators
    #
    $r00_acc_list = "$TI_PATH_TO_BASE_MONO/r00/acc.list"; 
    (isip_open(fp_r00_acc_list, ">$r00_acc_list") == 1) or 
	isip_die("ERROR: couldn't open $r00_acc_list\n"); 

    $r01_acc_list = "$TI_PATH_TO_BASE_MONO/r01/acc.list"; 
    (isip_open(fp_r01_acc_list, ">$r01_acc_list") == 1) or 
	isip_die("ERROR: couldn't open  $r01_acc_list\n"); 
    
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
	print fp_r00_acc_list 
	    "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/acc/acc.bin\n";	
	print fp_r01_acc_list 
	    "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/acc/acc.bin\n";
    }
    
    (isip_close(fp_r00_acc_list) == 1) or
	isip_die("ERROR: couldn't close $r00_acc_list\n"); 
    
    (isip_close(fp_r01_acc_list) == 1) or
	    isip_die("ERROR: couldn't close $r01_acc_list\n"); 
    
    # input files and the location they should copied to in the main base
    # experiment directory. we'll later copy odd files in the list to even 
    # files in the list
    #
    my @TI_BASE_MONO_COPY_PAIRS =
	("$TI_PATH_TO_MFC/transcriptions/mono_trans_no_sp.text",
	 "$TI_PATH_TO_BASE_MONO/transcriptions/mono_trans_no_sp.text",
	 "$TI_PATH_TO_MFC/transcriptions/mono_trans_with_sp.text",
	 "$TI_PATH_TO_BASE_MONO/transcriptions/mono_trans_with_sp.text",
	 "$TI_PATH_TO_TRAIN_DATA/model_lengths.text",
	 "$TI_PATH_TO_BASE_MONO/r00/hmm0/model_lengths.text",
	 "$TI_PATH_TO_TRAIN_DATA/monophones.text",
	 "$TI_PATH_TO_BASE_MONO/r00/hmm0/monophones.text",
	 "$TI_PATH_TO_TRAIN_DATA/lexicon.text",
	 "$TI_PATH_TO_BASE_MONO/transcriptions/lexicon.text");

    # copy over the necessary data files to the start directories
    #
    for ($j = 0; $j <= $#TI_BASE_MONO_COPY_PAIRS; $j+= 2) {
	(isip_copy($TI_BASE_MONO_COPY_PAIRS[$j], 
		   $TI_BASE_MONO_COPY_PAIRS[$j+1]) == 1) or return 0;
    }

    # copy the input files to the  the location they should copied to 
    # in the base experiment directory for training jobs on the cpus.
    #
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
	(isip_copy("$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/transcriptions/mono_trans_no_sp.text", "$TI_PATH_TO_BASE_MONO_CPUS[$i]/transcriptions/mono_trans_no_sp.text") == 1) or return 0;
	
	(isip_copy("$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/transcriptions/mono_trans_with_sp.text","$TI_PATH_TO_BASE_MONO_CPUS[$i]/transcriptions/mono_trans_with_sp.text") == 1) or return 0;
    }
            
    print "INFO: bw_build_base_mono_models - all necessary paths now exist\n";

    # set the locations of some needed files
    #
    $model_lengths_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/model_lengths.text";
    $models_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/models.text";
    $monophones_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/monophones.text";
    $lexicon_file = "$TI_PATH_TO_BASE_MONO/transcriptions/lexicon.text";
    $phones_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/phones.text";
    $init_transitions_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/transitions.text";
    $init_states_text_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/states.text";
    $init_states_bin_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/states.bin";
    $vfloor_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/vfloor.text";
    $model_transcript =
	"$TI_PATH_TO_BASE_MONO/transcriptions/mono_trans_no_sp.text";
    $word_transcript =
	"$TI_PATH_TO_BASE_MONO/transcriptions/word_transcription.text";
    $occupancy_file = "";

    # set the values for training
    #
    $vfloor_value = "0.0002";
    $transcription_mode = "phone";
    $context_mode = "monophone";
    $min_mpd = "1000";
    $beam_width = $train_beam_pruning;
    $min_occupancy = "1.0e-10";
    $min_model_count = "3";
    
    if ($mode eq $TI_MODE_ACC) {
	$compute_delta = "yes";
	$compute_acc = "yes";
	$delta_win = "2";
    }
    elsif ($mode eq $TI_MODE_DELTA) {
	$compute_delta = "yes";
	$compute_acc = "no";
	$delta_win = "2";
    }
    else {
	$compute_delta = "no";
	$compute_acc = "no";
	$delta_win = "";
    }
    
    
    $input_feature_format = $feature_format;

    # print a debugging statement
    #
    print "INFO: bw_build_base_mono_models - building initial base monophone";
    print " models\n";
    
    # create the initial left-to-right hmms with a call to create_models
    #
    (run_create_models($model_lengths_file, $models_file, 
		       $cpus_train_list[0]) == 1) or return 0;
	
    # create the phone cluster mapping with a call to create_triphone_map
    #
    (run_create_triphone_map($monophones_file, $monophones_file, "ci",
			     $models_file, $phones_file, $cpus_train_list[0]) 
     == 1) or return 0;

    # initialize the hmm statistics with a call to init_hmm
    #
    (run_init_hmm($TI_PATH_TO_TRAIN_LIST, $model_lengths_file,
		  $init_transitions_file, $init_states_text_file,
		  $TI_BINARY_MODE, $vfloor_file,
		  $vfloor_value, $num_features, $input_feature_format, 
		  $mode, $delta_win, $cpus_train_list[0]) == 1) or return 0;

    # convert the ascii states file to binary
    #
    (run_convert_mmf($TI_ASCII_MODE, $TI_BINARY_MODE, $init_states_text_file,
		     $init_states_bin_file, $cpus_train_list[0]) == 1) 
	or return 0;

    # print a debugging statement
    #
    print "INFO: bw_build_base_mono_models - initial base monophone models";
    print " complete.\n";

    # begin single mixture training
    #
    print "INFO: bw_build_base_mono_models - beginning four passes of";
    print " single-mixture training without the sp model.\n";

    # loop for four iterations
    # 
    for ($j = 1; $j <= 4; $j++) {
	
	# build the parameters that are required for both 
	# batch and combine mode
	#
	$input_dir = sprintf("%s%d", "$TI_PATH_TO_BASE_MONO/r00/hmm", $j-1);
	$input_states = "$input_dir/states.bin";
	$input_trans = "$input_dir/transitions.text";

	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j\n";

	for ($i = 0; $i <= $#cpus_train_list; $i++) {

	    # debug message
	    #
	    print "INFO: bw_build_base_mono_models - iteration $j cpu $sort_cpus_train[$i]\n";

	    # build the parameter file for training on cpus in batch mode 
	    #
	    $train_mode = "batch";
	    $input_dir_cpus = sprintf("%s%d", 
				      "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/hmm"
				      , $j-1);
	    $output_dir_cpus = sprintf("%s%d", 
				       "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/hmm", $j);
	    $mfc_list = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list";
	    $model_transcript_cpus = "$TI_PATH_TO_BASE_MONO_CPUS[$i]/transcriptions/mono_trans_no_sp.text";
	    $acc_file = "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r00/acc/acc.bin";
	    $acc_list_file = "";
	    $output_states = "";
	    $output_trans = "";
	    $param_file_cpus = "$input_dir_cpus/params.text";
	    $shell_script = "$input_dir_cpus/run_bw_train.sh";
	    $log_file = "$input_dir_cpus/bw_train.log";
	    $completion_file[$i] = "$input_dir_cpus/train_job_completed.text";

	    # create the shell script that would run the bw_train
	    #
	    run_create_shell_script_train($shell_script, $param_file_cpus, 
					  $log_file, $completion_file[$i], 
					  $cpus_train_list[$i], 
					  $walltime);

	    # write the file and run bw_train in batch mode
	    #
	    (run_bw_train($train_mode, $mfc_list, 
			  $model_transcript_cpus, $monophones_file, 
			  $models_file,
			  $phones_file, $lexicon_file, $input_states, 
			  $input_trans, $output_states, $output_trans,
			  $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
			  $min_mpd, $beam_width, $min_occupancy,
			  $min_model_count, $vfloor_file, $occupancy_file,
			  $acc_file, $acc_list_file, $param_file_cpus, 
			  $input_feature_format, $compute_delta, $compute_acc,
			  $delta_win, $cpus_train_list[$i], $shell_script,
			  $walltime)
	     == 1) 
		or return 0;
	}
	
	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j waiting for\n";
	print "the training jobs to complete\n";

	# wait for the jobs to finish 
	#
	$done = 0;
	while ($done == 0) {
	    sleep 1;
	    $done = 1;
	    for ($i = 0; $i <= $#cpus_train_list; $i++) {
		if ( ! -f $completion_file[$i]) {
		    $done = 0;
		}
	    }
	}

	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j combining accumulators\n";

	# build the parameter file to run bw_train in combine mode 
	#
	$train_mode = "combine";
	$mfc_list ="";
	$model_transcript = "";
	$acc_file = "";
	$acc_list_file = $r00_acc_list;
	$output_dir = sprintf("%s%d", "$TI_PATH_TO_BASE_MONO/r00/hmm", $j);
	$output_states = "$output_dir/states.bin";
	$output_trans = "$output_dir/transitions.text";
	$param_file = "$input_dir/params.text";
	$shell_script = "";

	# write the file and run bw_train in combine mode
	#
	(run_bw_train($train_mode, $mfc_list, 
		      $model_transcript, $monophones_file, $models_file,
		      $phones_file, $lexicon_file, $input_states, 
		      $input_trans, $output_states, $output_trans,
		      $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
		      $min_mpd, $beam_width, $min_occupancy,
		      $min_model_count, $vfloor_file, $occupancy_file,
		      $acc_file, $acc_list_file, $param_file, 
		      $input_feature_format, $compute_delta, $compute_acc,
                      $delta_win, $cpus_train_list[0], $shell_script) == 1) 
	    or return 0;
    }

    print "INFO: bw_build_base_mono_models - completed four passes of";
    print " single-mixture training.\n";

    # build sp model and silence model with skip transitions
    #
    print "INFO: bw_build_base_mono_models - tying sp model to central\n";
    print "INFO: bw_build_base_mono_models - state of silence model and\n";
    print "INFO: bw_build_base_mono_models - adding skip states to silence";
    print " model\n";
    
    # filter the transitions file to tie sp to silence and to add skip 
    # states to the sil model. the state-tying is already done by using
    # a different models file
    #
    $last_trans_file = "$TI_PATH_TO_BASE_MONO/r00/hmm4/transitions.text";
    $last_models_file = "$TI_PATH_TO_BASE_MONO/r00/hmm0/models.text";
    $last_states_file = "$TI_PATH_TO_BASE_MONO/r00/hmm4/states.bin";
    $next_states_file = "$TI_PATH_TO_BASE_MONO/r01/hmm0/states.bin";
    $next_trans_file = "$TI_PATH_TO_BASE_MONO/r01/hmm0/transitions.text";
    $next_models_file = "$TI_PATH_TO_BASE_MONO/r01/hmm0/models.text";
    (add_sp_and_sil_trans($last_trans_file, $next_trans_file,
		  $models_file, $next_models_file) == 1) or return 0;

    # copy the states file over
    #
    (isip_copy($last_states_file, $next_states_file) == 1) or return 0;

    # update the pointer to the models file and the phones file
    #
    $models_file = $next_models_file;
    $phones_file = "$TI_PATH_TO_BASE_MONO/r01/hmm0/phones.text";

    # call create_triphone_map
    #
    (run_create_triphone_map($monophones_file, $monophones_file, "ci",
			     $models_file, $phones_file, $cpus_train_list[0])
     == 1) or return 0;

    # begin single mixture training with sp
    #
    print "INFO: bw_build_base_mono_models - beginning four passes of";
    print " single-mixture training.\n";

    # loop for four iterations
    # 
    for ($j = 1; $j <= 4; $j++) {
	
	# build the parameters that are required for both batch and
	# combine mode
	#
	$input_dir = sprintf("%s%d", "$TI_PATH_TO_BASE_MONO/r01/hmm", $j-1);
	$input_states = "$input_dir/states.bin";
	$input_trans = "$input_dir/transitions.text";

	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j\n";

	for ($i = 0; $i <= $#cpus_train_list; $i++) {

	    # debug message
	    #
	    print "INFO: bw_build_base_mono_models - iteration $j cpu $sort_cpus_train[$i]\n";
	    
	    
	    # build the parameter file for training on cpus in batch mode 
	    #
	    $train_mode = "batch";
	    $input_dir_cpus = sprintf("%s%d", 
				      "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm"
				      , $j-1);
	    $output_dir_cpus = sprintf("%s%d", 
				       "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm", $j);
	    $mfc_list = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list";
	    $acc_file = "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/acc/acc.bin";
	    $acc_list_file = "";
	    $output_states = "";
	    $output_trans = "";
	    $param_file_cpus = "$input_dir_cpus/params.text";
	    $shell_script = "$input_dir_cpus/run_bw_train.sh";
	    $log_file = "$input_dir_cpus/bw_train.log";
	    $completion_file[$i] = "$input_dir_cpus/train_job_completed.text";

	    # change the transcription file to use the one with 
	    # embedded sp models.
	    #
	    $model_transcript_cpus = "$TI_PATH_TO_BASE_MONO_CPUS[$i]/transcriptions/mono_trans_with_sp.text";
	    
	    # create the shell script that would run the bw_train
	    #
	    run_create_shell_script_train($shell_script, $param_file_cpus, 
					  $log_file, $completion_file[$i], 
					  $cpus_train_list[$i]);

	    # write the file and run bw_train in batch mode
	    #
	    (run_bw_train($train_mode, $mfc_list, 
			  $model_transcript_cpus, $monophones_file, 
			  $models_file,
			  $phones_file, $lexicon_file, $input_states, 
			  $input_trans, $output_states, $output_trans,
			  $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
			  $min_mpd, $beam_width, $min_occupancy,
			  $min_model_count, $vfloor_file, $occupancy_file,
			  $acc_file, $acc_list_file, $param_file_cpus,
			  $input_feature_format, $compute_delta, $compute_acc,
			  $delta_win, $cpus_train_list[$i], $shell_script) 
	     == 1) 
		or return 0;
	}

	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j waiting for\n";
	print "the training jobs to complete\n";
	
	# wait for the jobs to finish 
	#
	$done = 0;
	while ($done == 0) {
	    sleep 1;
	    $done = 1;
	    for ($i = 0; $i <= $#cpus_train_list; $i++) {
		if ( ! -f $completion_file[$i]) {
		    $done = 0;
		}
	    }
	}
	
	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j combining accumulators\n";
	
	# build the parameter file to run bw_train in combine mode 
	#
	$train_mode = "combine";
	$mfc_list = "";
	$model_transcript = "";
	$acc_file = "";
	$acc_list_file = $r01_acc_list;
	$output_dir = sprintf("%s%d", "$TI_PATH_TO_BASE_MONO/r01/hmm", $j);
	$output_states = "$output_dir/states.bin";
	$output_trans = "$output_dir/transitions.text";
	$param_file = "$input_dir/params.text";
	$shell_script = "";

	# write the file and run bw_train in combine mode
	#
	(run_bw_train($train_mode, $mfc_list, 
		      $model_transcript, $monophones_file, $models_file,
		      $phones_file, $lexicon_file, $input_states, 
		      $input_trans, $output_states, $output_trans,
		      $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
		      $min_mpd, $beam_width, $min_occupancy,
		      $min_model_count, $vfloor_file, $occupancy_file,
		      $acc_file, $acc_list_file, $param_file, 
		      $input_feature_format, $compute_delta, $compute_acc,
                      $delta_win, $cpus_train_list[0], $shell_script) == 1) 
	    or return 0;
    }
    
    print "INFO: bw_build_base_mono_models - completed four passes of";
    print " single-mixture training.\n";
    
    # skip the forced alignment
    #
    print "INFO: bw_build_base_mono_models - skipping force alignment\n";
##    print "INFO: bw_build_base_mono_models - aligning models to generate\n";
##    print "INFO: bw_build_base_mono_models -  best pronunciation variant.\n";

    # set the files used and created during forced alignment
    #
##    $output_list = "$TI_PATH_TO_BASE_MONO/alignments/output_list.text";
##    $align_states_file = "$TI_PATH_TO_BASE_MONO/r01/hmm4/states.bin";
##    $align_transition_file = "$TI_PATH_TO_BASE_MONO/r01/hmm4/transitions.text";
##    $align_output_dir = "$TI_PATH_TO_BASE_MONO/alignments/output";
##    $align_transcript="$TI_PATH_TO_BASE_MONO/transcriptions/aligned_trans.text";
##    $align_mfcc_list="$TI_PATH_TO_BASE_MONO/transcriptions/aligned_mfcc.text";
##    $align_params = "$TI_PATH_TO_BASE_MONO/alignments/params.text";

    # define the search parameters used during alignment
    #
##    $state_beam = "400";
##    $model_beam = "400";
##    $word_beam = "400";
##    $mapmi_limit = "2000";
##    $num_allowed_words = "100";

    # build the output file list from the template file
    #
##    $output_list_template = "$TI_PATH_TO_TRAIN_DATA/aligned_output.list";
##    $output_list_string = "__WSJ_PATH_TO_ALIGN__";

##    (file_replace_string($output_list_string, $align_output_dir,
##			 $output_list_template, $output_list) == 1)
##	or return 0;
    
    # run the alignment routine
    #
##    (run_model_alignment($TI_TRAIN_MFC_FILE_LIST, $output_list, 
##			 $word_transcript, $monophones_file, $models_file,
##			 $phones_file, $lexicon_file, $align_states_file,
##			 $align_transition_file, $TI_BINARY_MODE,
##			 $context_mode, $state_beam, $model_beam, $word_beam,
##			 $mapmi_limit, $num_allowed_words, $align_output_dir,
##			 $align_transcript, $align_mfcc_list,
##			 $align_params) == 1) or return 0;
    
##    print "INFO: bw_build_base_mono_models - forced alignment completed.";
##    print " further training will use\n";
##    print "INFO: bw_build_base_mono_models - the following files as feature";
##    print " file list and transcription\n";
##    print "INFO: bw_build_base_mono_models - list respectively:\n";
##    print "INFO: bw_build_base_mono_models - $align_mfcc_list\n";
##    print "INFO: bw_build_base_mono_models - $align_transcript\n";
    

    # continue building the models with the aligned transcriptions
    #
##    print "INFO: bw_build_base_mono_models - continuing single-mixture";
##    print " training with aligned\n";
##    print "INFO: bw_build_base_mono_models - transcriptions.\n";


    # loop for remaining iterations
    # 
    for ($j = 5; $j <= 9; $j++) {
	
	# build the parameters that are required for both batch and
	# combine mode
	#
	$input_dir = sprintf("%s%d", "$TI_PATH_TO_BASE_MONO/r01/hmm", $j-1);
	$input_states = "$input_dir/states.bin";
	$input_trans = "$input_dir/transitions.text";

	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j\n";
	
	for ($i = 0; $i <= $#cpus_train_list; $i++) {
	    
	    # debug message
	    #
	    print "INFO: bw_build_base_mono_models - iteration $j cpu $sort_cpus_train[$i]\n";
	    
	    
	    # build the parameter file for training on cpus in batch mode 
	    #
	    $train_mode = "batch";
	    $input_dir_cpus = sprintf("%s%d", 
				      "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm"
				      , $j-1);
	    $output_dir_cpus = sprintf("%s%d", 
				       "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/hmm", $j);
	    $mfc_list = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list";
	    $acc_file = "$TI_PATH_TO_BASE_MONO_CPUS[$i]/r01/acc/acc.bin";
	    $acc_list_file = "";
	    $output_states = "";
	    $output_trans = "";
	    $param_file_cpus = "$input_dir_cpus/params.text";
	    $shell_script = "$input_dir_cpus/run_bw_train.sh";
	    $log_file = "$input_dir_cpus/bw_train.log";
	    $completion_file[$i] = "$input_dir_cpus/train_job_completed.text";

 	    # change the transcription file to use the one with 
	    # embedded sp models.
	    #
	    $model_transcript_cpus = "$TI_PATH_TO_BASE_MONO_CPUS[$i]/transcriptions/mono_trans_with_sp.text";

	    # create the shell script that would run the bw_train
	    #
	    run_create_shell_script_train($shell_script, $param_file_cpus, 
					  $log_file, $completion_file[$i], 
					  $cpus_train_list[$i]);

	    # write the file and run bw_train in batch mode
	    #
	    (run_bw_train($train_mode, $mfc_list,
			  $model_transcript_cpus, $monophones_file,
			  $models_file, 
			  $phones_file, $lexicon_file,
			  $input_states, $input_trans, $output_states,
			  $output_trans, $TI_BINARY_MODE, $TI_BINARY_MODE,
			  $context_mode, $min_mpd, $beam_width, $min_occupancy,
			  $min_model_count, $vfloor_file, $occupancy_file,
			  $acc_file, $acc_list_file, $param_file_cpus, 
			  $input_feature_format, $compute_delta, $compute_acc,
			  $delta_win, $cpus_train_list[$i], $shell_script) 
	     == 1) 
		or return 0;
	}
	
    	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j waiting for\n";
	print "the training jobs to complete\n";
	
	# wait for the jobs to finish 
	#
	$done = 0;
	while ($done == 0) {
	    sleep 1;
	    $done = 1;
	    for ($i = 0; $i <= $#cpus_train_list; $i++) {
		if ( ! -f $completion_file[$i]) {
		    $done = 0;
		}
	    }
	}
	
	# debug message
	#
	print "INFO: bw_build_base_mono_models - iteration $j combining accumulators\n";
	
	# build the parameter file to run bw_train in combine mode 
	#
	$train_mode = "combine";
	$mfc_list = "";
	$model_transcript = "";
	$acc_file = "";
	$acc_list_file = $r01_acc_list;
	$output_dir = sprintf("%s%d", "$TI_PATH_TO_BASE_MONO/r01/hmm", $j);
	$output_states = "$output_dir/states.bin";
	$output_trans = "$output_dir/transitions.text";
	$param_file = "$input_dir/params.text";
	$shell_script = "";

	# write the file and run bw_train in combine mode
	#
	(run_bw_train($train_mode, $mfc_list, 
		      $model_transcript, $monophones_file, $models_file,
		      $phones_file, $lexicon_file, $input_states, 
		      $input_trans, $output_states, $output_trans,
		      $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
		      $min_mpd, $beam_width, $min_occupancy,
		      $min_model_count, $vfloor_file, $occupancy_file,
		      $acc_file, $acc_list_file, $param_file, 
		      $input_feature_format, $compute_delta, $compute_acc,
                      $delta_win, $cpus_train_list[0], $shell_script) == 1) 
	    or return 0;
    }
    
    # copy final models
    #
    print "INFO: bw_build_base_mono_models - copying final models to";
    print " $TI_PATH_TO_FINAL_MODELS\n";

    # copy the necessary files
    #
    (isip_copy($monophones_file, 
	       "$TI_PATH_TO_FINAL_MODELS/monophones.text") == 1) or return 0;
    (isip_copy($models_file,
	       "$TI_PATH_TO_FINAL_MODELS/models.text") == 1)  or return 0;
    (isip_copy($phones_file,
	       "$TI_PATH_TO_FINAL_MODELS/phones.text") == 1) or return 0;
    (isip_copy($lexicon_file,
	       "$TI_PATH_TO_FINAL_MODELS/lexicon.text") == 1) or return 0;
    (isip_copy($vfloor_file,
	       "$TI_PATH_TO_FINAL_MODELS/vfloor.text") == 1) or return 0;
    (isip_copy("$TI_PATH_TO_BASE_MONO/r01/hmm9/states.bin",
	       "$TI_PATH_TO_FINAL_MODELS/states.bin") == 1) or return 0;
    (isip_copy("$TI_PATH_TO_BASE_MONO/r01/hmm9/transitions.text",
	       "$TI_PATH_TO_FINAL_MODELS/transitions.text") == 1) or return 0;
    
    # exit gracefully
    #
    return 1;
}


# subroutine: bw_build_xwrd_models
#
# arguments: 
#  @_ = (mixtures, cpus_train_list);
#
# description: this routine builds a set of cross-word triphone mixture models
#              from scratch on multiple cpus
#
# return: 0 if error, 1 if no error
#
sub bw_build_xwrd_models ($,@) {

    # parse the argument list
    #
    my $mixtures = shift @_;
    my @cpus_train_list = @_;

    # create sub-directory name for training cpus
    #
    $j = 0;
    for ($i = 0; $i <= $#cpus_train_list; $i++) {
	
	if ($cpus_train_list[$i] ne $cpus_train_list[$i-1]) {
	    $j = 0;
	}
	
	if (($cpus_train_list[$i] eq $cpus_train_list[$i-1]) && ($i != 0)) {
	    $j++; 
	}
	
	$cpus_train_num[$i] = join("_", $cpus_train_list[$i], $j);
    }
    
    # don't sort the training cpus
    #
    @sort_cpus_train = @cpus_train_num;

    # a path to the data directory
    #
    $TI_PATH_TO_TRAIN_DATA = "$TI_PATH_TO_TRAIN/xwrd_tri";

    # a path to the working main directories
    #
    my $TI_PATH_TO_XWRD = "$TI_PATH_TO_BW/xwrd_tri";
    my $TI_PATH_TO_BASE_MONO = "$TI_PATH_TO_BW/monophone/base";

    # a path to the working directories for training jobs on cpus
    #
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
	$TI_PATH_TO_XWRD_CPUS[$i] 
	    = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/train/baum_welch/xwrd_tri";
	$TI_PATH_TO_BASE_MONO_CPUS[$i] 
	    = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/train/baum_welch/monophone/base";
    }
    
    # a path to the final models directories
    #
    my $TI_PATH_TO_FINAL_MODELS = "$TI_PATH_TO_XWRD/final_models";
    my $TI_PATH_TO_BASE_FINAL_MODELS = "$TI_PATH_TO_BASE_MONO/final_models";

    # the files that must be present for this function to be complete
    #
    my @TI_XWRD_MODEL_DEPENDENCIES = 
	("$TI_PATH_TO_FINAL_MODELS/monophones.text",
	 "$TI_PATH_TO_FINAL_MODELS/models.text",
	 "$TI_PATH_TO_FINAL_MODELS/phones.text",
	 "$TI_PATH_TO_FINAL_MODELS/lexicon.text",
	 "$TI_PATH_TO_FINAL_MODELS/vfloor.text",
	 "$TI_PATH_TO_FINAL_MODELS/special_models.text", 
	 "$TI_PATH_TO_FINAL_MODELS/states.bin",
	 "$TI_PATH_TO_FINAL_MODELS/transitions.text",
	 "$TI_PATH_TO_XWRD/state_tying/output_trees.text");

    # check that we don't already have these models
    #
    $found_all_files = 1;
    foreach $file (@TI_XWRD_MODEL_DEPENDENCIES) {
	(file_exists($file) == 1) or $found_all_files = 0;
    }
    
    if ($found_all_files == 1) {
	print "INFO: bw_build_xwrd_models - pre-existing cross-word";
	print " triphone models found...skipping\n\n\n";
	return 1;
    }
    
    # build directories as needed
    #
    print "INFO: bw_build_xwrd_models - building necessary paths and files\n";
    
    # check the main base directory
    #
    if ( ! -d "$TI_PATH_TO_XWRD") {
	(isip_mkdir($TI_PATH_TO_XWRD) == 1) or return 0;
    }

    # check the base directory for training on cpus
    #
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
       if ( ! -d "$TI_PATH_TO_XWRD_CPUS[$i]") {
	   (isip_mkdir($TI_PATH_TO_XWRD_CPUS[$i]) == 1) or return 0;
       }
   }
    
    # main directories to create for this experiment
    #
    my @TI_XWRD_MODEL_DIRS = 
	("$TI_PATH_TO_FINAL_MODELS",
	 "$TI_PATH_TO_XWRD/transcriptions",
	 "$TI_PATH_TO_XWRD/r02/hmm0",
	 "$TI_PATH_TO_XWRD/r02/hmm1",
	 "$TI_PATH_TO_XWRD/r02/hmm2",
	 "$TI_PATH_TO_XWRD/r02/hmm3",
	 "$TI_PATH_TO_XWRD/r02/hmm4",
	 "$TI_PATH_TO_XWRD/r02/hmm5",
	 "$TI_PATH_TO_XWRD/r02/hmm6",
	 "$TI_PATH_TO_XWRD/r02/hmm7",
	 "$TI_PATH_TO_XWRD/r02/hmm8",
	 "$TI_PATH_TO_XWRD/r02/hmm9",
	 "$TI_PATH_TO_XWRD/r02/hmm10",
	 "$TI_PATH_TO_XWRD/r03/hmm0",
	 "$TI_PATH_TO_XWRD/r03/hmm20",
	 "$TI_PATH_TO_XWRD/r03/hmm21",
	 "$TI_PATH_TO_XWRD/r03/hmm22",
 	 "$TI_PATH_TO_XWRD/r03/hmm23",
	 "$TI_PATH_TO_XWRD/r03/hmm24",
	 "$TI_PATH_TO_XWRD/r03/hmm40",
	 "$TI_PATH_TO_XWRD/r03/hmm41",
	 "$TI_PATH_TO_XWRD/r03/hmm42",
 	 "$TI_PATH_TO_XWRD/r03/hmm43",
	 "$TI_PATH_TO_XWRD/r03/hmm44",
	 "$TI_PATH_TO_XWRD/r03/hmm80",
	 "$TI_PATH_TO_XWRD/r03/hmm81",
	 "$TI_PATH_TO_XWRD/r03/hmm82",
 	 "$TI_PATH_TO_XWRD/r03/hmm83",
	 "$TI_PATH_TO_XWRD/r03/hmm84",
	 "$TI_PATH_TO_XWRD/r03/hmm160",
	 "$TI_PATH_TO_XWRD/r03/hmm161",
	 "$TI_PATH_TO_XWRD/r03/hmm162",
 	 "$TI_PATH_TO_XWRD/r03/hmm163",
	 "$TI_PATH_TO_XWRD/r03/hmm164",
	 "$TI_PATH_TO_XWRD/state_tying");
    
    # create all other necessary directories
    #
    foreach $dir (@TI_XWRD_MODEL_DIRS) {
	if ( ! -d "$dir") {
	    (isip_mkdir($dir) == 1) or return 0;
	}
    }
    
    # sub-directories in base directories for training jobs on cpus
    # to create for this experiment
    #
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {
	my @TI_XWRD_MODEL_DIRS_CPUS = 
	    ("$TI_PATH_TO_XWRD_CPUS[$i]/transcriptions",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm0",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm1",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm2",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm3",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm4",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm5",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm6",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm7",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm8",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm9",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm10",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r02/acc",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm0",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm20",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm21",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm22",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm23",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm24",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm40",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm41",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm42",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm43",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm44",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm80",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm81",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm82",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm83",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm84",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm160",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm161",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm162",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm163",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm164",
	     "$TI_PATH_TO_XWRD_CPUS[$i]/r03/acc");
	
	# create all other necessary directories
	#
	foreach $dir (@TI_XWRD_MODEL_DIRS_CPUS) {
	    if ( ! -d "$dir") {
		(isip_mkdir($dir) == 1) or return 0;
	    }
	}
    }
	
    # create the accumulator list that will be used for combining the 
    # accumulators
    #
    $r02_acc_list = "$TI_PATH_TO_XWRD/r02/acc.list"; 
    (isip_open(fp_r02_acc_list, ">$r02_acc_list") == 1) or 
	isip_die("ERROR: couldn't open $r02_acc_list\n"); 
    
    $r03_acc_list = "$TI_PATH_TO_XWRD/r03/acc.list"; 
    (isip_open(fp_r03_acc_list, ">$r03_acc_list") == 1) or 
	isip_die("ERROR: couldn't open  $r03_acc_list\n"); 
    
    for ($i = 0; $i <= $#sort_cpus_train; $i++) {	
	print fp_r02_acc_list 
	    "$TI_PATH_TO_XWRD_CPUS[$i]/r02/acc/acc.bin\n";

	print fp_r03_acc_list 
	    "$TI_PATH_TO_XWRD_CPUS[$i]/r03/acc/acc.bin\n";
    }
    
    (isip_close(fp_r02_acc_list) == 1) or
	isip_die("ERROR: couldn't close $r02_acc_list\n"); 
    (isip_close(fp_r03_acc_list) == 1) or
	    isip_die("ERROR: couldn't close $r03_acc_list\n");     
    

    # input files and the location they should copied to in the experiment
    # directory. the even-indexed items are packaged with the system
    # the odd-indexed items are for this experiment only
    #
    my @TI_XWRD_COPY_PAIRS = ("$TI_PATH_TO_BASE_MONO/r00/hmm0/monophones.text",
		      "$TI_PATH_TO_XWRD/r02/hmm0/monophones.text",
		      "$TI_PATH_TO_BASE_FINAL_MODELS/transitions.text",
		      "$TI_PATH_TO_XWRD/r02/hmm0/transitions.text",
		      "$TI_PATH_TO_BASE_MONO/transcriptions/lexicon.text",
		      "$TI_PATH_TO_XWRD/transcriptions/lexicon.text",
		      "$TI_PATH_TO_BASE_FINAL_MODELS/vfloor.text",
		      "$TI_PATH_TO_XWRD/r02/hmm0/vfloor.text",
		      "$TI_PATH_TO_TRAIN_DATA/special_models.text", 
		      "$TI_PATH_TO_XWRD/state_tying/special_models.text",
		      "$TI_PATH_TO_TRAIN/state_tying/isip_questions.text",
		      "$TI_PATH_TO_XWRD/state_tying/isip_questions.text");
    
    # the first thing we need to do is build the base monophone models
    #
#    print "INFO: bw_build_xwrd_models - building base monophone models\n"; 
#    (bw_build_base_mono_models(@cpus_train_list) == 1) or return 0;

    # copy over the necessary data files to the start directories
    #
    for ($j = 0; $j <= $#TI_XWRD_COPY_PAIRS; $j+= 2) {
	(isip_copy($TI_XWRD_COPY_PAIRS[$j],
		   $TI_XWRD_COPY_PAIRS[$j+1]) == 1) or return 0;
    }

    print "INFO: bw_build_xwrd_models - all necessary paths now exist\n";

    # set the locations of some needed files
    #
    $xwrd_transcript_file =
	"$TI_PATH_TO_XWRD/transcriptions/xwrd_transcriptions.text";
    $mono_transcript_file =
	"$TI_PATH_TO_BASE_MONO/transcriptions/mono_trans_with_sp.text";
    $mono_models_file = "$TI_PATH_TO_BASE_FINAL_MODELS/models.text";
    $mono_states_file = "$TI_PATH_TO_BASE_FINAL_MODELS/states.bin";
    $mono_transitions_file = "$TI_PATH_TO_BASE_FINAL_MODELS/transitions.text";

    $xwrd_triphone_list="$TI_PATH_TO_XWRD/transcriptions/xwrd_triphones.text";
    $mfcc_list = "$TI_PATH_TO_TRAIN_LIST";
    $models_file = "$TI_PATH_TO_XWRD/r02/hmm0/models.text";
    $monophones_file = "$TI_PATH_TO_XWRD/r02/hmm0/monophones.text";
    $lexicon_file = "$TI_PATH_TO_XWRD/transcriptions/lexicon.text";
    $phones_file = "$TI_PATH_TO_XWRD/r02/hmm0/phones.text";
    $init_transitions_file = "$TI_PATH_TO_XWRD/r02/hmm0/transitions.text";
    $init_states_text_file = "$TI_PATH_TO_XWRD/r02/hmm0/states.text";
    $init_states_bin_file = "$TI_PATH_TO_XWRD/r02/hmm0/states.bin";
    $vfloor_file = "$TI_PATH_TO_XWRD/r02/hmm0/vfloor.text";
    $model_transcript = $xwrd_transcript_file;
    $acc_file = "";
    $acc_list_file = "";


    # set the values for training
    #
    $vfloor_value = "0.0002";
    $transcription_mode = "phone";
    $context_mode = "cross_word";
    $min_mpd = "1000";
    $min_occupancy = "1.0e-10";
    $min_model_count = "3";

    if ($mode eq $TI_MODE_ACC) {
	$compute_delta = "yes";
	$compute_acc = "yes";
	$delta_win = "2";
    }
    elsif ($mode eq $TI_MODE_DELTA) {
	$compute_delta = "yes";
	$compute_acc = "no";
	$delta_win = "2";
    }
    else {
	$compute_delta = "no";
	$compute_acc = "no";
	$delta_win = "";
    }
    
    # set the mixture sequence
    #
    if ($mixtures == $TI_NUM_MIX_1) {
	@num_mixes = ();
    }

    elsif ($mixtures == $TI_NUM_MIX_2) {
	@num_mixes = (2);
    }

    elsif ($mixtures == $TI_NUM_MIX_4) {
	@num_mixes = (2, 4);
    }

    elsif ($mixtures == $TI_NUM_MIX_8) {
	@num_mixes = (2, 4, 8);
    }

    elsif ($mixtures == $TI_NUM_MIX_16) {
	@num_mixes = (2, 4, 8, 16);
    }

    else {
	isip_die("ERROR: Failed to build cross-word mixture models\n");
    }

    # print a debugging message
    #
    print "INFO: bw_build_xwrd_models - building cross-word transcriptions";
    print " file:\n";
    print "INFO: bw_build_xwrd_models - $xwrd_transcript_file ";
    print "INFO: bw_build_xwrd_models - from $mono_transcript_file\n";

    # build cross-word triphone transcriptions
    #
    (build_xwrd_transcription($mono_transcript_file,
			      $xwrd_transcript_file) == 1) or return 0;
    
    # print a debugging message
    #
    print "INFO: bw_build_xwrd_models - dividing the cross-word\n"; 
    print "transcription equal to the number of corresponding cpus\n";

    # divide the cross-word triphone transcriptions according to the 
    # number of cpu's
    #
    (create_N_lists(($#sort_cpus_train + 1), $xwrd_transcript_file, 
		    "$TI_PATH_TO_XWRD/transcriptions", 
		    "xwrd_transcriptions", "text")
     == 1) 
	or isip_die("ERROR: bw_build_xwrd_models - cross-word triphones transcription not divided into N ways...exiting");
    
    # print a debugging message
    #
    print "INFO: bw_build_xwrd_models - moving transcriptions to\n";
    print " respective subdirectories corresponding to the cpus..\n";
    
    # moving the transcriptions to the cpus directory
    # from the main directory
    #
    for( $i = 0; $i <= $#sort_cpus_train; $i++) {
	(isip_move("$TI_PATH_TO_XWRD/transcriptions/xwrd_transcriptions_$i.text", "$TI_PATH_TO_XWRD_CPUS[$i]/transcriptions/xwrd_transcriptions.text"))
	    or isip_die("ERROR: bw_build_xwrd_models - cross-word triphone sub-transcriptions not moved...exiting");
    }
    
    # build the list of unique cross-word triphones
    #
    print "INFO: bw_build_xwrd_models - building a list of all unique";
    print " triphones in the\n";
    print "INFO: bw_build_xwrd_models - training set and writing to";
    print "$xwrd_triphone_list\n";
    (build_phone_list($xwrd_transcript_file,
		      $xwrd_triphone_list) == 1)  or return 0;

    # initializing triphone models from their central monophone models
    #
    print "INFO: bw_build_xwrd_models - initializing triphone models from\n";
    print "their central monophone models\n";
    $itri_param_file = "$TI_PATH_TO_XWRD/r02/hmm0/params_init_triphones.text";

    (run_init_triphones($monophones_file, $mono_models_file, $mono_states_file,
			$mono_transitions_file, $xwrd_triphone_list,
			$models_file, $init_states_bin_file,
			$itri_param_file, $cpus_train_list[0]) 
     == 1) or return 0;
    
    # creating the new triphone map
    #
    # creating the new triphone map
    #
    (run_create_triphone_map($monophones_file, $xwrd_triphone_list, "cd",
			     $models_file, $phones_file)
     == 1) or return 0;

    # run four passes of single mixture triphone training
    #
    print "INFO: bw_build_xwrd_models - beginning four passes of";
    print " single-mixture triphone training.\n";

    # loop for four iterations
    # 
    for ($j = 1; $j <= 4; $j++) {
	
	# build the parameters that are required for both
	# batch and combine mode
	#
	$input_dir = sprintf("%s%d", "$TI_PATH_TO_XWRD/r02/hmm", $j-1);
	$input_states = "$input_dir/states.bin";
	$input_trans = "$input_dir/transitions.text";

	# debug message
	#
	print "INFO: bw_build_xwrd_models - iteration $j\n";

	for ($i = 0; $i <= $#cpus_train_list; $i++) {

	    # debug message
	    #
	    print "INFO: bw_build_xwrd_models - iteration $j cpu $sort_cpus_train[$i]\n";

	    # build the parameter file for training on cpus in batch mode 
	    #
	    $train_mode = "batch";
	    $input_dir_cpus = sprintf("%s%d", 
				      "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm"
				      , $j-1);
	    $output_dir_cpus = sprintf("%s%d", 
				       "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm", $j);
	    $mfc_list = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list";
	    $model_transcript_cpus = "$TI_PATH_TO_XWRD_CPUS[$i]/transcriptions/xwrd_transcriptions.text";
	    $occupancy_file = "";
	    $acc_file = "$TI_PATH_TO_XWRD_CPUS[$i]/r02/acc/acc.bin";
	    $acc_list_file = "";
	    $output_states = "";
	    $output_trans = "";
	    $param_file_cpus = "$input_dir_cpus/params.text";
	    $shell_script = "$input_dir_cpus/run_bw_train.sh";
	    $log_file = "$input_dir_cpus/bw_train.log";
	    $completion_file[$i] = "$input_dir_cpus/train_job_completed.text";
	    
	    # create the shell script that would run the bw_train
	    #
	    run_create_shell_script_train($shell_script, $param_file_cpus, 
					  $log_file, $completion_file[$i], 
					  $cpus_train_list[$i]);


	    # write the file and run bw_train in batch mode
	    #
	    (run_bw_train($train_mode, $mfc_list, 
			  $model_transcript_cpus,
			  $monophones_file, $models_file, $phones_file,
			  $lexicon_file, $input_states, $input_trans,
			  $output_states, $output_trans, $TI_BINARY_MODE,
			  $TI_BINARY_MODE, $context_mode, $min_mpd,
			  $beam_width, $min_occupancy, $min_model_count,
			  $vfloor_file, $occupancy_file, $acc_file,
			  $acc_list_file, $param_file_cpus, 
			  $input_feature_format, $compute_delta, $compute_acc,
			  $delta_win, $cpus_train_list[$i], $shell_script)
	     == 1) or return 0;
	}


        # debug message
	#
	print "INFO: bw_build_xwrd_models - iteration $j waiting for\n";
	print "the training jobs to complete\n";
	
	# wait for the jobs to finish 
	#
	$done = 0;
	while ($done == 0) {
	    sleep 1;
	    $done = 1;
	    for ($i = 0; $i <= $#cpus_train_list; $i++) {
		if ( ! -f $completion_file[$i]) {
		    $done = 0;
		}
	    }
	}
	
	# debug message
	#
	print "INFO: bw_build_xwrd_models - iteration $j combining accumulators\n";

	# build the parameter file to run bw_train in combine mode 
	#
	$train_mode = "combine";
	$mfc_list = "";
	$model_transcript = "";
	$acc_file = "";
	$acc_list_file = $r02_acc_list;
	$output_dir = sprintf("%s%d", "$TI_PATH_TO_XWRD/r02/hmm", $j);
	$output_states = "$output_dir/states.bin";
	$output_trans = "$output_dir/transitions.text";
	$param_file = "$input_dir/params.text";
	$shell_script = "";

	# write the occupancied to the occupancy file that will be used 
	# for state tying
	#
	$occupancy_file = "$TI_PATH_TO_XWRD/r02/hmm4/state_occupancy.text";

	# write the file and run bw_train in combine mode
	#
	(run_bw_train($train_mode, $mfc_list, 
		      $model_transcript, $monophones_file, $models_file,
		      $phones_file, $lexicon_file, $input_states, 
		      $input_trans, $output_states, $output_trans,
		      $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
		      $min_mpd, $beam_width, $min_occupancy,
		      $min_model_count, $vfloor_file, $occupancy_file,
		      $acc_file, $acc_list_file, $param_file, 
		      $input_feature_format, $compute_delta, $compute_acc,
                      $delta_win, $cpus_train_list[0], $shell_script) == 1) 
	    or return 0;
    }

    print "INFO: bw_build_xwrd_models - tying cross-word models\n";
    
    # create the occupancy list file
    #
    $occ_list_file = "$TI_PATH_TO_XWRD/r02/hmm4/occ_list.text";
    (isip_open(fp_occ_list, ">$occ_list_file") == 1) or return 0;
    
    # print out the state occupancy file name to the occupancy list
    #
    print fp_occ_list "$occupancy_file\n";
    (isip_close(fp_occ_list) == 1) or return 0;
    
    # declare variables specific to state-tying
    #
    $last_states_file = "$TI_PATH_TO_XWRD/r02/hmm4/states.bin";
    $tied_tree_file = "$TI_PATH_TO_XWRD/state_tying/output_trees.text";
    $tied_models_file = "$TI_PATH_TO_XWRD/state_tying/models.text";
    $tied_phones_file = "$TI_PATH_TO_XWRD/state_tying/phones.text";
    $tied_states_file = "$TI_PATH_TO_XWRD/state_tying/states.bin";
    $tied_cluster_list =
	"$TI_PATH_TO_XWRD/state_tying/xwrd_triphones.clist";
    $questions_file = "$TI_PATH_TO_XWRD/state_tying/isip_questions.text";
    $special_models_file =
	"$TI_PATH_TO_XWRD/state_tying/special_models.text";
    $num_context = "3";
    $param_file = "$TI_PATH_TO_XWRD/state_tying/params.text";
    
    # tie the states
    #
    (run_tie_state_train($monophones_file, $models_file, 
			 $last_states_file, $occ_list_file,
			 $tied_tree_file, $tied_models_file,
			 $tied_states_file, $tied_cluster_list, 
			 $TI_BINARY_MODE, $TI_BINARY_MODE, $questions_file,
			 $special_models_file, $split_threshold, 
			 $merge_threshold, $num_occ_threshold, $num_context, 
			 $param_file, $cpus_train_list[$0]) == 1) or return 0;
    
    # create the new triphone map
    #
    (run_create_triphone_map($monophones_file, $tied_cluster_list, "cd",
			     $tied_models_file,
			     $tied_phones_file, $sort_cpus_train[0])
     == 1) or return 0;
    
    # copy data into hmm5
    #
    (isip_copy($tied_states_file,
	       "$TI_PATH_TO_XWRD/r02/hmm5/states.bin") == 1) or return 0;
    (isip_copy("$TI_PATH_TO_XWRD/r02/hmm4/transitions.text",
	       "$TI_PATH_TO_XWRD/r02/hmm5/transitions.text") == 1)
	or return 0;
    
    # continue training triphone models. if state-tying was used, we'll run
    # four passes, otherwise we'll run 5. update the models and phones files
    # if necessary
    #     
    $start_iteration = 6; 
    $last_iteration = 10;
    $models_file = "$TI_PATH_TO_XWRD/state_tying/models.text";
    $phones_file = "$TI_PATH_TO_XWRD/state_tying/phones.text";

    # print an informative message
    #
    print "INFO: bw_build_xwrd_models - continuing cross-word triphone";
    print " training\n";

    # stop writing occupancies
    #
    $occupancy_file = "";

    # loop for four iterations
    # 
    for ($j = $start_iteration; $j <= $last_iteration; $j++) {
	
	# build the parameters that are required for both
	# batch and combine mode
	#
	$input_dir = sprintf("%s%d", "$TI_PATH_TO_XWRD/r02/hmm", $j-1);
	$input_states = "$input_dir/states.bin";
	$input_trans = "$input_dir/transitions.text";
	
	# debug message
	#
	print "INFO: bw_build_xwrd_models - iteration $j\n";

	for ($i = 0; $i <= $#cpus_train_list; $i++) {
	    
	    # debug message
	    #
	    print "INFO: bw_build_xwrd_models - iteration $j cpu $sort_cpus_train[$i]\n";

	    # build the parameter file for training on cpus in batch mode 
	    #
	    $train_mode = "batch";
	    $input_dir_cpus = sprintf("%s%d", 
				      "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm"
				      , $j-1);
	    $output_dir_cpus = sprintf("%s%d", 
				       "$TI_PATH_TO_XWRD_CPUS[$i]/r02/hmm", $j);
	    $mfc_list = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list";
	    $model_transcript_cpus = "$TI_PATH_TO_XWRD_CPUS[$i]/transcriptions/xwrd_transcriptions.text";
	    $acc_file = "$TI_PATH_TO_XWRD_CPUS[$i]/r02/acc/acc.bin";
	    $acc_list_file = "";
	    $output_states = "";
	    $output_trans = "";
	    $param_file_cpus = "$input_dir_cpus/params.text";
	    $shell_script = "$input_dir_cpus/run_bw_train.sh";
	    $log_file = "$input_dir_cpus/bw_train.log";
	    $completion_file[$i] = "$input_dir_cpus/train_job_completed.text";
	    
	    # create the shell script that would run the bw_train
	    #
	    run_create_shell_script_train($shell_script, $param_file_cpus, 
					  $log_file, $completion_file[$i], 
					  $cpus_train_list[$i]);


	    # write the file and run bw_train in batch mode
	    #
	    (run_bw_train($train_mode, $mfc_list, 
			  $model_transcript_cpus,
			  $monophones_file, $models_file, $phones_file,
			  $lexicon_file, $input_states, $input_trans,
			  $output_states, $output_trans, $TI_BINARY_MODE,
			  $TI_BINARY_MODE, $context_mode, $min_mpd,
			  $beam_width, $min_occupancy, $min_model_count,
			  $vfloor_file, $occupancy_file, $acc_file,
			  $acc_list_file, $param_file_cpus, 
			  $input_feature_format, $compute_delta, $compute_acc,
			  $delta_win, $cpus_train_list[$i], $shell_script)
	     == 1) or return 0;
	}


	# debug message
	#
	print "INFO: bw_build_xwrd_models - iteration $j waiting for\n";
	print "the training jobs to complete\n";
	
	# wait for the jobs to finish 
	#
	$done = 0;
	while ($done == 0) {
	    sleep 1;
	    $done = 1;
	    for ($i = 0; $i <= $#cpus_train_list; $i++) {
		if ( ! -f $completion_file[$i]) {
		    $done = 0;
		}
	    }
	}
	
	# debug message
	#
	print "INFO: bw_build_xwrd_models - iteration $j combining accumulators\n";

	# build the parameter file to run bw_train in combine mode 
	#
	$train_mode = "combine";
	$mfc_list = "";
	$model_transcript = "";
	$acc_file = "";
	$acc_list_file = $r02_acc_list;
	$output_dir = sprintf("%s%d", "$TI_PATH_TO_XWRD/r02/hmm", $j);
	$output_states = "$output_dir/states.bin";
	$output_trans = "$output_dir/transitions.text";
	$param_file = "$input_dir/params.text";
	$shell_script = "";

	# write the file and run bw_train in combine mode
	#
	(run_bw_train($train_mode, $mfc_list, 
		      $model_transcript, $monophones_file, $models_file,
		      $phones_file, $lexicon_file, $input_states, 
		      $input_trans, $output_states, $output_trans,
		      $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
		      $min_mpd, $beam_width, $min_occupancy,
		      $min_model_count, $vfloor_file, $occupancy_file,
		      $acc_file, $acc_list_file, $param_file, 
		      $input_feature_format, $compute_delta, $compute_acc,
                      $delta_win, $cpus_train_list[0], $shell_script) == 1) 
	    or return 0;
    }

    # copy the data to the mixture generation portion
    # 
    (isip_copy("$TI_PATH_TO_XWRD/r02/hmm10/transitions.text",
	       "$TI_PATH_TO_XWRD/r03/hmm0/transitions.text") == 1) or return 0;
    (isip_copy("$TI_PATH_TO_XWRD/r02/hmm10/states.bin",
	       "$TI_PATH_TO_XWRD/r03/hmm0/states.bin") == 1) or return 0;    

    # begin multi-mixture training
    #
    print "INFO: bw_build_xwrd_models - beginning mixture step-up";
    print " training upto $mixtures four passes each\n";

    # loop over the number of mixtures
    #
    $last_mix = 0;
    for ($mix_num = 0; $mix_num <= $#num_mixes; $mix_num++) {

	# find the mixture we should generate
	#
	$this_mix = $num_mixes[$mix_num];
	
	# debugging message
	#
	print "INFO: bw_build_xwrd_models - mixing up to $this_mix";
	print " mixtures\n";

	# create the filenames
	#
	$last_mix_file = sprintf("%s%d%s", "$TI_PATH_TO_XWRD/r03/hmm", 
				 $last_mix, "/states.bin");
	$this_mix_text_file = sprintf("%s%d%d%s", "$TI_PATH_TO_XWRD/r03/hmm", 
				      $this_mix, 0, "/states.text");
	$this_mix_bin_file = sprintf("%s%d%d%s", "$TI_PATH_TO_XWRD/r03/hmm", 
				     $this_mix, 0, "/states.bin");
	$last_mix_trans_file = sprintf("%s%d%s", "$TI_PATH_TO_XWRD/r03/hmm", 
				       $last_mix, "/transitions.text");
	$this_mix_trans_file = sprintf("%s%d%d%s", "$TI_PATH_TO_XWRD/r03/hmm", 
				      $this_mix, 0, "/transitions.text");

	# run the create_mixtures function
	#
	(run_create_mixtures($last_mix_file, $this_mix_text_file,
			     $this_mix, $cpus_train_list[0]) == 1) or return 0;

	# convert the ouput to binary and copy the transitions file over
	#
	(run_convert_mmf($TI_ASCII_MODE, $TI_BINARY_MODE, $this_mix_text_file,
			 $this_mix_bin_file, $cpus_train_list[0]) == 1) 
	    or return 0;

	(isip_copy($last_mix_trans_file,
		   $this_mix_trans_file) == 1) or return 0;

	# reset the last_mix flag
	#
	$last_mix = $this_mix * 10 + 4;

	# loop for four iterations
	# 
	for ($j = 1; $j <= 4; $j++) {
	
	    # build the parameters that are required for both
	    # batch and combine mode
	    #
	    $input_dir = sprintf("%s%d", "$TI_PATH_TO_XWRD/r03/hmm",
				 $this_mix * 10 + $j-1);
	    $input_states = "$input_dir/states.bin";
	    $input_trans = "$input_dir/transitions.text";

	    # debug message
	    #
	    print "INFO: bw_build_xwrd_models - iteration $j\n";
	    
	    for ($i = 0; $i <= $#cpus_train_list; $i++) {

		# debug message
		#
		print "INFO: bw_build_xwrd_models - iteration $j cpu $sort_cpus_train[$i]\n";
		
		# build the parameter file for training on cpus in batch mode 
		#
		$train_mode = "batch";
		$input_dir_cpus = sprintf("%s%d", 
					  "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm"
					  , $this_mix * 10 + $j-1);
		$output_dir_cpus = sprintf("%s%d", 
					   "$TI_PATH_TO_XWRD_CPUS[$i]/r03/hmm",
					  $this_mix *10 + $j);
		$mfc_list = "$TI_PATH_TO_CUR_DIR/$sort_cpus_train[$i]/data_generation/lists/train_mfc.list";
		$model_transcript_cpus = "$TI_PATH_TO_XWRD_CPUS[$i]/transcriptions/xwrd_transcriptions.text";
		$acc_file = "$TI_PATH_TO_XWRD_CPUS[$i]/r03/acc/acc.bin";
		$acc_list_file = "";
		$output_states = "";
		$output_trans = "";
		$param_file_cpus = "$input_dir_cpus/params.text";
		$shell_script = "$input_dir_cpus/run_bw_train.sh";
		$log_file = "$input_dir_cpus/bw_train.log";
		$completion_file[$i] = "$input_dir_cpus/train_job_completed.text";
	    
		# create the shell script that would run the bw_train
		#
		run_create_shell_script_train($shell_script, $param_file_cpus, 
					      $log_file, $completion_file[$i], 
					      $cpus_train_list[$i]);


		# write the file and run bw_train in batch mode
		#
		(run_bw_train($train_mode, $mfc_list, 
			      $model_transcript_cpus,
			      $monophones_file, $models_file, $phones_file,
			      $lexicon_file, $input_states, $input_trans,
			      $output_states, $output_trans, $TI_BINARY_MODE,
			      $TI_BINARY_MODE, $context_mode, $min_mpd,
			      $beam_width, $min_occupancy, $min_model_count,
			      $vfloor_file, $occupancy_file, $acc_file,
			      $acc_list_file, $param_file_cpus, 
			      $input_feature_format, $compute_delta, 
			      $compute_acc,
			      $delta_win, $cpus_train_list[$i], $shell_script)
		 == 1) or return 0;
	    }

	    # debug message
	    #
	    print "INFO: bw_build_xwrd_models - iteration $j waiting for\n";
	    print "the training jobs to complete\n";
	    
	    # wait for the jobs to finish 
	    #
	    $done = 0;
	    while ($done == 0) {
		sleep 1;
		$done = 1;
		for ($i = 0; $i <= $#cpus_train_list; $i++) {
		    if ( ! -f $completion_file[$i]) {
			$done = 0;
		    }
		}
	    }
	    
	    # debug message
	    #
	    print "INFO: bw_build_xwrd_models - iteration $j combining accumulators\n";
	    
	    # build the parameter file to run bw_train in combine mode 
	    #
	    $train_mode = "combine";
	    $mfc_list = "";
	    $model_transcript = "";
	    $acc_file = "";
	    $acc_list_file = $r03_acc_list;
	    $output_dir = sprintf("%s%d", "$TI_PATH_TO_XWRD/r03/hmm",
				  $this_mix * 10 + $j);
	    $output_states = "$output_dir/states.bin";
	    $output_trans = "$output_dir/transitions.text";
	    $param_file = "$input_dir/params.text";
	    $shell_script = "";
	    
	    # write the file and run bw_train in combine mode
	    #
	    (run_bw_train($train_mode, $mfc_list, 
			  $model_transcript, $monophones_file, $models_file,
			  $phones_file, $lexicon_file, $input_states, 
			  $input_trans, $output_states, $output_trans,
			  $TI_BINARY_MODE, $TI_BINARY_MODE, $context_mode,
			  $min_mpd, $beam_width, $min_occupancy,
			  $min_model_count, $vfloor_file, $occupancy_file,
			  $acc_file, $acc_list_file, $param_file, 
			  $input_feature_format, $compute_delta, $compute_acc,
			  $delta_win, $cpus_train_list[0], 
			  $shell_script) == 1) or return 0;
	}
    }
    
    print "INFO: bw_build_xwrd_models - completed building $mixtures mixture";
    print " models.\n";
    
    # copy final models
    #
    print "INFO: bw_build_xwrd_models - copying final models to";
    print " $TI_PATH_TO_FINAL_MODELS\n";

    # copy the necessary files
    #
    (isip_copy($monophones_file, 
	       "$TI_PATH_TO_FINAL_MODELS/monophones.text") == 1) or return 0;
    (isip_copy($models_file,
	       "$TI_PATH_TO_FINAL_MODELS/models.text") == 1)  or return 0;
    (isip_copy($phones_file,
	       "$TI_PATH_TO_FINAL_MODELS/phones.text") == 1) or return 0;
    (isip_copy($lexicon_file,
	       "$TI_PATH_TO_FINAL_MODELS/lexicon.text") == 1) or return 0;
    (isip_copy($special_models_file,
	       "$TI_PATH_TO_FINAL_MODELS/special_models.text") == 1)
	or return 0;
    (isip_copy($vfloor_file,
	       "$TI_PATH_TO_FINAL_MODELS/vfloor.text") == 1) or return 0;

    if ($mixtures == $TI_NUM_MIX_1) {
	(isip_copy("$TI_PATH_TO_XWRD/r02/hmm10/states.bin",
		   "$TI_PATH_TO_FINAL_MODELS/states.bin") == 1) or return 0;
	(isip_copy("$TI_PATH_TO_XWRD/r02/hmm10/transitions.text",
		   "$TI_PATH_TO_FINAL_MODELS/transitions.text") == 1) or return 0;
    }
    
    elsif ($mixtures == $TI_NUM_MIX_2) {
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm24/states.bin",
		   "$TI_PATH_TO_FINAL_MODELS/states.bin") == 1) or return 0;
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm24/transitions.text",
		   "$TI_PATH_TO_FINAL_MODELS/transitions.text") == 1) or return 0;
    }
    
    elsif ($mixtures == $TI_NUM_MIX_4) {
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm44/states.bin",
		   "$TI_PATH_TO_FINAL_MODELS/states.bin") == 1) or return 0;
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm44/transitions.text",
		   "$TI_PATH_TO_FINAL_MODELS/transitions.text") == 1) or return 0;
    }
    
    elsif ($mixtures == $TI_NUM_MIX_8) {
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm84/states.bin",
		   "$TI_PATH_TO_FINAL_MODELS/states.bin") == 1) or return 0;
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm84/transitions.text",
		   "$TI_PATH_TO_FINAL_MODELS/transitions.text") == 1) or return 0;
    }
    
    elsif ($mixtures == $TI_NUM_MIX_16) {
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm164/states.bin",
		   "$TI_PATH_TO_FINAL_MODELS/states.bin") == 1) or return 0;
	(isip_copy("$TI_PATH_TO_XWRD/r03/hmm164/transitions.text",
		   "$TI_PATH_TO_FINAL_MODELS/transitions.text") == 1) or return 0;
    }
    
    else {
	isip_die("Failed to copy $mixtures cross-word models to the final models directory\n");
    }

#----------------------------------------------------------------------------
#
# building state-tied testing models since we are dealing with 
# context-dependent phone models
#
#----------------------------------------------------------------------------

# find the state-tying path
#
$state_tie_path = "$TI_PATH_TO_XWRD/state_tying";
    
# print an informative message
#
print "INFO: wsj_train - creating test models from the state-tying";
print " decision trees in $state_tie_path\n";

# set the parameters needed for state-tying
#
$monophones_file = "$TI_PATH_TO_FINAL_MODELS/monophones.text";
$states_file = "$TI_PATH_TO_XWRD/final_models/states.bin";
$special_models_file = "$TI_PATH_TO_XWRD/final_models/special_models.text";
$questions_file = "$state_tie_path/isip_questions.text";
$num_context = "3";
$input_tree_file = "$state_tie_path/output_trees.text";
$input_phones_list = "$TI_PATH_TO_TRAIN/xwrd_tri/all_xwrd_triphones.list";
$output_models_file = "$state_tie_path/tied_models.text";
$phones_file = "$state_tie_path/tied_phones.text";
$output_cluster_list = "$state_tie_path/tied_models.clist";
$param_file = "$state_tie_path/params_state_tying.text";

# run tie_state
#
(run_tie_state_test($monophones_file, $models_file, 
		    $states_file, $input_tree_file, 
		    $input_phones_list, $output_models_file,
		    $output_cluster_list, $TI_BINARY_MODE,
		    $TI_BINARY_MODE, $questions_file,
		    $special_models_file, $split_threshold,
		    $merge_threshold, $num_occ_threshold, $num_context,
		    $param_file, $cpus_train_list[0]) == 1)
    or isip_die("ERROR: failed to create tied models for testing");

# set the new models file
#
$models_file = $output_models_file;

# create the new triphone map
#
(run_create_triphone_map($monophones_file, $output_cluster_list, "cd",
			 $models_file, $phones_file) == 1) or return 0;

print "INFO: bw_build_xwrd_models - Completed Training\n";


# exit gracefully
#
return 1;
}
